@@ -25,7 +25,7 @@
 
 /*
  * Escapes the field content string to avoid confusion with file comments.
- * Returned field must be free()d by caller.
+ * Returned field must be g_free()d by caller.
  */
 static char *escape(const char *str)
 {
@@ -52,7 +52,7 @@ static char *escape(const char *str)
 
     // Copy the input string to the resultant string, and escape all
     // occurences of \# and #.
-    char *result = (char*)xmalloc(strlen(str) + 1 + count);
+    char *result = (char*)g_malloc(strlen(str) + 1 + count);
 
     const char *src = str;
     char *dest = result;
@@ -131,9 +131,8 @@ static void write_crash_report_field(FILE *fp, problem_data_t *problem_data,
     if (!(value->flags & CD_FLAG_ISEDITABLE))
         fprintf(fp, _("# This field is read only\n"));
 
-    char *escaped_content = escape(value->content);
+    g_autofree char *escaped_content = escape(value->content);
     fprintf(fp, "%s\n", escaped_content);
-    free(escaped_content);
 }
 
 /*
@@ -230,18 +229,18 @@ static int read_crash_report_field(const char *text, problem_data_t *report,
     char newvalue[length + 1];
     strncpy(newvalue, textfield, length);
     newvalue[length] = '\0';
-    strtrim(newvalue);
+    g_strstrip(newvalue);
 
     char oldvalue[strlen(value->content) + 1];
     strcpy(oldvalue, value->content);
-    strtrim(oldvalue);
+    g_strstrip(oldvalue);
 
     // Return if no change in the contents detected.
     if (strcmp(newvalue, oldvalue) == 0)
         return 0;
 
-    free(value->content);
-    value->content = xstrdup(newvalue);
+    g_free(value->content);
+    value->content = g_strdup(newvalue);
     return 1;
 }
 
@@ -352,7 +351,7 @@ static int run_report_editor(problem_data_t *problem_data)
     FILE *fp = fdopen(fd, "w");
     if (!fp) /* errno is set */
     {
-        die_out_of_memory();
+        libreport_die_out_of_memory();
     }
 
     write_crash_report(problem_data, fp);
@@ -375,14 +374,13 @@ static int run_report_editor(problem_data_t *problem_data)
         return 2;
     }
 
-    off_t size = fstat_st_size_or_die(fileno(fp));
+    off_t size = libreport_fstat_st_size_or_die(fileno(fp));
     if (size > INT_MAX/4)
 	    size = INT_MAX/4; /* paranoia */
-    char *text = xmalloc(size + 1);
+    g_autofree char *text = g_malloc(size + 1);
     if (fread(text, 1, size, fp) != size)
     {
         error_msg("Can't read '%s'", filename);
-        free(text);
         fclose(fp);
         return 2;
     }
@@ -398,7 +396,6 @@ static int run_report_editor(problem_data_t *problem_data)
     remove_comments_and_unescape(text);
     // Updates the crash report from the file text.
     int report_changed = read_crash_report(problem_data, text);
-    free(text);
     if (report_changed)
         puts(_("\nThe report has been updated"));
     else
@@ -414,73 +411,67 @@ static void ask_for_missing_settings(const char *event_name)
 {
     for (int i = 0; i < 3; ++i)
     {
-        GHashTable *error_table = validate_event(event_name);
-        if (!error_table)
+        GList *err_list = NULL, *iter = NULL;
+        err_list = get_options_with_err_msg(event_name);
+        if(!err_list)
             return;
 
         event_config_t *event_config = get_event_config(event_name);
 
-        GHashTableIter iter;
-        char *opt_name, *err_msg;
-        g_hash_table_iter_init(&iter, error_table);
-        while (g_hash_table_iter_next(&iter, (void**)&opt_name, (void**)&err_msg))
+        for (iter = err_list; iter; iter = iter->next)
         {
-            event_option_t *opt = get_event_option_from_list(opt_name,
-                                                             event_config->options);
+            invalid_option_t *err_data = (invalid_option_t *)iter->data;
+            event_option_t *opt = get_event_option_from_list(err_data->invopt_name, event_config->options);
 
             free(opt->eo_value);
             opt->eo_value = NULL;
 
-            char *question = xasprintf("%s %s:",
+            g_autofree char *question = g_strdup_printf("%s %s:",
                                              ec_get_screen_name(event_config) ? ec_get_screen_name(event_config) : event_name,
                                              (opt->eo_label) ? opt->eo_label : opt->eo_name);
             switch (opt->eo_type) {
             case OPTION_TYPE_TEXT:
             case OPTION_TYPE_NUMBER:
-                opt->eo_value = ask(question);
+                opt->eo_value = libreport_ask(question);
                 break;
             case OPTION_TYPE_PASSWORD:
             {
-                opt->eo_value = ask_password(question);
-                /* Newline was not added by pressing Enter because ECHO was
-                   disabled, so add it now. */
-                puts("");
+                opt->eo_value = libreport_ask_password(question);
                 break;
             }
             case OPTION_TYPE_BOOL:
-                if (ask_yes_no(question))
-                    opt->eo_value = xstrdup("yes");
+                if (libreport_ask_yes_no(question))
+                    opt->eo_value = g_strdup("yes");
                 else
-                    opt->eo_value = xstrdup("no");
+                    opt->eo_value = g_strdup("no");
 
                 break;
             case OPTION_TYPE_HINT_HTML: /* TODO? */
             case OPTION_TYPE_INVALID:
                 break;
             };
-
-            free(question);
         }
 
-        g_hash_table_destroy(error_table);
+        g_list_free_full(err_list, (GDestroyNotify)free_invalid_options);
 
-        error_table = validate_event(event_name);
-        if (!error_table)
+        err_list = get_options_with_err_msg(event_name);
+        if (!err_list)
             return;
 
-        alert(_("Your input is not valid, because of:"));
-        g_hash_table_iter_init(&iter, error_table);
-        while (g_hash_table_iter_next(&iter, (void**)&opt_name, (void**)&err_msg))
+        libreport_alert(_("Your input is not valid, because of:"));
+        for (iter = err_list; iter; iter = iter -> next)
         {
-            char *msg = xasprintf(_("Bad value for '%s': %s"), opt_name, err_msg);
-            alert(msg);
-            free(msg);
+            invalid_option_t *err_data = (invalid_option_t *)iter->data;
+            g_autofree char *msg = g_strdup_printf(_("Bad value for '%s': %s"),
+                                    err_data->invopt_name,
+                                    err_data->invopt_error);
+            libreport_alert(msg);
         }
 
-        g_hash_table_destroy(error_table);
+        g_list_free_full(err_list, (GDestroyNotify)free_invalid_options);
     }
 
-    /* we ask for 3 times and still don't have valid infromation */
+    /* we ask for 3 times and still don't have valid information */
     error_msg_and_die("Invalid input, exiting.");
 }
 
@@ -493,7 +484,7 @@ struct logging_state {
 
 static char *do_log(char *log_line, void *param)
 {
-    client_log(log_line);
+    libreport_client_log(log_line);
     return log_line;
 }
 
@@ -525,15 +516,16 @@ static int is_not_reportable(problem_data_t *problem_data)
     if (not_reportable)
     {
         printf("%s\n", not_reportable);
-        return 1;
+        if (libreport_get_global_stop_on_not_reportable())
+            return 1;
     }
     return 0;
 }
 
 static int is_backtrace_rating_usable(event_config_t *config, problem_data_t *problem_data)
 {
-    char *usability_description = NULL;
-    char *usability_detail = NULL;
+    g_autofree char *usability_description = NULL;
+    g_autofree char *usability_detail = NULL;
     const bool usable_rating = check_problem_rating_usability(config, problem_data,
                                                               &usability_description,
                                                               &usability_detail);
@@ -544,9 +536,6 @@ static int is_backtrace_rating_usable(event_config_t *config, problem_data_t *pr
         printf("%s\n", usability_detail);
     }
 
-    free(usability_description);
-    free(usability_detail);
-
     return usable_rating;
 }
 
@@ -594,8 +583,7 @@ static int run_event_on_dir_name_batch(
                 const char *dump_dir_name,
                 const char *event_name)
 {
-    int retval = -1;
-    problem_data_t *problem_data = NULL;
+    g_autoptr(problem_data_t) problem_data = NULL;
 
     event_config_t *config = get_event_config(event_name);
     if (config)
@@ -603,10 +591,8 @@ static int run_event_on_dir_name_batch(
         if (config->ec_minimal_rating != 0)
         {
             problem_data = load_problem_data_if_not_yet(problem_data, dump_dir_name);
-            if (!problem_data)
-                goto ret;
-            if (!is_backtrace_rating_usable(config, problem_data))
-                goto ret;
+            if (!problem_data || !is_backtrace_rating_usable(config, problem_data))
+                return -1;
         }
 
         if (!config->ec_skip_review)
@@ -617,21 +603,12 @@ static int run_event_on_dir_name_batch(
 
             /* Is problem non-reportable? */
             problem_data = load_problem_data_if_not_yet(problem_data, dump_dir_name);
-            if (!problem_data)
-                goto ret;
-            if (is_not_reportable(problem_data))
-                goto ret;
+            if (!problem_data || is_not_reportable(problem_data))
+                return -1;
         }
     }
 
-    problem_data_free(problem_data);
-    problem_data = NULL;
-
-    retval = export_config_and_run_event(state, dump_dir_name, event_name);
-
- ret:
-    problem_data_free(problem_data);
-    return retval;
+    return export_config_and_run_event(state, dump_dir_name, event_name);
 }
 
 static int run_event_on_dir_name_interactively(
@@ -639,8 +616,7 @@ static int run_event_on_dir_name_interactively(
                 const char *dump_dir_name,
                 const char *event_name)
 {
-    int retval = -1;
-    problem_data_t *problem_data = NULL;
+    g_autoptr(problem_data_t) problem_data = NULL;
 
     event_config_t *config = get_event_config(event_name);
     if (config)
@@ -648,10 +624,8 @@ static int run_event_on_dir_name_interactively(
         if (config->ec_minimal_rating != 0)
         {
             problem_data = load_problem_data_if_not_yet(problem_data, dump_dir_name);
-            if (!problem_data)
-                goto ret;
-            if (!is_backtrace_rating_usable(config, problem_data))
-                goto ret;
+            if (!problem_data || !is_backtrace_rating_usable(config, problem_data))
+                return -1;
         }
 
         if (!config->ec_skip_review)
@@ -662,21 +636,18 @@ static int run_event_on_dir_name_interactively(
 
             /* Is problem non-reportable? */
             problem_data = load_problem_data_if_not_yet(problem_data, dump_dir_name);
-            if (!problem_data)
-                goto ret;
-            if (is_not_reportable(problem_data))
-                goto ret;
+            if (!problem_data || is_not_reportable(problem_data))
+                return -1;
         }
 
         if (config->ec_sending_sensitive_data)
         {
-            char *msg = xasprintf(_("Event '%s' requires permission to send possibly sensitive data."
+            g_autofree char *msg = g_strdup_printf(_("Event '%s' requires permission to send possibly sensitive data."
                                     " Do you want to continue?"),
                         ec_get_screen_name(config) ? ec_get_screen_name(config) : event_name);
-            bool ok = ask_yes_no(msg);
-            free(msg);
+            bool ok = libreport_ask_yes_no(msg);
             if (!ok)
-                goto ret;
+                return -1;
         }
 
         /* can't fail */
@@ -687,22 +658,14 @@ static int run_event_on_dir_name_interactively(
         {
             problem_data = load_problem_data_if_not_yet(problem_data, dump_dir_name);
             if (!problem_data)
-                goto ret;
-            retval = review_problem_data(dump_dir_name, problem_data);
+                return -1;
+            int retval = review_problem_data(dump_dir_name, problem_data);
             if (retval != 0)
-                /* review failed, an error message was already logged */
-                goto ret;
+                return retval;
         }
     }
 
-    problem_data_free(problem_data);
-    problem_data = NULL;
-
-    retval = export_config_and_run_event(state, dump_dir_name, event_name);
-
- ret:
-    problem_data_free(problem_data);
-    return retval;
+    return export_config_and_run_event(state, dump_dir_name, event_name);
 }
 
 static int choose_number_from_range(unsigned min, unsigned max, const char *message)
@@ -711,15 +674,19 @@ static int choose_number_from_range(unsigned min, unsigned max, const char *mess
     unsigned ii;
     for (ii = 0; ii < 3; ++ii)
     {
-        char *answer = ask(message);
+        char *answer = libreport_ask(message);
+        char *endptr;
 
-        picked = xatou(answer);
+        long picked_intermediate = g_ascii_strtoull(answer, &endptr, 10);
+        if (picked_intermediate >= 0 && picked_intermediate <= UINT_MAX && answer != endptr)
+            picked = (unsigned)picked_intermediate;
+        else
+            error_msg_and_die("expected number in range <0, %d>: '%s'", UINT_MAX, answer);
         if (min <= picked && picked <= max)
             return picked;
 
-        char *msg = xasprintf("%s (%u - %u)\n", _("You have chosen number out of range"), min, max);
-        alert(msg);
-        free(msg);
+        g_autofree char *msg = g_strdup_printf("%s (%u - %u)\n", _("You have chosen number out of range"), min, max);
+        libreport_alert(msg);
     }
 
     error_msg_and_die(_("Invalid input, exiting."));
@@ -732,7 +699,7 @@ static char *select_event_name(GList *list_options)
 
     /* Just one? */
     if (!list_options->next)
-        return xstrdup((char*)list_options->data);
+        return g_strdup((char*)list_options->data);
 
     int count = 0;
     for (GList *li = list_options; li; li = li->next)
@@ -745,14 +712,14 @@ static char *select_event_name(GList *list_options)
 
     const unsigned picked = choose_number_from_range(1, count, _("Select an event to run: "));
     GList *chosen = g_list_nth(list_options, picked - 1);
-    return xstrdup((char*)chosen->data);
+    return g_strdup((char*)chosen->data);
 }
 
 int select_and_run_one_event(const char *dump_dir_name, const char *pfx, int interactive)
 {
     GList *list_events = list_possible_events_glist(dump_dir_name, pfx);
-    char *event_name = select_event_name(list_events);
-    list_free_with_free(list_events);
+    g_autofree char *event_name = select_event_name(list_events);
+    g_list_free_full(list_events, free);
 
     struct run_event_state *run_state = new_run_event_state();
     run_state->logging_callback = do_log;
@@ -763,61 +730,80 @@ int select_and_run_one_event(const char *dump_dir_name, const char *pfx, int int
 
     free_run_event_state(run_state);
 
-    free(event_name);
-
     return r;
 }
 
-/*
- * Run event from chain. Perform the following steps for each event from chain.
- * 1. Terminates a chain run if a backtrace is not usable.
- * 2. Asks for missing settings.
- * 3. Performs review of problem data if event requires review.
- * 4. Runs events commands.
- * 5. Terminates a chain run if any event from the chain requires termination
- *    (i.e. if it existed with EXIT_STOP_EVENT_RUN).
- * 6. Terminates a chain run if any error occurs.
- * 7. Continues with processing of next event.
- */
-int run_event_chain(const char *dump_dir_name, GList *chain, int interactive)
+static int run_event_chain_real(struct run_event_state *run_state,
+                                const char             *dump_dir_name,
+                                GList                  *chain,
+                                int                     interactive)
 {
-    struct logging_state l_state;
-
-    struct run_event_state *run_state = new_run_event_state();
-    run_state->logging_callback = do_log2;
-    run_state->logging_param = &l_state;
-
+    struct logging_state *l_state = g_malloc0(sizeof(struct logging_state));
     int retval = 0;
+
+    /* Blergh. */
+    run_state->logging_callback = do_log2;
+    run_state->logging_param = l_state;
+
     for (GList *eitem = chain; eitem; eitem = g_list_next(eitem))
     {
-        l_state.output_was_produced = 0;
+        l_state->output_was_produced = false;
         const char *event_name = eitem->data;
-        retval = interactive
-                ? run_event_on_dir_name_interactively(run_state, dump_dir_name, event_name)
-                : run_event_on_dir_name_batch(run_state, dump_dir_name, event_name)
-                ;
+        log_info("running commands for event '%s'", event_name);
+
+        if (interactive)
+            retval = run_event_on_dir_name_interactively(run_state, dump_dir_name,
+			    event_name);
+        else
+            retval = run_event_on_dir_name_batch(run_state, dump_dir_name,
+			    event_name);
 
         if (retval < 0)
-            /* Nothing was run (bad backtrace, user declined, etc... */
+            /* Nothing was run (bad backtrace, user declined, etc.) */
             break;
         if (retval == 0 && run_state->children_count == 0)
         {
-            printf("Error: no processing is specified for event '%s'\n", event_name);
-            retval = 1;
+            /* Continue running the chain even if no actions were executed for
+             * this event.
+             */
+            log_warning("no actions matching this problem found for event '%s'",
+                    event_name);
+            continue;
         }
-        else
-        /* If program failed, or if it finished successfully without saying anything... */
-        if (retval != 0 || !l_state.output_was_produced)
+        else if (retval != 0 || !l_state->output_was_produced)
         {
-            char *msg = exit_status_as_string(event_name, run_state->process_status);
+            /* Program failed, or finished successfully with no output. */
+            g_autofree char *msg = libreport_exit_status_as_string(event_name, run_state->process_status);
             fputs(msg, stdout);
-            free(msg);
         }
         if (retval != 0)
             break;
     }
 
+    return retval;
+}
+
+/*
+ * Run events from a chain. Perform the following steps for each event:
+ * 1. Terminate the chain run if the backtrace is not usable.
+ * 2. Ask for missing settings.
+ * 3. Perform review of problem data if the event requires review.
+ * 4. Run event's commands.
+ * 5. Terminate the run if any event from the chain requires termination
+ *    (i.e. if it exited with EXIT_STOP_EVENT_RUN).
+ * 6. Terminate the run if any error occurred.
+ * 7. Continue processing next event.
+ */
+int run_event_chain(const char *dump_dir_name, GList *chain, int interactive)
+{
+    /* Expand *-wildcards in the names of events in the chain. */
+    GList *expanded_chain = expand_event_chain_wildcards(chain);
+    struct run_event_state *run_state = new_run_event_state();
+
+    int retval = run_event_chain_real(run_state, dump_dir_name, expanded_chain, interactive);
+
     free_run_event_state(run_state);
+    g_list_free_full(expanded_chain, free);
 
     return retval;
 }
@@ -836,7 +822,7 @@ static workflow_t *select_workflow(GHashTable *workflows)
     if (wf_cnt == 1)
     {
         workflow_t *wf_selected = (workflow_t *)wf_list->data;
-        log_notice("autoselected workflow: '%s'", (char *)wf_get_name(wf_selected));
+        log_info("autoselected workflow: '%s'", (char *)wf_get_name(wf_selected));
         g_list_free(wf_list);
         return wf_selected;
     }
@@ -861,7 +847,29 @@ static workflow_t *select_workflow(GHashTable *workflows)
 
 int select_and_run_workflow(const char *dump_dir_name, GHashTable *workflows, int interactive)
 {
-    workflow_t *value = select_workflow(workflows);
+    workflow_t *workflow;
+    GList *events;
+    const char *workflow_name;
+    char *environment_variable = NULL;
+    struct run_event_state *run_state;
+    int retval;
 
-    return ((value == NULL) ? -1 : run_event_chain(dump_dir_name, wf_get_event_names(value), interactive));
+    workflow = select_workflow(workflows);
+    if (NULL == workflow)
+    {
+        return -1;
+    }
+
+    events = wf_get_event_names(workflow);
+
+    workflow_name = wf_get_name(workflow);
+    environment_variable = g_strdup_printf("LIBREPORT_WORKFLOW=%s", workflow_name);
+
+    run_state = new_run_event_state();
+    g_ptr_array_add(run_state->extra_environment, environment_variable);
+
+    retval = run_event_chain_real(run_state, dump_dir_name, events, interactive);
+    free_run_event_state(run_state);
+
+    return retval;
 }
