@@ -66,7 +66,7 @@
 
    For example, "report_Bugzilla" item will have the lookup attribute
    like ("libreportEventConfig", "report_Bugzilla") and the secret value
-   like ("Bugzilla_URL=https://bugzilla.redhat.com\0Bugzilla_Login=foo\0")
+   like ("Bugzilla_URL=https://bugzilla.redhat.com\0Bugzilla_APIKey=foo\0")
 
 
 
@@ -208,9 +208,8 @@ static GVariant *dbus_call_sync(GDBusProxy *proxy, const gchar *method, GVariant
 static bool is_dbus_remote_error(GError *error, const char *type)
 {
     /* returns a malloced string and return value can be NULL */
-    char *remote_type = g_dbus_error_get_remote_error(error);
+    g_autofree char *remote_type = g_dbus_error_get_remote_error(error);
     const bool result = (remote_type && strcmp(type, remote_type) == 0);
-    g_free(remote_type);
     return result;
 }
 
@@ -220,7 +219,7 @@ static bool is_dbus_remote_error(GError *error, const char *type)
 
 static struct secrets_object *secrets_object_new_from_proxy(GDBusProxy *proxy)
 {
-    struct secrets_object *obj = xzalloc(sizeof(*obj));
+    struct secrets_object *obj = g_new0(struct secrets_object, 1);
     obj->proxy = proxy;
     obj->interface_name = g_dbus_proxy_get_interface_name(proxy);
 
@@ -305,7 +304,7 @@ static enum secrets_service_state secrets_service_connect(void)
 
     if (!g_connection)
     {
-        log("Failed to open connection to D-Bus session bus: %s", error->message);
+        log_warning("Failed to open connection to D-Bus session bus: %s", error->message);
         g_error_free(error);
         return SBS_UNAVAILABLE;
     }
@@ -1259,7 +1258,7 @@ static void load_event_options_from_item(GDBusProxy *session,
             if (option)
             {
                 free(option->eo_value);
-                option->eo_value = xstrdup(value);
+                option->eo_value = g_strdup(value);
                 log_info("loaded event option : '%s' => '%s'", name, option->eo_value);
             }
         }
@@ -1333,7 +1332,7 @@ static void load_event_options_from_item(GDBusProxy *session,
             }
             nelems -= sz;
 
-            char *name = xstrdup(data);
+            g_autofree char *name = g_strdup(data);
             char *value = strchr(name, SECRETS_OPTION_VALUE_DELIMITER);
             if (!value)
             {
@@ -1347,12 +1346,11 @@ static void load_event_options_from_item(GDBusProxy *session,
             if (option)
             {
                 free(option->eo_value);
-                option->eo_value = xstrdup(value);
+                option->eo_value = g_strdup(value);
                 log_info("loaded event option : '%s' => '%s'", name, option->eo_value);
             }
 
 next_option:
-            free(name);
             data += sz;
         }
 
@@ -1399,7 +1397,7 @@ static void load_event_config(gpointer key, gpointer value, gpointer user_data)
     event_config_t *const ec = (event_config_t *)value;
     struct secrets_object *const collection = (struct secrets_object *)user_data;
 
-    if (is_event_config_user_storage_available())
+    if (libreport_is_event_config_user_storage_available())
         load_settings(g_session_proxy, collection, event_name, ec);
 }
 
@@ -1491,7 +1489,7 @@ static void save_event_config(const char *event_name,
 /* Public interface                                                           */
 /******************************************************************************/
 
-bool is_event_config_user_storage_available()
+bool libreport_is_event_config_user_storage_available()
 {
     INITIALIZE_LIBREPORT();
 
@@ -1507,21 +1505,19 @@ bool is_event_config_user_storage_available()
  * @param name Event name
  * @param config Event config
  */
-void load_single_event_config_data_from_user_storage(event_config_t *config)
+void libreport_load_single_event_config_data_from_user_storage(event_config_t *config)
 {
     INITIALIZE_LIBREPORT();
 
-    GHashTable *tmp = g_hash_table_new_full(
+    g_autoptr(GHashTable) tmp = g_hash_table_new_full(
                 /*hash_func*/ g_str_hash,
                 /*key_equal_func:*/ g_str_equal,
                 /*key_destroy_func:*/ g_free,
                 /*value_destroy_func:*/ NULL);
 
-    g_hash_table_insert(tmp, xstrdup(ec_get_name(config)), config);
+    g_hash_table_insert(tmp, g_strdup(ec_get_name(config)), config);
 
-    load_event_config_data_from_user_storage(tmp);
-
-    g_hash_table_destroy(tmp);
+    libreport_load_event_config_data_from_user_storage(tmp);
 
     return;
 }
@@ -1531,11 +1527,11 @@ void load_single_event_config_data_from_user_storage(event_config_t *config)
  *
  * @param event_config_list Events configs
  */
-void load_event_config_data_from_user_storage(GHashTable *event_config_list)
+void libreport_load_event_config_data_from_user_storage(GHashTable *event_config_list)
 {
     INITIALIZE_LIBREPORT();
 
-    if (is_event_config_user_storage_available())
+    if (libreport_is_event_config_user_storage_available())
     {
         bool dismissed = false;
         struct secrets_object *collection = NULL;
@@ -1591,13 +1587,13 @@ void load_event_config_data_from_user_storage(GHashTable *event_config_list)
  * @param event_config Event data
  * @param store_passwords If TRUE stores options with passwords, otherwise skips them
  */
-void save_event_config_data_to_user_storage(const char *event_name,
+void libreport_save_event_config_data_to_user_storage(const char *event_name,
                                             const event_config_t *event_config,
                                             bool store_passwords)
 {
     INITIALIZE_LIBREPORT();
 
-    if (is_event_config_user_storage_available())
+    if (libreport_is_event_config_user_storage_available())
         save_event_config(event_name, event_config->options, store_passwords);
     else
     {
