@@ -38,7 +38,7 @@ static char*
 check_curl_error(CURLcode err, const char* msg)
 {
     if (err)
-        return xasprintf("%s: %s", msg, curl_easy_strerror(err));
+        return g_strdup_printf("%s: %s", msg, curl_easy_strerror(err));
     return NULL;
 }
 
@@ -79,7 +79,8 @@ xcurl_easy_setopt_off_t(CURL *handle, CURLoption option, curl_off_t parameter)
 
 CURLcode curl_easy_perform_with_proxy(CURL *handle, const char *url)
 {
-    GList *proxy_list, *li;
+    GList *proxy_list = NULL;
+    GList *li = NULL;
     CURLcode curl_err;
 
     proxy_list = get_proxy_list(url);
@@ -101,7 +102,7 @@ CURLcode curl_easy_perform_with_proxy(CURL *handle, const char *url)
         curl_err = curl_easy_perform(handle);
     }
 
-    list_free_with_free(proxy_list);
+    g_list_free_full(proxy_list, g_free);
 
     return curl_err;
 }
@@ -112,7 +113,7 @@ CURLcode curl_easy_perform_with_proxy(CURL *handle, const char *url)
 
 post_state_t *new_post_state(int flags)
 {
-    post_state_t *state = (post_state_t *)xzalloc(sizeof(*state));
+    post_state_t *state = g_new0(post_state_t, 1);
     state->flags = flags;
     return state;
 }
@@ -126,12 +127,12 @@ void free_post_state(post_state_t *state)
     if (headers)
     {
         while (*headers)
-            free(*headers++);
-        free(state->headers);
+            g_free(*headers++);
+        g_free(state->headers);
     }
-    free(state->curl_error_msg);
+    g_free(state->curl_error_msg);
     free(state->body);
-    free(state);
+    g_free(state);
 }
 
 char *find_header_in_post_state(post_state_t *state, const char *str)
@@ -143,7 +144,7 @@ char *find_header_in_post_state(post_state_t *state, const char *str)
         while (*headers)
         {
             if (strncmp(*headers, str, len) == 0)
-                return skip_whitespace(*headers + len);
+                return libreport_skip_whitespace(*headers + len);
             headers++;
         }
     }
@@ -161,7 +162,7 @@ save_headers(void *buffer_pv, size_t count, size_t nmemb, void *ptr)
     post_state_t* state = (post_state_t*)ptr;
     size_t size = count * nmemb;
 
-    char *h = xstrndup((char*)buffer_pv, size);
+    char *h = g_strndup((char*)buffer_pv, size);
     strchrnul(h, '\r')[0] = '\0';
     strchrnul(h, '\n')[0] = '\0';
 
@@ -182,13 +183,13 @@ save_headers(void *buffer_pv, size_t count, size_t nmemb, void *ptr)
         if (headers)
         {
             while (*headers)
-                free(*headers++);
+                g_free(*headers++);
         }
         cnt = 0;
     }
 
     log_debug("save_headers: header %d: '%s'", cnt, h);
-    state->headers = (char**)xrealloc(state->headers, (cnt+2) * sizeof(state->headers[0]));
+    state->headers = (char**)g_realloc(state->headers, (cnt+2) * sizeof(state->headers[0]));
     state->headers[cnt] = h;
     state->header_cnt = ++cnt;
     state->headers[cnt] = NULL;
@@ -223,8 +224,8 @@ static size_t fread_with_reporting(void *ptr, size_t size, size_t nmemb, void *u
     {
         last_t = t;
         report_interval *= 2;
-        off_t sz = fstat_st_size_or_die(fileno(fp));
-        log(_("Uploaded: %llu of %llu kbytes"),
+        off_t sz = libreport_fstat_st_size_or_die(fileno(fp));
+        log_warning(_("Uploaded: %llu of %llu kbytes"),
                 (unsigned long long)cur_pos / 1024,
                 (unsigned long long)sz / 1024);
     }
@@ -235,7 +236,7 @@ static size_t fread_with_reporting(void *ptr, size_t size, size_t nmemb, void *u
 
 static int curl_debug(CURL *handle, curl_infotype it, char *buf, size_t bufsize, void *unused)
 {
-    if (logmode == 0)
+    if (libreport_logmode == 0)
         return 0;
 
     unsigned orig_bufsize = bufsize;
@@ -267,25 +268,25 @@ static int curl_debug(CURL *handle, curl_infotype it, char *buf, size_t bufsize,
     switch (it) {
     case CURLINFO_TEXT: /* The data is informational text. */
         /* Here eol is always "^J" or "", not printing it */
-        log("curl: %.*s", (int) bufsize, buf);
+        log_warning("curl: %.*s", (int) bufsize, buf);
         break;
     case CURLINFO_HEADER_IN: /* The data is header (or header-like) data received from the peer. */
-        log("curl rcvd header: '%.*s%s'", (int) bufsize, buf, eol);
+        log_warning("curl rcvd header: '%.*s%s'", (int) bufsize, buf, eol);
         break;
     case CURLINFO_HEADER_OUT: /* The data is header (or header-like) data sent to the peer. */
-        log("curl sent header: '%.*s%s'", (int) bufsize, buf, eol);
+        log_warning("curl sent header: '%.*s%s'", (int) bufsize, buf, eol);
         break;
     case CURLINFO_DATA_IN: /* The data is protocol data received from the peer. */
-        if (g_verbose >= 3)
-            log("curl rcvd data: '%.*s%s'", (int) bufsize, buf, eol);
+        if (libreport_g_verbose >= 3)
+            log_warning("curl rcvd data: '%.*s%s'", (int) bufsize, buf, eol);
         else
-            log("curl rcvd data %u bytes", orig_bufsize);
+            log_warning("curl rcvd data %u bytes", orig_bufsize);
         break;
     case CURLINFO_DATA_OUT: /* The data is protocol data sent to the peer. */
-        if (g_verbose >= 3)
-            log("curl sent data: '%.*s%s'", (int) bufsize, buf, eol);
+        if (libreport_g_verbose >= 3)
+            log_warning("curl sent data: '%.*s%s'", (int) bufsize, buf, eol);
         else
-            log("curl sent data %u bytes", orig_bufsize);
+            log_warning("curl sent data %u bytes", orig_bufsize);
         break;
     default:
         break;
@@ -328,7 +329,7 @@ post(post_state_t *state,
     // Shut off the built-in progress meter completely
     xcurl_easy_setopt_long(handle, CURLOPT_NOPROGRESS, 1);
 
-    if (g_verbose >= 2)
+    if (libreport_g_verbose >= 2)
     {
         // "Display a lot of verbose information about its operations.
         // Very useful for libcurl and/or protocol debugging and understanding.
@@ -351,7 +352,13 @@ post(post_state_t *state,
         xcurl_easy_setopt_ptr(handle, CURLOPT_PASSWORD, (state->password ? state->password : ""));
     }
 
-    if (data_size != POST_DATA_FROMFILE_PUT)
+    /* set SSH public and private keyfile if configured */
+    if (state->client_ssh_public_keyfile)
+        xcurl_easy_setopt_ptr(handle, CURLOPT_SSH_PUBLIC_KEYFILE, state->client_ssh_public_keyfile);
+    if (state->client_ssh_private_keyfile)
+        xcurl_easy_setopt_ptr(handle, CURLOPT_SSH_PRIVATE_KEYFILE, state->client_ssh_private_keyfile);
+
+    if (data_size != POST_DATA_FROMFILE_PUT && data_size != POST_DATA_GET)
     {
         // Do a HTTP POST. This also makes curl use
         // a "Content-Type: application/x-www-form-urlencoded" header.
@@ -402,15 +409,7 @@ post(post_state_t *state,
         const char *basename = strrchr(data, '/');
         if (basename) basename++;
         else basename = data;
-#if 0
-        // Simple way, without custom reader function
-        CURLFORMcode curlform_err = curl_formadd(&post, &last,
-                        CURLFORM_PTRNAME, "file", // element name
-                        CURLFORM_FILE, data, // filename to read from
-                        CURLFORM_CONTENTTYPE, content_type,
-                        CURLFORM_FILENAME, basename, // filename to put in the form
-                        CURLFORM_END);
-#else
+
         data_file = fopen(data, "r");
         if (!data_file)
         {
@@ -433,7 +432,7 @@ post(post_state_t *state,
                         CURLFORM_CONTENTTYPE, content_type,
                         CURLFORM_FILENAME, basename, // filename to put in the form
                         CURLFORM_END);
-#endif
+
         if (curlform_err != 0)
 //FIXME:
             error_msg_and_die("out of memory or read error (curl_formadd error code: %d)", (int)curlform_err);
@@ -459,7 +458,7 @@ post(post_state_t *state,
             error_msg_and_die("out of memory or read error (curl_formadd error code: %d)", (int)curlform_err);
         xcurl_easy_setopt_ptr(handle, CURLOPT_HTTPPOST, post);
     }
-    else
+    else if (data_size != POST_DATA_GET)
     {
         // ...from a blob in memory
         xcurl_easy_setopt_ptr(handle, CURLOPT_POSTFIELDS, data);
@@ -474,12 +473,11 @@ post(post_state_t *state,
     if (data_size != POST_DATA_FROMFILE_AS_FORM_DATA
         && data_size != POST_DATA_STRING_AS_FORM_DATA)
     {
-        char *content_type_header = xasprintf("Content-Type: %s", content_type);
+        g_autofree char *content_type_header = g_strdup_printf("Content-Type: %s", content_type);
         // Note: curl_slist_append() copies content_type_header
         httpheader_list = curl_slist_append(httpheader_list, content_type_header);
         if (!httpheader_list)
             error_msg_and_die("out of memory");
-        free(content_type_header);
     }
 
     for (; additional_headers && *additional_headers; additional_headers++)
@@ -501,21 +499,6 @@ post(post_state_t *state,
 // apparently with POST->GET remapping - which server didn't like at all.
 // Attempted to suppress remapping on 305 using CURLOPT_POSTREDIR of -1,
 // but it still did not work.
-#if 0
-    // Please handle 301/302 redirects for me
-    xcurl_easy_setopt_long(handle, CURLOPT_FOLLOWLOCATION, 1);
-    xcurl_easy_setopt_long(handle, CURLOPT_MAXREDIRS, 10);
-    // Bitmask to control how libcurl acts on redirects after POSTs.
-    // Bit 0 set (value CURL_REDIR_POST_301) makes libcurl
-    // not convert POST requests into GET requests when following
-    // a 301 redirection. Bit 1 (value CURL_REDIR_POST_302) makes libcurl
-    // maintain the request method after a 302 redirect.
-    // CURL_REDIR_POST_ALL is a convenience define that sets both bits.
-    // The non-RFC behaviour is ubiquitous in web browsers, so the library
-    // does the conversion by default to maintain consistency.
-    // However, a server may require a POST to remain a POST.
-    xcurl_easy_setopt_long(handle, CURLOPT_POSTREDIR, -1L /*CURL_REDIR_POST_ALL*/ );
-#endif
 
     // Prepare for saving information
     if (state->flags & POST_WANT_HEADERS)
@@ -542,6 +525,8 @@ post(post_state_t *state,
         xcurl_easy_setopt_ptr(handle, CURLOPT_SSLCERT, state->client_cert_path);
         xcurl_easy_setopt_ptr(handle, CURLOPT_SSLKEY, state->client_key_path);
     }
+    if (state->cert_authority_cert_path)
+        xcurl_easy_setopt_ptr(handle, CURLOPT_CAINFO, state->cert_authority_cert_path);
 
     // This is the place where everything happens.
     // Here errors are not limited to "out of memory", can't just die.
@@ -585,40 +570,61 @@ post(post_state_t *state,
 /* Unlike post_file(),
  * this function will use PUT, not POST if url is "http(s)://..."
  */
-char *upload_file(const char *url, const char *filename)
+char *libreport_upload_file(const char *url, const char *filename)
 {
     post_state_t *state = new_post_state(POST_WANT_ERROR_MSG);
-    char *retval = upload_file_ext(state, url, filename, UPLOAD_FILE_NOFLAGS);
+    char *retval = libreport_upload_file_ext(state, url, filename, UPLOAD_FILE_NOFLAGS);
     free_post_state(state);
 
     return retval;
 }
 
-char *upload_file_ext(post_state_t *state, const char *url, const char *filename, int flags)
+char *libreport_upload_file_ext(post_state_t *state, const char *url, const char *filename, int flags)
 {
     /* we don't want to print the whole url as it may contain password
      * rhbz#856960
-     * there can be '@' in the login or password so let's try to find the
-     * first '@' from the end
+     *
+     * jfilak:
+     * We want to print valid URLs in useful messages.
+     *
+     * The old code had this approach:
+     *   there can be '@' in the login or password so let's try to find the
+     *   first '@' from the end
+     *
+     * The new implementation decomposes URI to its base elements and uses only
+     * scheme and hostname for the logging purpose. These elements should not
+     * contain any sensitive information.
      */
-    const char *clean_url = strrchr(url, '@');
-    if (clean_url)
-        clean_url++;
-    else
-        clean_url = url;
-
-    char *whole_url;
-    unsigned len = strlen(url);
-    if (len > 0 && url[len-1] == '/')
-        whole_url = concat_path_file(url, strrchr(filename, '/') ? : filename);
-    else
-        whole_url = xstrdup(url);
-
-
     const char *username_bck = state->username;
     const char *password_bck = state->password;
-    char *username = NULL;
-    char *password = NULL;
+
+    char *whole_url = NULL;
+    g_autofree char *scheme = NULL;
+    g_autofree char *hostname = NULL;
+    g_autofree char *username = NULL;
+    g_autofree char *password = NULL;
+    g_autofree char *clean_url = NULL;
+
+    if (libreport_uri_userinfo_remove(url, &clean_url, &scheme, &hostname, &username, &password, NULL) != 0)
+        goto finito;
+
+    if (scheme == NULL || hostname == NULL)
+    {
+        log_warning(_("Ignoring URL without scheme and hostname"));
+        goto finito;
+    }
+
+    if (username && (state->username == NULL || state->username[0] == '\0'))
+    {
+        state->username = username;
+        state->password = password;
+    }
+
+    unsigned len = strlen(clean_url);
+    if (len > 0 && clean_url[len-1] == '/')
+        whole_url = g_build_filename(clean_url, strrchr(filename, '/') ? : filename, NULL);
+    else
+        whole_url = g_strdup(clean_url);
 
     /* work around bug in libssh2(curl with scp://)
      * libssh2_aget_disconnect() calls close(0)
@@ -632,7 +638,9 @@ char *upload_file_ext(post_state_t *state, const char *url, const char *filename
      */
   do_post:
 
-    log(_("Sending %s to %s"), filename, clean_url);
+    /* Do not include the path part of the URL as it can contain sensitive data
+     * in case of typos */
+    log_warning(_("Sending %s to %s//%s"), filename, scheme, hostname);
     post(state,
                 whole_url,
                 /*content_type:*/ "application/octet-stream",
@@ -656,16 +664,14 @@ char *upload_file_ext(post_state_t *state, const char *url, const char *filename
                 (state->curl_result == CURLE_LOGIN_DENIED
                  || state->curl_result == CURLE_REMOTE_ACCESS_DENIED))
         {
-            char *msg = xasprintf(_("Please enter user name for '%s':"), clean_url);
-            free(username);
-            username = ask(msg);
-            free(msg);
+            char *msg = g_strdup_printf(_("Please enter user name for '%s//%s':"), scheme, hostname);
+            username = libreport_ask(msg);
+            g_free(msg);
             if (username != NULL && username[0] != '\0')
             {
-                msg = xasprintf(_("Please enter password for '%s':"), username);
-                free(password);
-                password = ask_password(msg);
-                free(msg);
+                msg = g_strdup_printf(_("Please enter password for '%s//%s@%s':"), scheme, username, hostname);
+                password = libreport_ask_password(msg);
+                g_free(msg);
                 /* What about empty password? */
                 if (password != NULL && password[0] != '\0')
                 {
@@ -678,21 +684,17 @@ char *upload_file_ext(post_state_t *state, const char *url, const char *filename
                 }
             }
         }
-
-        free(whole_url);
         whole_url = NULL;
     }
     else
     {
         /* This ends up a "reporting status message" in abrtd */
-        log(_("Successfully sent %s to %s"), filename, clean_url);
+        log_warning(_("Successfully created %s"), whole_url);
     }
 
     close(stdin_bck);
 
-    free(password);
-    free(username);
-
+finito:
     state->username = username_bck;
     state->password = password_bck;
 
