@@ -25,7 +25,6 @@
 #define LONG_DESCR_ELEMENT      "long-description"
 #define ALLOW_EMPTY_ELEMENT     "allow-empty"
 #define NOTE_HTML_ELEMENT       "note-html"
-#define CREATES_ELEMENT         "creates-items"
 #define OPTION_ELEMENT          "option"
 //#define ACTION_ELEMENT        "action"
 #define NAME_ELEMENT            "name"
@@ -33,6 +32,9 @@
 #define MINIMAL_RATING_ELEMENT  "minimal-rating"
 #define GUI_REVIEW_ELEMENTS     "gui-review-elements"
 #define SENDING_SENSITIVE_DATA_ELEMENT  "sending-sensitive-data"
+#define SUPPORTS_RESTRICTED_ACCESS_ELEMENT "support-restricted-access"
+#define RESTRICTED_ACCESS_OPTION_ATTR "optionname"
+#define REQUIRES_DETAILS        "requires-details"
 
 #define REQUIRES_ELEMENT        "requires-items"
 #define EXCL_BY_DEFAULT_ELEMENT "exclude-items-by-default"
@@ -93,7 +95,7 @@ static char *get_element_lang(struct my_parse_data *parse_data, const gchar **at
             if (strcmp(att_values[i], parse_data->cur_locale) == 0)
             {
                 log_debug("found translation for: %s", parse_data->cur_locale);
-                return xstrdup(att_values[i]);
+                return g_strdup(att_values[i]);
             }
 
             /* try to match shorter locale
@@ -103,13 +105,13 @@ static char *get_element_lang(struct my_parse_data *parse_data, const gchar **at
              && strncmp(att_values[i], parse_data->cur_locale, short_locale_end - parse_data->cur_locale) == 0
             ) {
                 log_debug("found translation for shortlocale: %s", parse_data->cur_locale);
-                return xstrndup(att_values[i], short_locale_end - parse_data->cur_locale);
+                return g_strndup(att_values[i], short_locale_end - parse_data->cur_locale);
             }
         }
     }
     /* if the element has no attribute then it's a default non-localized value */
     if (i == 0)
-        return xstrdup("");
+        return g_strdup("");
     /* if the element is in different language than the current locale */
     return NULL;
 }
@@ -135,7 +137,7 @@ static void consume_cur_option(struct my_parse_data *parse_data)
      * (strcmp would segfault, etc), so provide invented name:
      */
     if (!opt->eo_name)
-        opt->eo_name = xasprintf("%u", (unsigned)g_list_length(event_config->values->options));
+        opt->eo_name = g_strdup_printf("%u", (unsigned)g_list_length(event_config->values->options));
 
     GList *elem = g_list_find_custom(event_config->values->options, opt->eo_name, cmp_event_option_name_with_string);
     if (elem)
@@ -147,17 +149,16 @@ static void consume_cur_option(struct my_parse_data *parse_data)
             /* ...and it already has a value, which
              * overrides xml-defined default one:
              */
-            free(opt->eo_value);
+            g_free(opt->eo_value);
             opt->eo_value = old_opt->eo_value;
             old_opt->eo_value = NULL;
         }
-        //log("xml: replacing '%s' value:'%s'->'%s'", opt->eo_name, old_opt->eo_value, opt->eo_value);
+
         free_event_option(old_opt);
         elem->data = opt;
     }
     else
     {
-        //log("xml: new value %s='%s'", opt->eo_name, opt->eo_value);
         event_config->values->options = g_list_append(event_config->values->options, opt);
     }
 }
@@ -170,8 +171,6 @@ static void start_element(GMarkupParseContext *context,
                   gpointer user_data,
                   GError **error)
 {
-    //log("start: %s", element_name);
-
     struct my_parse_data *parse_data = user_data;
     if (strcmp(element_name, ADV_OPTIONS_ELEMENT) == 0)
     {
@@ -196,8 +195,8 @@ static void start_element(GMarkupParseContext *context,
             log_info("attr: %s:%s", attribute_names[i], attribute_values[i]);
             if (strcmp(attribute_names[i], "name") == 0)
             {
-                free(opt->eo_name);
-                opt->eo_name = xstrdup(attribute_values[i]);
+                g_free(opt->eo_name);
+                opt->eo_name = g_strdup(attribute_values[i]);
             }
             else if (strcmp(attribute_names[i], "type") == 0)
             {
@@ -222,7 +221,7 @@ static void start_element(GMarkupParseContext *context,
         }
 
         GList *head = parse_data->event_config.values->ec_imported_event_names;
-        parse_data->event_config.values->ec_imported_event_names = g_list_append(head, xstrdup(attribute_values[0]));
+        parse_data->event_config.values->ec_imported_event_names = g_list_append(head, g_strdup(attribute_values[0]));
     }
     else
     if (strcmp(element_name, LABEL_ELEMENT) == 0
@@ -231,9 +230,23 @@ static void start_element(GMarkupParseContext *context,
      || strcmp(element_name, NAME_ELEMENT) == 0
      || strcmp(element_name, NOTE_HTML_ELEMENT) == 0
     ) {
-        free(parse_data->attribute_lang);
+        g_free(parse_data->attribute_lang);
         parse_data->attribute_lang = get_element_lang(parse_data, attribute_names, attribute_values);
     }
+    else
+    if (strcmp(element_name, SUPPORTS_RESTRICTED_ACCESS_ELEMENT) == 0)
+    {
+        if (attribute_names[0] == NULL
+            || attribute_names[1] != NULL
+            || strcmp(attribute_names[0], RESTRICTED_ACCESS_OPTION_ATTR) != 0)
+        {
+            error_msg("XML event configuration error: '%s' element misses attribute '%s'",
+                    SUPPORTS_RESTRICTED_ACCESS_ELEMENT, RESTRICTED_ACCESS_OPTION_ATTR);
+            return;
+        }
+
+        parse_data->event_config.values->ec_restricted_access_option = g_strdup(attribute_values[0]);
+    }
 }
 
 // Called for close tags </foo>
@@ -244,7 +257,7 @@ static void end_element(GMarkupParseContext *context,
 {
     struct my_parse_data *parse_data = user_data;
 
-    free(parse_data->attribute_lang);
+    g_free(parse_data->attribute_lang);
     parse_data->attribute_lang = NULL;
 
     if (strcmp(element_name, ADV_OPTIONS_ELEMENT) == 0)
@@ -273,7 +286,7 @@ static void text(GMarkupParseContext *context,
     event_config_t *ui = parse_data->event_config.values;
 
     const gchar *inner_element = g_markup_parse_context_get_element(context);
-    char *text_copy = xstrndup(text, text_len);
+    char *text_copy = g_strndup(text, text_len);
     event_option_t *opt = parse_data->cur_option.values;
     if (opt)
     {
@@ -293,9 +306,8 @@ static void text(GMarkupParseContext *context,
                             (strcmp(parse_data->attribute_lang, parse_data->cur_locale) == 0);
 
                         log_info("new label:'%s'", text_copy);
-                        free(opt->eo_label);
-                        opt->eo_label = text_copy;
-                        text_copy = NULL;
+                        g_free(opt->eo_label);
+                        opt->eo_label = g_steal_pointer(&text_copy);
                     }
                 }
             }
@@ -309,9 +321,8 @@ static void text(GMarkupParseContext *context,
         else if (strcmp(inner_element, DEFAULT_VALUE_ELEMENT) == 0)
         {
             log_info("default value:'%s'", text_copy);
-            free(opt->eo_value);
-            opt->eo_value = text_copy;
-            text_copy = NULL;
+            g_free(opt->eo_value);
+            opt->eo_value = g_steal_pointer(&text_copy);
         }
         else if (strcmp(inner_element, NOTE_HTML_ELEMENT) == 0)
         {
@@ -329,9 +340,8 @@ static void text(GMarkupParseContext *context,
                             (strcmp(parse_data->attribute_lang, parse_data->cur_locale) == 0);
 
                         log_info("html note:'%s'", text_copy);
-                        free(opt->eo_note_html);
-                        opt->eo_note_html = text_copy;
-                        text_copy = NULL;
+                        g_free(opt->eo_note_html);
+                        opt->eo_note_html = g_steal_pointer(&text_copy);
                     }
                 }
             }
@@ -339,38 +349,12 @@ static void text(GMarkupParseContext *context,
         else if (strcmp(inner_element, ALLOW_EMPTY_ELEMENT) == 0)
         {
             log_info("allow-empty:'%s'", text_copy);
-            opt->eo_allow_empty = string_to_bool(text_copy);
+            opt->eo_allow_empty = libreport_string_to_bool(text_copy);
         }
-        /*
-        if (strcmp(inner_element, DESCRIPTION_ELEMENT) == 0)
-        {
-            log_info("tooltip:'%s'", text_copy);
-            free(opt->eo_description);
-            opt->eo_description = text_copy;
-            text_copy = NULL;
-        }
-        */
     }
     else
     {
-        /* we're not in option, so the description is for the event */
-        /*
-        if (strcmp(inner_element, ACTION_ELEMENT) == 0)
-        {
-            log_info("action description:'%s'", text_copy);
-            free(ui->action);
-            ui->action = text_copy;
-            text_copy = NULL;
-        }
-        */
-        if (strcmp(inner_element, CREATES_ELEMENT) == 0)
-        {
-            log_info("ec_creates_items:'%s'", text_copy);
-            free(ui->ec_creates_items);
-            ui->ec_creates_items = text_copy;
-            text_copy = NULL;
-        }
-        else if (strcmp(inner_element, NAME_ELEMENT) == 0)
+        if (strcmp(inner_element, NAME_ELEMENT) == 0)
         {
             if (parse_data->attribute_lang != NULL) /* if it isn't for other locale */
             {
@@ -387,8 +371,7 @@ static void text(GMarkupParseContext *context,
 
                         log_info("event name:'%s'", text_copy);
                         ec_set_screen_name(ui, text_copy);
-                        free(text_copy);
-                        text_copy = NULL;
+                        g_clear_pointer(&text_copy, g_free);
                     }
                 }
             }
@@ -411,8 +394,7 @@ static void text(GMarkupParseContext *context,
                             (strcmp(parse_data->attribute_lang, parse_data->cur_locale) == 0);
 
                         ec_set_description(ui, text_copy);
-                        free(text_copy);
-                        text_copy = NULL;
+                        g_clear_pointer(&text_copy, g_free);
                     }
                 }
             }
@@ -436,39 +418,34 @@ static void text(GMarkupParseContext *context,
                             (strcmp(parse_data->attribute_lang, parse_data->cur_locale) == 0);
 
                         ec_set_long_desc(ui, text_copy);
-                        free(text_copy);
-                        text_copy = NULL;
+                        g_clear_pointer(&text_copy, g_free);
                     }
                 }
             }
         }
         else if (strcmp(inner_element, REQUIRES_ELEMENT) == 0)
         {
-            free(ui->ec_requires_items);
-            ui->ec_requires_items = text_copy;
-            text_copy = NULL;
+            g_free(ui->ec_requires_items);
+            ui->ec_requires_items = g_steal_pointer(&text_copy);
         }
         else if (strcmp(inner_element, EXCL_BY_DEFAULT_ELEMENT) == 0)
         {
-            free(ui->ec_exclude_items_by_default);
-            ui->ec_exclude_items_by_default = text_copy;
-            text_copy = NULL;
+            g_free(ui->ec_exclude_items_by_default);
+            ui->ec_exclude_items_by_default = g_steal_pointer(&text_copy);
         }
         else if (strcmp(inner_element, INCL_BY_DEFAULT_ELEMENT) == 0)
         {
-            free(ui->ec_include_items_by_default);
-            ui->ec_include_items_by_default = text_copy;
-            text_copy = NULL;
+            g_free(ui->ec_include_items_by_default);
+            ui->ec_include_items_by_default = g_steal_pointer(&text_copy);
         }
         else if (strcmp(inner_element, EXCL_ALWAYS_ELEMENT) == 0)
         {
-            free(ui->ec_exclude_items_always);
-            ui->ec_exclude_items_always = text_copy;
-            text_copy = NULL;
+            g_free(ui->ec_exclude_items_always);
+            ui->ec_exclude_items_always = g_steal_pointer(&text_copy);
         }
         else if (strcmp(inner_element, EXCL_BINARY_ELEMENT) == 0)
         {
-            ui->ec_exclude_binary_items = string_to_bool(text_copy);
+            ui->ec_exclude_binary_items = libreport_string_to_bool(text_copy);
         }
         else if (strcmp(inner_element, MINIMAL_RATING_ELEMENT) == 0)
         {
@@ -477,20 +454,29 @@ static void text(GMarkupParseContext *context,
             ui->ec_minimal_rating = strtol(text_copy, &endptr, 10);
             if (errno != 0 || endptr == text_copy || *endptr != '\0')
             {
-                log("invalid minimal-rating number '%s', set to default 4", text_copy);
+                log_warning("invalid minimal-rating number '%s', set to default 4", text_copy);
                 ui->ec_minimal_rating = 4;
             }
         }
         else if (strcmp(inner_element, GUI_REVIEW_ELEMENTS) == 0)
         {
-            ui->ec_skip_review = !string_to_bool(text_copy);
+            ui->ec_skip_review = !libreport_string_to_bool(text_copy);
         }
         else if (strcmp(inner_element, SENDING_SENSITIVE_DATA_ELEMENT) == 0)
         {
-            ui->ec_sending_sensitive_data = string_to_bool(text_copy);
+            ui->ec_sending_sensitive_data = libreport_string_to_bool(text_copy);
+        }
+        else if (strcmp(inner_element, SUPPORTS_RESTRICTED_ACCESS_ELEMENT) == 0)
+        {
+            ui->ec_supports_restricted_access = libreport_string_to_bool(text_copy);
+        }
+        else if (strcmp(inner_element, REQUIRES_DETAILS) == 0)
+        {
+            ui->ec_requires_details = libreport_string_to_bool(text_copy);
         }
     }
-    free(text_copy);
+
+    g_free(text_copy);
 }
 
   // Called for strings that should be re-saved verbatim in this same
@@ -518,15 +504,20 @@ static void error(GMarkupParseContext *context,
 /* this function takes 2 parameters
  * ui -> pointer to event_config_t
  * filename -> filename to read
- * event_config_t contains list of options, which is malloced by hits function
+ * event_config_t contains list of options, which is malloced by this function
  * and must be freed by the caller
  */
-
 void load_event_description_from_file(event_config_t *event_config, const char* filename)
 {
-    log_notice("loading event: '%s'", filename);
-    struct my_parse_data parse_data = { {event_config, false, false, false}, {NULL, false, false}, NULL, NULL };
-    parse_data.cur_locale = xstrdup(setlocale(LC_ALL, NULL));
+    log_info("loading event: '%s'", filename);
+    struct my_parse_data parse_data = {
+        { event_config, false, false, false },
+        { NULL, false, false },
+        NULL,
+        NULL,
+        false
+    };
+    parse_data.cur_locale = g_strdup(setlocale(LC_ALL, NULL));
     strchrnul(parse_data.cur_locale, '.')[0] = '\0';
 
     GMarkupParser parser;
@@ -552,10 +543,12 @@ void load_event_description_from_file(event_config_t *event_config, const char*
         }
         fclose(fin);
     }
+    else
+        perror_msg("cannot load event description from file: '%s'", filename);
 
     g_markup_parse_context_free(context);
 
     consume_cur_option(&parse_data); /* just in case */
-    free(parse_data.attribute_lang); /* just in case */
-    free(parse_data.cur_locale);
+    g_free(parse_data.attribute_lang); /* just in case */
+    g_free(parse_data.cur_locale);
 }
