@@ -23,15 +23,15 @@ static void free_problem_item(void *ptr)
     if (ptr)
     {
         struct problem_item *item = (struct problem_item *)ptr;
-        free(item->content);
-        free(item);
+        g_free(item->content);
+        g_free(item);
     }
 }
 
 char *problem_item_format(struct problem_item *item)
 {
     if (!item)
-        return xstrdup("(nullitem)");
+        return g_strdup("(nullitem)");
 
     if (item->flags & CD_FLAG_UNIXTIME)
     {
@@ -46,7 +46,7 @@ char *problem_item_format(struct problem_item *item)
             char timeloc[256];
             int success = strftime(timeloc, sizeof(timeloc), "%c", localtime(&time));
             if (success)
-                return xstrdup(timeloc);
+                return g_strdup(timeloc);
         }
     }
     return NULL;
@@ -54,20 +54,27 @@ char *problem_item_format(struct problem_item *item)
 
 int problem_item_get_size(struct problem_item *item, unsigned long *size)
 {
+    if (item->size != PROBLEM_ITEM_UNINITIALIZED_SIZE)
+    {
+        *size = item->size;
+        return 0;
+    }
+
     if (item->flags & CD_FLAG_TXT)
     {
-        *size = strlen(item->content);
+        *size = item->size = strlen(item->content);
         return 0;
     }
 
     /* else if (item->flags & CD_FLAG_BIN) */
+
     struct stat statbuf;
     statbuf.st_size = 0;
 
     if (stat(item->content, &statbuf) != 0)
         return -errno;
 
-    *size = statbuf.st_size;
+    *size = item->size = statbuf.st_size;
     return 0;
 }
 
@@ -114,9 +121,7 @@ void problem_data_add_basics(problem_data_t *pd)
             problem_data_add_text_noteditable(pd, FILENAME_UUID, duphash);
         else
         {
-            /* start hash */
-            sha1_ctx_t sha1ctx;
-            sha1_begin(&sha1ctx);
+            g_autoptr(GChecksum) hash = g_checksum_new(G_CHECKSUM_SHA1);
 
             /*
              * To avoid spurious hash differences, sort keys so that elements are
@@ -136,16 +141,16 @@ void problem_data_add_basics(problem_data_t *pd)
                  */
                 if (item->flags & CD_FLAG_BIN)
                     continue;
-                sha1_hash(&sha1ctx, item->content, strlen(item->content));
+                g_checksum_update(hash, (unsigned char *)item->content,
+                                  strlen(item->content));
             }
             g_list_free(list);
 
-            /* end hash */
-            char hash_bytes[SHA1_RESULT_LEN];
-            sha1_end(&sha1ctx, hash_bytes);
-            char hash_str[SHA1_RESULT_LEN*2 + 1];
-            bin2hex(hash_str, hash_bytes, SHA1_RESULT_LEN)[0] = '\0';
+            // This string is owned by the hash object and will be freed
+            // automatically once it goes out of scope.
+            const char *hash_str = g_checksum_get_string(hash);
 
+            // hash_str is copied in this call.
             problem_data_add_text_noteditable(pd, FILENAME_UUID, hash_str);
         }
     }
@@ -181,20 +186,32 @@ void problem_data_add_current_process_data(problem_data_t *pd)
     }
 }
 
-void problem_data_add(problem_data_t *problem_data,
+struct problem_item *problem_data_add_ext(problem_data_t *problem_data,
                 const char *name,
                 const char *content,
-                unsigned flags)
+                unsigned flags,
+                unsigned long size)
 {
     if (!(flags & CD_FLAG_BIN))
         flags |= CD_FLAG_TXT;
     if (!(flags & CD_FLAG_ISEDITABLE))
         flags |= CD_FLAG_ISNOTEDITABLE;
 
-    struct problem_item *item = (struct problem_item *)xzalloc(sizeof(*item));
-    item->content = xstrdup(content);
+    struct problem_item *item = g_new0(struct problem_item, 1);
+    item->content = g_strdup(content);
     item->flags = flags;
-    g_hash_table_replace(problem_data, xstrdup(name), item);
+    item->size = size;
+    g_hash_table_replace(problem_data, g_strdup(name), item);
+
+    return item;
+}
+
+void problem_data_add(problem_data_t *problem_data,
+                const char *name,
+                const char *content,
+                unsigned flags)
+{
+    problem_data_add_ext(problem_data, name, content, flags, PROBLEM_ITEM_UNINITIALIZED_SIZE);
 }
 
 void problem_data_add_text_noteditable(problem_data_t *problem_data,
@@ -255,8 +272,10 @@ static const char *const editable_files[] = {
     //FILENAME_EXECUTABLE ,
     //FILENAME_BINARY     ,
     FILENAME_OPEN_FDS   , /* user might want to hide sensitive file names */
+    FILENAME_MOUNTINFO  , /* user might want to hide sensitive file paths */
     //FILENAME_LIMITS     ,
     FILENAME_CMDLINE    ,
+    FILENAME_CONTAINER_CMDLINE,
     //FILENAME_CGROUP     ,
     //FILENAME_COREDUMP   ,
     FILENAME_BACKTRACE  ,
@@ -282,51 +301,54 @@ static const char *const editable_files[] = {
     //FILENAME_EVENT_LOG  ,
     FILENAME_KICKSTART_CFG,
     FILENAME_ANACONDA_TB,
+    FILENAME_CPUINFO    , /* user might want to hide hw details not available to public */
     NULL
 };
 static bool is_editable_file(const char *file_name)
 {
-    return is_in_string_list(file_name, editable_files);
+    return libreport_is_in_string_list(file_name, editable_files);
 }
 
-/* When is_text_file() returns this special pointer value,
- * the file in question is "text, but very large".
- */
-#define HUGE_TEXT ((char*)(long)1)
-
 static const char *const always_text_files[] = {
     FILENAME_CMDLINE  ,
     FILENAME_BACKTRACE,
     FILENAME_OS_RELEASE,
     NULL
 };
-static char* is_text_file(const char *name, ssize_t *sz)
+static int is_text_file_at(int dir_fd, const char *name, char **content, ssize_t *sz, int *file_fd)
 {
     /* We were using magic.h API to check for file being text, but it thinks
      * that file containing just "0" is not text (!!)
      * So, we do it ourself.
      */
 
-    int fd = open(name, O_RDONLY);
+    int fd = secure_openat_read(dir_fd, name);
     if (fd < 0)
-        return NULL; /* it's not text (because it does not exist! :) */
+        return fd; /* it's not text (because it does not exist! :) */
 
     off_t size = lseek(fd, 0, SEEK_END);
     if (size < 0)
     {
         close(fd);
-        return NULL; /* it's not text (because there is an I/O error) */
+        return -EIO; /* it's not text (because there is an I/O error) */
     }
     lseek(fd, 0, SEEK_SET);
 
-    unsigned char *buf = xmalloc(*sz);
-    ssize_t r = full_read(fd, buf, *sz);
-    close(fd);
+    unsigned char *buf = g_malloc(*sz);
+    ssize_t r = libreport_full_read(fd, buf, *sz);
+
     if (r < 0)
     {
-        free(buf);
-        return NULL; /* it's not text (because we can't read it) */
+        close(fd);
+        g_free(buf);
+        return -EIO; /* it's not text (because we can't read it) */
     }
+
+    if (file_fd == NULL)
+        close(fd);
+    else
+        *file_fd = fd;
+
     if (r < *sz)
         buf[r] = '\0';
     *sz = r;
@@ -336,7 +358,7 @@ static char* is_text_file(const char *name, ssize_t *sz)
     if (base)
     {
         base++;
-        if (is_in_string_list(base, always_text_files))
+        if (libreport_is_in_string_list(base, always_text_files))
             goto text;
     }
 
@@ -364,8 +386,8 @@ static char* is_text_file(const char *name, ssize_t *sz)
             /* We don't like NULs and other control chars very much.
              * Not text for sure!
              */
-            free(buf);
-            return NULL;
+            g_free(buf);
+            return CD_FLAG_BIN;
         }
         if (buf[i] == 0x7f)
             bad_chars++;
@@ -386,16 +408,88 @@ static char* is_text_file(const char *name, ssize_t *sz)
     if ((total_chars / bad_chars) >= RATIO)
         goto text; /* looks like text to me */
 
-    free(buf);
-    return NULL; /* it's binary */
+    g_free(buf);
+    return CD_FLAG_BIN; /* it's binary */
 
  text:
     if (size > CD_MAX_TEXT_SIZE)
     {
-        free(buf);
-        return HUGE_TEXT;
+        g_free(buf);
+        return CD_FLAG_BIN | CD_FLAG_BIGTXT;
     }
-    return (char*)buf;
+
+    *content = /* cast from (unsigned char *) to */(char *)buf;
+    return CD_FLAG_TXT;
+}
+
+
+static int _problem_data_load_dump_dir_element(struct dump_dir *dd, const char *name, char **content, int *type_flags, int *fd)
+{
+    int file_fd = -1;
+    int *file_fd_ptr = fd == NULL ? &file_fd : fd;
+
+#define IS_TEXT_FILE_AT_PROBE_SIZE 4*1024
+
+    ssize_t sz = IS_TEXT_FILE_AT_PROBE_SIZE;
+    char *text = NULL;
+    int r = is_text_file_at(dd->dd_fd, name, &text, &sz, file_fd_ptr);
+
+    if (r < 0)
+        return r;
+
+    *type_flags = r;
+
+    if ((r == CD_FLAG_BIN) || (r == (CD_FLAG_BIN | CD_FLAG_BIGTXT)))
+        goto finito;
+
+    if (r != CD_FLAG_TXT)
+    {
+        error_msg("Unrecognized is_text_file_at() return value");
+        abort();
+    }
+
+    if (sz >= IS_TEXT_FILE_AT_PROBE_SIZE) /* did is_text_file() read entire file? */
+    {
+        /* no, it didn't, we need to read it all */
+        g_free(text);
+        lseek(*file_fd_ptr, 0, SEEK_SET);
+        text = libreport_xmalloc_read(*file_fd_ptr, NULL);
+    }
+
+#undef IS_TEXT_FILE_AT_PROBE_SIZE
+
+    /* Strip '\n' from one-line elements: */
+    char *nl = strchr(text, '\n');
+    if (nl && nl[1] == '\0')
+        *nl = '\0';
+
+    /* Sanitize possibly corrupted utf8.
+     * Of control chars, allow only tab and newline.
+     */
+    char *sanitized = libreport_sanitize_utf8(text,
+            (SANITIZE_ALL & ~SANITIZE_LF & ~SANITIZE_TAB)
+    );
+
+    if (sanitized != NULL)
+    {
+        g_free(text);
+        text = sanitized;
+    }
+
+finito:
+    if (file_fd >= 0)
+        close(file_fd);
+
+    *content = text;
+    return 0;
+}
+
+int problem_data_load_dump_dir_element(struct dump_dir *dd, const char *name, char **content, int *type_flags, int *fd)
+{
+    if (!libreport_str_is_correct_filename(name))
+        return -EINVAL;
+
+    return _problem_data_load_dump_dir_element(dd, name, content, type_flags, fd);
 }
 
 void problem_data_load_from_dump_dir(problem_data_t *problem_data, struct dump_dir *dd, char **excluding)
@@ -406,92 +500,65 @@ void problem_data_load_from_dump_dir(problem_data_t *problem_data, struct dump_d
     dd_init_next_file(dd);
     while (dd_get_next_file(dd, &short_name, &full_name))
     {
-        if (excluding && is_in_string_list(short_name, (const char *const *)excluding))
+        if (excluding && libreport_is_in_string_list(short_name, (const char *const *)excluding))
         {
-            //log("Excluded:'%s'", short_name);
+            //log_warning("Excluded:'%s'", short_name);
             goto next;
         }
 
         if (short_name[0] == '#'
          || (short_name[0] && short_name[strlen(short_name) - 1] == '~')
         ) {
-            //log("Excluded (editor backup file):'%s'", short_name);
+            //log_warning("Excluded (editor backup file):'%s'", short_name);
             goto next;
         }
 
-        ssize_t sz = 4*1024;
-        char *text = is_text_file(full_name, &sz);
-        if (!text || text == HUGE_TEXT)
-        {
-            int flag = !text ? CD_FLAG_BIN : (CD_FLAG_BIN+CD_FLAG_BIGTXT);
-            problem_data_add(problem_data,
-                    short_name,
-                    full_name,
-                    flag + CD_FLAG_ISNOTEDITABLE
-            );
-            goto next;
-        }
-
-        char *content;
-        if (sz < 4*1024) /* did is_text_file read entire file? */
-        {
-            /* yes */
-            content = text;
-        }
-        else
-        {
-            /* no, need to read it all */
-            free(text);
-            content = dd_load_text(dd, short_name);
-        }
-        /* Strip '\n' from one-line elements: */
-        char *nl = strchr(content, '\n');
-        if (nl && nl[1] == '\0')
-            *nl = '\0';
-
-        /* Sanitize possibly corrupted utf8.
-         * Of control chars, allow only tab and newline.
-         */
-        char *sanitized = sanitize_utf8(content,
-                (SANITIZE_ALL & ~SANITIZE_LF & ~SANITIZE_TAB)
-        );
-        if (sanitized)
-        {
-            free(content);
-            content = sanitized;
-        }
-
-        bool editable = is_editable_file(short_name);
+        char *content = NULL;
         int flags = 0;
-        if (editable)
-            flags |= CD_FLAG_TXT | CD_FLAG_ISEDITABLE;
+        int r = _problem_data_load_dump_dir_element(dd, short_name, &content, &flags, /*fd*/NULL);
+        if (r < 0)
+        {
+            error_msg("Failed to load element %s: %s", short_name, strerror(-r));
+            goto next;
+        }
+
+        if (flags & CD_FLAG_TXT)
+        {
+            if (is_editable_file(short_name))
+                flags |= CD_FLAG_ISEDITABLE;
+            else
+                flags |= CD_FLAG_ISNOTEDITABLE;
+
+            static const char *const list_files[] = {
+                FILENAME_UID       ,
+                FILENAME_PACKAGE   ,
+                FILENAME_CMDLINE   ,
+                FILENAME_TIME      ,
+                FILENAME_COUNT     ,
+                FILENAME_REASON    ,
+                NULL
+            };
+            if (libreport_is_in_string_list(short_name, list_files))
+                flags |= CD_FLAG_LIST;
+
+            if (strcmp(short_name, FILENAME_TIME) == 0)
+                flags |= CD_FLAG_UNIXTIME;
+        }
         else
-            flags |= CD_FLAG_TXT | CD_FLAG_ISNOTEDITABLE;
-
-        static const char *const list_files[] = {
-            FILENAME_UID       ,
-            FILENAME_PACKAGE   ,
-            FILENAME_CMDLINE   ,
-            FILENAME_TIME      ,
-            FILENAME_COUNT     ,
-            FILENAME_REASON    ,
-            NULL
-        };
-        if (is_in_string_list(short_name, list_files))
-            flags |= CD_FLAG_LIST;
-
-        if (strcmp(short_name, FILENAME_TIME) == 0)
-            flags |= CD_FLAG_UNIXTIME;
+        {
+            content = full_name;
+            full_name = NULL;
+        }
 
         problem_data_add(problem_data,
                 short_name,
                 content,
                 flags
         );
-        free(content);
+        g_free(content);
  next:
-        free(short_name);
-        free(full_name);
+        g_free(short_name);
+        g_free(full_name);
     }
 }
 
@@ -507,15 +574,15 @@ problem_data_t *create_problem_data_for_reporting(const char *dump_dir_name)
     struct dump_dir *dd = dd_opendir(dump_dir_name, /*flags:*/ 0);
     if (!dd)
         return NULL; /* dd_opendir already emitted error msg */
-    string_vector_ptr_t exclude_items = get_global_always_excluded_elements();
+    string_vector_ptr_t exclude_items = libreport_get_global_always_excluded_elements();
     problem_data_t *problem_data = problem_data_new();
     problem_data_load_from_dump_dir(problem_data, dd, exclude_items);
     dd_close(dd);
-    string_vector_free(exclude_items);
+    g_strfreev(exclude_items);
     return problem_data;
 }
 
-void log_problem_data(problem_data_t *problem_data, const char *pfx)
+void libreport_log_problem_data(problem_data_t *problem_data, const char *pfx)
 {
     GHashTableIter iter;
     char *name;
@@ -523,7 +590,7 @@ void log_problem_data(problem_data_t *problem_data, const char *pfx)
     g_hash_table_iter_init(&iter, problem_data);
     while (g_hash_table_iter_next(&iter, (void**)&name, (void**)&value))
     {
-        log("%s[%s]:'%s' 0x%x",
+        log_warning("%s[%s]:'%s' 0x%x",
                 pfx, name,
                 value->content,
                 value->flags
@@ -531,7 +598,7 @@ void log_problem_data(problem_data_t *problem_data, const char *pfx)
     }
 }
 
-gint cmp_problem_data(gconstpointer a, gconstpointer b, gpointer filename)
+gint libreport_cmp_problem_data(gconstpointer a, gconstpointer b, gpointer filename)
 {
     problem_data_t *a_data = *(problem_data_t **) a;
     const char *a_time_str = problem_data_get_content_or_NULL(a_data, filename);
@@ -552,12 +619,12 @@ gint cmp_problem_data(gconstpointer a, gconstpointer b, gpointer filename)
 }
 
 static bool problem_data_get_osinfo_from_items(problem_data_t *problem_data,
-        map_string_t *osinfo, const char *osinfo_name, const char *release_name)
+        GHashTable *osinfo, const char *osinfo_name, const char *release_name)
 {
     char *data = problem_data_get_content_or_NULL(problem_data, osinfo_name);
     if (data)
     {
-        parse_osinfo(data, osinfo);
+        libreport_parse_osinfo(data, osinfo);
         return true;
     }
 
@@ -565,11 +632,11 @@ static bool problem_data_get_osinfo_from_items(problem_data_t *problem_data,
     if (!data)
         return false;
 
-    insert_map_string(osinfo, xstrdup(OSINFO_PRETTY_NAME), xstrdup(data));
+    g_hash_table_insert(osinfo, g_strdup(OSINFO_PRETTY_NAME), g_strdup(data));
     return true;
 }
 
-void problem_data_get_osinfo(problem_data_t *problem_data, map_string_t *osinfo)
+void problem_data_get_osinfo(problem_data_t *problem_data, GHashTable *osinfo)
 {
     char *rootdir = problem_data_get_content_or_NULL(problem_data, FILENAME_ROOTDIR);
     if (rootdir &&
@@ -580,3 +647,38 @@ void problem_data_get_osinfo(problem_data_t *problem_data, map_string_t *osinfo)
     problem_data_get_osinfo_from_items(problem_data, osinfo,
                 FILENAME_OS_INFO, FILENAME_OS_RELEASE);
 }
+
+static const gchar *const reproducible_names[_PROBLEM_REPRODUCIBLE_MAX_] =
+{
+    "Not sure how to reproduce the problem",
+    "The problem is reproducible",
+    "The problem occurs regularly",
+};
+
+int get_problem_data_reproducible(problem_data_t *problem_data)
+{
+    const char *reproducible_str = problem_data_get_content_or_NULL(problem_data, FILENAME_REPRODUCIBLE);
+    if (reproducible_str == NULL)
+    {
+        log_info("Cannot return Reproducible type: missing "FILENAME_REPRODUCIBLE);
+        return -1;
+    }
+
+    for (int i = 0; i < _PROBLEM_REPRODUCIBLE_MAX_; ++i)
+        if (strcmp(reproducible_str, reproducible_names[i]) == 0)
+            return i;
+
+    error_msg("Cannot return Reproducible type: invalid format of '%s'", FILENAME_REPRODUCIBLE);
+    return -1;
+}
+
+const char *get_problem_data_reproducible_name(int reproducible)
+{
+    if (reproducible < 0 || reproducible >= _PROBLEM_REPRODUCIBLE_MAX_)
+    {
+        error_msg("Cannot return Reproducible name: invalid code: %d", reproducible);
+        return NULL;
+    }
+
+    return reproducible_names[reproducible];
+}
