@@ -21,6 +21,7 @@
 #include "internal_libreport.h"
 
 #include <satyr/stacktrace.h>
+#include <satyr/thread.h>
 #include <satyr/abrt.h>
 
 #include <assert.h>
@@ -33,7 +34,7 @@
  * |
  * |Package:: package
  * |
- * |%attach: %binary, backtrace
+ * |%attach:: %binary, backtrace
  * |
  * |%additional_info::
  * |%reporter%
@@ -98,8 +99,8 @@
  * }
  */
 struct section_t {
-    char *name;      ///< name or output text (%summar, 'Package version:');
-    GList *items;    ///< list of file names and special items (%reporter, %binar, ...)
+    char *name;      ///< name or output text (%summary, 'Package version:');
+    GList *items;    ///< list of file names and special items (%reporter, %binary, ...)
     GList *children; ///< list of sub sections (struct section_t)
 };
 
@@ -108,8 +109,8 @@ typedef struct section_t section_t;
 static section_t *
 section_new(const char *name)
 {
-    section_t *self = xmalloc(sizeof(*self));
-    self->name = xstrdup(name);
+    section_t *self = g_malloc(sizeof(*self));
+    self->name = g_strdup(name);
     self->items = NULL;
     self->children = NULL;
 
@@ -122,11 +123,11 @@ section_free(section_t *self)
     if (self == NULL)
         return;
 
-    free(self->name);
+    g_free(self->name);
     g_list_free_full(self->items, free);
     g_list_free_full(self->children, (GDestroyNotify)section_free);
 
-    free(self);
+    g_free(self);
 }
 
 static int
@@ -144,7 +145,7 @@ split_string_on_char(const char *str, char ch)
     for (;;)
     {
         const char *delim = strchrnul(str, ch);
-        list = g_list_prepend(list, xstrndup(str, delim - str));
+        list = g_list_prepend(list, g_strndup(str, delim - str));
         if (*delim == '\0')
             break;
         str = delim + 1;
@@ -203,10 +204,10 @@ load_stream(FILE *fp)
     sections = g_list_append(sections, master);
 
     char *line;
-    while ((line = xmalloc_fgetline(fp)) != NULL)
+    while ((line = libreport_xmalloc_fgetline(fp)) != NULL)
     {
         /* Skip comments */
-        char first = *skip_whitespace(line);
+        char first = *libreport_skip_whitespace(line);
         if (first == '#')
             goto free_line;
 
@@ -216,11 +217,10 @@ load_stream(FILE *fp)
         if (len && line[len-1] == '\\')
         {
             line[len-1] = '\0';
-            char *next_line = xmalloc_fgetline(fp);
+            g_autofree char *next_line = libreport_xmalloc_fgetline(fp);
             if (next_line)
             {
-                line = append_to_malloced_string(line, next_line);
-                free(next_line);
+                line = libreport_append_to_malloced_string(line, next_line);
                 goto check_continuation;
             }
         }
@@ -259,7 +259,7 @@ load_stream(FILE *fp)
         if (summary_line)
         {
             /* %summary is special */
-            item_list = g_list_append(NULL, xstrdup(skip_whitespace(value)));
+            item_list = g_list_append(NULL, g_strdup(libreport_skip_whitespace(value)));
         }
         else
         {
@@ -285,7 +285,7 @@ load_stream(FILE *fp)
             master->children = g_list_prepend(master->children, sec);
 
  free_line:
-        free(line);
+        g_free(line);
     }
 
     /* If master equals sec, then master's children list was not yet reversed.
@@ -304,13 +304,15 @@ load_stream(FILE *fp)
 
 #define MAX_OPT_DEPTH 10
 static int
-format_percented_string(const char *str, problem_data_t *pd, FILE *result)
+format_percented_string(const char *str, problem_data_t *pd, FILE *result, char *fmt_file)
 {
     long old_pos[MAX_OPT_DEPTH] = { 0 };
     int okay[MAX_OPT_DEPTH] = { 1 };
     long len = 0;
     int opt_depth = 1;
 
+    GList *missing_items = NULL;
+
     while (*str) {
         switch (*str) {
         default:
@@ -344,7 +346,8 @@ format_percented_string(const char *str, problem_data_t *pd, FILE *result)
                 opt_depth--;
                 if (!okay[opt_depth])
                 {
-                    fseek(result, old_pos[opt_depth], SEEK_SET);
+                    if (fseek(result, old_pos[opt_depth], SEEK_SET) < 0)
+                        perror_msg_and_die("fseek");
                     len = old_pos[opt_depth];
                 }
                 str += 2;
@@ -363,15 +366,35 @@ format_percented_string(const char *str, problem_data_t *pd, FILE *result)
 
             *nextpercent = '\0';
             const problem_item *item = problem_data_get_item_or_NULL(pd, str);
-            *nextpercent = '%';
 
-            if (item && (item->flags & CD_FLAG_TXT))
+            if (item)
             {
-                fputs(item->content, result);
-                len += strlen(item->content);
+                if (item->flags & CD_FLAG_TXT)
+                {
+                    fputs(item->content, result);
+                    len += strlen(item->content);
+                }
+                else if (fmt_file)
+                {
+                    error_msg_and_die("In format file '%s':\n"
+                                      "\t'%s' is not a text file",
+                                      fmt_file, str);
+                }
+                else
+                    error_msg_and_die("'%s' is not a text file", str);
             }
             else
+            {
                 okay[opt_depth - 1] = 0;
+                if (opt_depth > 1)
+                    log_debug("Missing content element: '%s'", str);
+                if (opt_depth == 1)
+                {
+                    log_debug("Missing top-level element: '%s'", str);
+                    missing_items = g_list_prepend(missing_items, g_strdup(str));
+                }
+            }
+            *nextpercent = '%';
             str = nextpercent + 1;
             break;
         }
@@ -384,7 +407,23 @@ format_percented_string(const char *str, problem_data_t *pd, FILE *result)
 
     if (!okay[0])
     {
-        error_msg("Undefined variable outside of [[ ]] bracket");
+        /* Items are stored in reverse order and then reversed due to how GLib stores
+         * the list internally.
+         */
+        missing_items = g_list_reverse(missing_items);
+
+        for (GList *item = missing_items; item != NULL; item = item->next)
+        {
+            if (fmt_file)
+                error_msg("In format file '%s':\n"
+                          "\tUndefined variable '%s' outside [[ ]] brackets",
+                          fmt_file, (char *)item->data);
+            else
+                error_msg("Undefined variable '%s' outside [[ ]] brackets",
+                          (char *)item->data);
+        }
+
+        g_list_free_full(missing_items, g_free);
     }
 
     return 0;
@@ -393,7 +432,7 @@ format_percented_string(const char *str, problem_data_t *pd, FILE *result)
 /* BZ comment generation */
 
 static int
-append_text(struct strbuf *result, const char *item_name, const char *content, bool print_item_name)
+append_text(GString *result, const char *item_name, const char *content, bool print_item_name)
 {
     char *eol = strchrnul(content, '\n');
     if (eol[0] == '\0' || eol[1] == '\0')
@@ -403,12 +442,12 @@ append_text(struct strbuf *result, const char *item_name, const char *content, b
         if (pad < 0)
             pad = 0;
         if (print_item_name)
-            strbuf_append_strf(result,
+            g_string_append_printf(result,
                     eol[0] == '\0' ? "%s: %*s%s\n" : "%s: %*s%s",
                     item_name, pad, "", content
             );
         else
-            strbuf_append_strf(result,
+            g_string_append_printf(result,
                     eol[0] == '\0' ? "%s\n" : "%s",
                     content
             );
@@ -417,11 +456,11 @@ append_text(struct strbuf *result, const char *item_name, const char *content, b
     {
         /* multi-line item */
         if (print_item_name)
-            strbuf_append_strf(result, "%s:\n", item_name);
+            g_string_append_printf(result, "%s:\n", item_name);
         for (;;)
         {
             eol = strchrnul(content, '\n');
-            strbuf_append_strf(result,
+            g_string_append_printf(result,
                     /* For %bare_multiline_item, we don't want to print colons */
                     (print_item_name ? ":%.*s\n" : "%.*s\n"),
                     (int)(eol - content), content
@@ -435,22 +474,29 @@ append_text(struct strbuf *result, const char *item_name, const char *content, b
 }
 
 static int
-append_short_backtrace(struct strbuf *result, problem_data_t *problem_data, size_t max_text_size, bool print_item_name)
+append_short_backtrace(GString *result, problem_data_t *problem_data, bool print_item_name, problem_report_settings_t *settings)
 {
-    const problem_item *item = problem_data_get_item_or_NULL(problem_data,
-                                                             FILENAME_BACKTRACE);
-    if (!item)
-        return 0; /* "I did not print anything" */
-    if (!(item->flags & CD_FLAG_TXT))
-        return 0; /* "I did not print anything" */
+    const problem_item *backtrace_item = problem_data_get_item_or_NULL(problem_data,
+                                                                       FILENAME_BACKTRACE);
+    const problem_item *core_stacktrace_item = NULL;
+    if (!backtrace_item || !(backtrace_item->flags & CD_FLAG_TXT))
+    {
+        backtrace_item = NULL;
 
-    char *truncated = NULL;
+        core_stacktrace_item = problem_data_get_item_or_NULL(problem_data,
+                                                             FILENAME_CORE_BACKTRACE);
 
-    if (strlen(item->content) >= max_text_size)
+        if (!core_stacktrace_item || !(core_stacktrace_item->flags & CD_FLAG_TXT))
+            return 0;
+    }
+
+    g_autofree char *truncated = NULL;
+
+    if (core_stacktrace_item || strlen(backtrace_item->content) >= settings->prs_shortbt_max_text_size)
     {
         log_debug("'backtrace' exceeds the text file size, going to append its short version");
 
-        char *error_msg = NULL;
+        g_autofree char *error_msg = NULL;
         const char *type = problem_data_get_content_or_NULL(problem_data, FILENAME_TYPE);
         if (!type)
         {
@@ -463,29 +509,32 @@ append_short_backtrace(struct strbuf *result, problem_data_t *problem_data, size
          * by default for CCpp crashes.
          */
         enum sr_report_type report_type = sr_abrt_type_from_type(type);
-        if (strcmp(type, "CCpp") == 0)
+        if (backtrace_item && strcmp(type, "CCpp") == 0)
         {
             log_debug("Successfully identified 'CCpp' abrt type");
             report_type = SR_REPORT_GDB;
         }
 
-        struct sr_stacktrace *backtrace = sr_stacktrace_parse(report_type,
-                item->content, &error_msg);
+        const char *content = backtrace_item ? backtrace_item->content : core_stacktrace_item->content;
+        struct sr_stacktrace *backtrace = sr_stacktrace_parse(report_type, content, &error_msg);
 
         if (!backtrace)
         {
-            log(_("Can't parse backtrace: %s"), error_msg);
-            free(error_msg);
+            log_warning(_("Can't parse backtrace: %s"), error_msg);
             return 0;
         }
 
-        /* Get optimized thread stack trace for 10 top most frames */
-        truncated = sr_stacktrace_to_short_text(backtrace, 10);
+        /* normalize */
+        struct sr_thread *thread = sr_stacktrace_find_crash_thread(backtrace);
+        sr_thread_normalize(thread);
+
+        /* Get optimized thread stack trace for max_frames top most frames */
+        truncated = sr_stacktrace_to_short_text(backtrace, settings->prs_shortbt_max_frames);
         sr_stacktrace_free(backtrace);
 
         if (!truncated)
         {
-            log(_("Can't generate stacktrace description (no crash thread?)"));
+            log_warning(_("Can't generate stacktrace description (no crash thread?)"));
             return 0;
         }
     }
@@ -494,17 +543,17 @@ append_short_backtrace(struct strbuf *result, problem_data_t *problem_data, size
         log_debug("'backtrace' is small enough to be included as is");
     }
 
+    /* full item content  */
     append_text(result,
                 /*item_name:*/ truncated ? "truncated_backtrace" : FILENAME_BACKTRACE,
-                /*content:*/   truncated ? truncated             : item->content,
+                /*content:*/   truncated ? truncated             : backtrace_item->content,
                 print_item_name
     );
-    free(truncated);
     return 1;
 }
 
 static int
-append_item(struct strbuf *result, const char *item_name, problem_data_t *pd, GList *comment_fmt_spec)
+append_item(GString *result, const char *item_name, problem_data_t *pd, GList *comment_fmt_spec, problem_report_settings_t *settings)
 {
     bool print_item_name = (strncmp(item_name, "%bare_", strlen("%bare_")) != 0);
     if (!print_item_name)
@@ -518,10 +567,9 @@ append_item(struct strbuf *result, const char *item_name, problem_data_t *pd, GL
         if (!(item->flags & CD_FLAG_TXT))
             return 0; /* "I did not print anything" */
 
-        char *formatted = problem_item_format(item);
+        g_autofree char *formatted = problem_item_format(item);
         char *content = formatted ? formatted : item->content;
         append_text(result, item_name, content, print_item_name);
-        free(formatted);
         return 1; /* "I printed something" */
     }
 
@@ -529,7 +577,7 @@ append_item(struct strbuf *result, const char *item_name, problem_data_t *pd, GL
 
     /* Compat with previously-existed ad-hockery: %short_backtrace */
     if (strcmp(item_name, "%short_backtrace") == 0)
-        return append_short_backtrace(result, pd, CD_TEXT_ATT_SIZE_BZ, print_item_name);
+        return append_short_backtrace(result, pd, print_item_name, settings);
 
     /* Compat with previously-existed ad-hockery: %reporter */
     if (strcmp(item_name, "%reporter") == 0)
@@ -541,7 +589,7 @@ append_item(struct strbuf *result, const char *item_name, problem_data_t *pd, GL
     bool text      = (strcmp(item_name+1, "text"     ) == 0);
     if (!oneline && !multiline && !text)
     {
-        log("Unknown or unsupported element specifier '%s'", item_name);
+        log_warning("Unknown or unsupported element specifier '%s'", item_name);
         return 0; /* "I did not print anything" */
     }
 
@@ -571,13 +619,12 @@ append_item(struct strbuf *result, const char *item_name, problem_data_t *pd, GL
         if (is_explicit_or_forbidden(name, comment_fmt_spec))
             continue;
 
-        char *formatted = problem_item_format(item);
+        g_autofree char *formatted = problem_item_format(item);
         char *content = formatted ? formatted : item->content;
         char *eol = strchrnul(content, '\n');
         bool is_oneline = (eol[0] == '\0' || eol[1] == '\0');
         if (oneline == is_oneline)
             printed |= append_text(result, name, content, print_item_name);
-        free(formatted);
     }
     if (text && oneline)
     {
@@ -600,7 +647,7 @@ append_item(struct strbuf *result, const char *item_name, problem_data_t *pd, GL
     } while (0)
 
 static void
-format_section(section_t *section, problem_data_t *pd, GList *comment_fmt_spec, FILE *result)
+format_section(section_t *section, problem_data_t *pd, GList *comment_fmt_spec, FILE *result, problem_report_settings_t *settings)
 {
     int empty_lines = -1;
 
@@ -610,7 +657,7 @@ format_section(section_t *section, problem_data_t *pd, GList *comment_fmt_spec,
         if (child->items)
         {
             /* "Text: item[,item]..." */
-            struct strbuf *output = strbuf_new();
+            g_autoptr(GString) output = g_string_new(NULL);
             GList *item = child->items;
             while (item)
             {
@@ -618,14 +665,12 @@ format_section(section_t *section, problem_data_t *pd, GList *comment_fmt_spec,
                 item = item->next;
                 if (str[0] == '-') /* "-name", ignore it */
                     continue;
-                append_item(output, str, pd, comment_fmt_spec);
+                append_item(output, str, pd, comment_fmt_spec, settings);
             }
 
             if (output->len != 0)
                 add_to_section_output((child->name[0] ? "%s:\n%s" : "%s%s"),
-                                      child->name, output->buf);
-
-            strbuf_free(output);
+                                      child->name, output->str);
         }
         else
         {
@@ -651,7 +696,7 @@ get_special_items(const char *item_name, problem_data_t *pd, GList *comment_fmt_
     bool binary    = (strcmp(item_name+1, "binary"   ) == 0);
     if (!oneline && !multiline && !text && !binary)
     {
-        log("Unknown or unsupported element specifier '%s'", item_name);
+        log_warning("Unknown or unsupported element specifier '%s'", item_name);
         return NULL;
     }
 
@@ -680,10 +725,10 @@ get_special_items(const char *item_name, problem_data_t *pd, GList *comment_fmt_
             char *eol = strchrnul(content, '\n');
             bool is_oneline = (eol[0] == '\0' || eol[1] == '\0');
             if (text || oneline == is_oneline)
-                result = g_list_append(result, xstrdup(name));
+                result = g_list_append(result, g_strdup(name));
         }
         else if ((item->flags & CD_FLAG_BIN) && binary)
-            result = g_list_append(result, xstrdup(name));
+            result = g_list_append(result, g_strdup(name));
     }
 
     g_list_free(sorted_names); /* names themselves are not freed */
@@ -708,7 +753,7 @@ get_attached_files(problem_data_t *pd, GList *items, GList *comment_fmt_spec)
 
         if (item_name[0] != '%')
         {
-            result = g_list_append(result, xstrdup(item_name));
+            result = g_list_append(result, g_strdup(item_name));
             continue;
         }
 
@@ -747,7 +792,7 @@ struct memstream_buffer
 static struct memstream_buffer *
 memstream_buffer_new()
 {
-    struct memstream_buffer *self = xmalloc(sizeof(*self));
+    struct memstream_buffer *self = g_new(struct memstream_buffer, 1);
 
     self->msb_buffer = NULL;
     self->msb_stream = open_memstream(&(self->msb_buffer), &(self->msb_size));
@@ -767,7 +812,7 @@ memstream_buffer_free(struct memstream_buffer *self)
     free(self->msb_buffer);
     self->msb_buffer = DESTROYED_POINTER;
 
-    free(self);
+    g_free(self);
 }
 
 static FILE *
@@ -785,6 +830,7 @@ memstream_get_string(struct memstream_buffer *self)
     assert(self->msb_stream != NULL);
 
     fflush(self->msb_stream);
+    self->msb_buffer[self->msb_size] = '\0';
 
     return self->msb_buffer;
 }
@@ -823,7 +869,7 @@ struct problem_report
 static problem_report_t *
 problem_report_new()
 {
-    problem_report_t *self = xmalloc(sizeof(*self));
+    problem_report_t *self = g_malloc(sizeof(*self));
 
     self->pr_sec_summ = memstream_buffer_new();
     self->pr_sec_desc = memstream_buffer_new();
@@ -869,7 +915,7 @@ problem_report_add_custom_section(problem_report_t *self, const char *name)
     }
 
     log_debug("Problem report enriched with section : '%s'", name);
-    g_hash_table_insert(self->pr_sec_custom, xstrdup(name), memstream_buffer_new());
+    g_hash_table_insert(self->pr_sec_custom, g_strdup(name), memstream_buffer_new());
     return 0;
 }
 
@@ -969,7 +1015,7 @@ problem_report_free(problem_report_t *self)
         self->pr_sec_custom = DESTROYED_POINTER;
     }
 
-    free(self);
+    g_free(self);
 }
 
 /*
@@ -984,9 +1030,9 @@ struct extra_section
 static struct extra_section *
 extra_section_new(const char *name, int flags)
 {
-    struct extra_section *self = xmalloc(sizeof(*self));
+    struct extra_section *self = g_new(struct extra_section, 1);
 
-    self->pfes_name = xstrdup(name);
+    self->pfes_name = g_strdup(name);
     self->pfes_flags = flags;
 
     return self;
@@ -998,10 +1044,10 @@ extra_section_free(struct extra_section *self)
     if (self == NULL)
         return;
 
-    free(self->pfes_name);
+    g_free(self->pfes_name);
     self->pfes_name = DESTROYED_POINTER;
 
-    free(self);
+    g_free(self);
 }
 
 static int
@@ -1020,14 +1066,38 @@ struct problem_formatter
     GList *pf_sections;         ///< parsed sections (struct section_t)
     GList *pf_extra_sections;   ///< user configured sections (struct extra_section)
     char  *pf_default_summary;  ///< default summary format
+    problem_report_settings_t pf_settings; ///< settings for report generating
+    char *fmt_file;
 };
 
+static problem_report_settings_t
+problem_report_settings_init(void)
+{
+    problem_report_settings_t settings = {
+        .prs_shortbt_max_frames = 10,
+        .prs_shortbt_max_text_size = CD_TEXT_ATT_SIZE_BZ,
+    };
+
+    return settings;
+}
+
+problem_report_settings_t problem_formatter_get_settings(const problem_formatter_t *self)
+{
+    return self->pf_settings;
+}
+
+void problem_formatter_set_settings(problem_formatter_t *self, problem_report_settings_t settings)
+{
+    self->pf_settings = settings;
+}
+
 problem_formatter_t *
 problem_formatter_new(void)
 {
-    problem_formatter_t *self = xzalloc(sizeof(*self));
+    problem_formatter_t *self = g_malloc0(sizeof(*self));
 
-    self->pf_default_summary = xstrdup("%reason%");
+    self->pf_default_summary = g_strdup("%reason%");
+    self->pf_settings = problem_report_settings_init();
 
     return self;
 }
@@ -1044,10 +1114,13 @@ problem_formatter_free(problem_formatter_t *self)
     g_list_free_full(self->pf_extra_sections, (GDestroyNotify)extra_section_free);
     self->pf_extra_sections = DESTROYED_POINTER;
 
-    free(self->pf_default_summary);
+    g_free(self->pf_default_summary);
     self->pf_default_summary = DESTROYED_POINTER;
 
-    free(self);
+    g_free(self->fmt_file);
+    self->fmt_file = DESTROYED_POINTER;
+
+    g_free(self);
 }
 
 static int
@@ -1149,6 +1222,7 @@ problem_formatter_load_file(problem_formatter_t *self, const char *path)
     }
 
     self->pf_sections = load_stream(fp);
+    self->fmt_file = g_strdup(path);
 
     if (fp != stdin)
         fclose(fp);
@@ -1160,6 +1234,8 @@ problem_formatter_load_file(problem_formatter_t *self, const char *path)
 int
 problem_formatter_generate_report(const problem_formatter_t *self, problem_data_t *data, problem_report_t **report)
 {
+    problem_report_settings_t settings = problem_formatter_get_settings(self);
+
     problem_report_t *pr = problem_report_new();
 
     for (GList *iter = self->pf_extra_sections; iter; iter = g_list_next(iter))
@@ -1175,7 +1251,7 @@ problem_formatter_generate_report(const problem_formatter_t *self, problem_data_
         {
             has_summary = true;
             format_percented_string((const char *)section->items->data, data,
-                                    problem_report_get_buffer(pr, PR_SEC_SUMMARY));
+                                    problem_report_get_buffer(pr, PR_SEC_SUMMARY), self->fmt_file);
         }
         /* %attach as well */
         else if (strcmp(section->name, "%attach") == 0)
@@ -1189,7 +1265,7 @@ problem_formatter_generate_report(const problem_formatter_t *self, problem_data_
             if (buffer != NULL)
             {
                 log_debug("Formatting section : '%s'", section->name);
-                format_section(section, data, self->pf_sections, buffer);
+                format_section(section, data, self->pf_sections, buffer, &settings);
             }
             else
                 log_warning("Unsupported section '%s'", section->name);
@@ -1201,7 +1277,7 @@ problem_formatter_generate_report(const problem_formatter_t *self, problem_data_
                     self->pf_default_summary);
 
         format_percented_string(self->pf_default_summary,
-                   data, problem_report_get_buffer(pr, PR_SEC_SUMMARY));
+                   data, problem_report_get_buffer(pr, PR_SEC_SUMMARY), self->fmt_file);
     }
 
     *report = pr;
