@@ -26,7 +26,7 @@ static void run_event_stdio_error_and_die(const char *error_line, void *param);
 
 struct run_event_state *new_run_event_state()
 {
-    struct run_event_state *state = xzalloc(sizeof(struct run_event_state));
+    struct run_event_state *state = g_new0(struct run_event_state, 1);
 
     state->logging_callback = run_event_stdio_log;
     state->error_callback = run_event_stdio_error_and_die;
@@ -35,21 +35,25 @@ struct run_event_state *new_run_event_state()
     state->ask_callback = run_event_stdio_ask;
     state->ask_yes_no_callback = run_event_stdio_ask_yes_no;
     state->ask_yes_no_yesforever_callback= run_event_stdio_ask_yes_no_yesforever;
+    state->ask_yes_no_save_result_callback= run_event_stdio_ask_yes_no_save_result;
     state->ask_password_callback = run_event_stdio_ask_password;
 
-    state->command_output = strbuf_new();
+    state->extra_environment = g_ptr_array_new_with_free_func(g_free);
+
+    state->command_output = g_string_new(NULL);
 
     return state;
 }
 
 void free_run_event_state(struct run_event_state *state)
 {
-    if (state)
-    {
-        strbuf_free(state->command_output);
-        free_commands(state);
-        free(state);
-    }
+    g_return_if_fail(state != NULL);
+
+    g_ptr_array_free(state->extra_environment, TRUE);
+    g_string_free(state->command_output, TRUE);
+    state->command_output = NULL;
+    free_commands(state);
+    g_free(state);
 }
 
 void make_run_event_state_forwarding(struct run_event_state *state)
@@ -59,13 +63,14 @@ void make_run_event_state_forwarding(struct run_event_state *state)
     state->ask_callback = run_event_stdio_ask;
     state->ask_yes_no_callback = run_event_stdio_ask_yes_no;
     state->ask_yes_no_yesforever_callback= run_event_stdio_ask_yes_no_yesforever;
+    state->ask_yes_no_save_result_callback= run_event_stdio_ask_yes_no_save_result;
     state->ask_password_callback = run_event_stdio_ask_password;
 
     /*
      * Not sure if we should reset even logging_callback and error_callback?
      */
 
-    xsetenv("REPORT_CLIENT_SLAVE", "1");
+    g_setenv("REPORT_CLIENT_SLAVE", "1", TRUE);
 }
 
 /* Asynchronous command execution */
@@ -109,24 +114,20 @@ void make_run_event_state_forwarding(struct run_event_state *state)
  *
  * List of commands machinery is encapsulated in struct run_event_state,
  * and public async API:
- *      prepare_commands(state, dir, event);
+ *      prepare_commands(state);
  *      spawn_next_command(state, dir, event, 0);
  *      free_commands(state);
  * does not expose the way we select rules to execute.
  */
-struct rule {
-    GList *conditions;
-    char *command; /* never NULL */
-};
 
-static void free_rule_list(GList *rule_list)
+void free_rule_list(GList *rule_list)
 {
     while (rule_list)
     {
         struct rule *cur_rule = rule_list->data;
-        list_free_with_free(cur_rule->conditions);
-        free(cur_rule->command);
-        free(cur_rule);
+        g_list_free_full(cur_rule->conditions, free);
+        g_free(cur_rule->command);
+        g_free(cur_rule);
 
         GList *next = rule_list->next;
         g_list_free_1(rule_list);
@@ -137,7 +138,7 @@ static void free_rule_list(GList *rule_list)
 /* Stop-gap measure against infinite recursion */
 #define MAX_recursion_depth 32
 
-static GList *load_rule_list(GList *rule_list,
+GList *load_rule_list(GList *rule_list,
                 const char *conf_file_name,
                 unsigned recursion_depth
 ) {
@@ -151,100 +152,158 @@ static GList *load_rule_list(GList *rule_list,
     /* Used only for better warning message */
     int line_counter = 0;
     /* Read and remember rules */
-    char *next_line = xmalloc_fgetline(conffile);
-    while (next_line)
+    char *next_line;
+    while (1)
     {
+        next_line = libreport_xmalloc_fgetline(conffile);
+        if(!next_line)
+        {
+            log_parser("EOF");
+            break;
+        }
+
+        if (*next_line == '\0' || *next_line == '#')
+        {
+            log_parser("empty or comment, skipping");
+            g_free(next_line);
+            continue;
+        }
+
+        log_parser("current line '%s'", next_line);
         /* Read and concatenate all lines in a rule */
         char *line = next_line;
         ++line_counter;
 
-        while (1)
+        if(strncmp(line, "EVENT", strlen("EVENT")) == 0)
         {
-            next_line = xmalloc_fgetline(conffile);
-            if (!next_line || !isblank(next_line[0]))
-                break;
+            log_parser("found EVENT");
 
-            ++line_counter;
-            char *old_line = line;
-            line = xasprintf("%s\n%s", line, next_line);
-            free(old_line);
-            free(next_line);
+            while (1)
+            {
+                long prev = ftell(conffile);
+                if (prev < 0)
+                    perror_msg_and_die("ftell");
+
+                next_line = libreport_xmalloc_fgetline(conffile);
+
+                log_parser("next_line is: '%s'", next_line ? next_line : "EOF");
+                /* stop merging new lines into this event
+                 * if we reach
+                 * EOF
+                 * line starting with:
+                 ** # (comment)
+                 ** EVENT
+                 ** include
+                 *
+                 * When adding another directive don't forget to add it to this if!
+                 */
+                if (    !next_line
+                     || *next_line == '#'
+                     || strncmp(next_line, "EVENT", strlen("EVENT")) == 0
+                     || strncmp(next_line, "include", strlen("include")) == 0
+                ){
+                    log_parser("found next EVENT or EOF, seeking back to prev line");
+
+                    if (fseek(conffile, prev, SEEK_SET) < 0)
+                        perror_msg_and_die("fseek");
+
+                    g_free(next_line);
+                    break;
+                }
+
+                ++line_counter;
+                char *tmp = g_strdup_printf("%s\n%s", line, next_line);
+                g_free(line);
+                g_free(next_line);
+                line = tmp;
+            }
+
+            char *p = libreport_skip_whitespace(line);
+
+            /* Rule has form: [VAR=VAL]... PROG [ARGS] */
+            struct rule *cur_rule = g_new0(struct rule, 1);
+
+            while (1) /* word loop */
+            {
+                log_parser("word loop, p is: '%s'", p);
+                char *end_word = libreport_skip_non_whitespace(p);
+                log_parser("end word: '%s'", end_word);
+
+                /* If there is no '=' in this word... */
+                char *line_val = strchr(p, '=');
+                log_parser("line val: '%s'", line_val);
+                if (!line_val || line_val >= end_word)
+                {
+                    log_parser("found start of a command");
+                    p = libreport_skip_whitespace(p);
+                    break; /* ...we found the start of a command */
+                }
+
+                char *const current_word = g_strndup(p, end_word - p);
+                cur_rule->conditions = g_list_append(cur_rule->conditions, current_word);
+                log_parser("adding condition '%s'", current_word);
+
+                /* Go to next word, but don't cross lines */
+                p = libreport_skip_blank(end_word);
+            } /* end of word loop */
+
+            if (g_list_length(cur_rule->conditions) == 0)
+            {
+                log_debug("Adding '%s' without conditions", p);
+            }
+            else if (libreport_g_verbose >= 3)
+            {
+                log_warning("Adding '%s'\nwith conditions:", p);
+                for (GList *c = cur_rule->conditions; c != NULL; c = g_list_next(c))
+                    log_warning("| %s", (const char *)c->data);
+            }
+
+            cur_rule->command = g_strdup(p);
+
+            rule_list = g_list_append(rule_list, cur_rule);
         }
-
-        char *p = skip_whitespace(line);
-        if (*p == '\0' || *p == '#')
-            goto next_line; /* empty or comment line, skip */
-
-        //log_debug("%s: line '%s'", __func__, p);
-
-        /* Handle "include" directive */
-        if (recursion_depth < MAX_recursion_depth
-         && strncmp(p, "include", strlen("include")) == 0
-         && isblank(p[strlen("include")])
+        else if (   recursion_depth < MAX_recursion_depth
+                 && strncmp(line, "include", strlen("include")) == 0
+                 && isblank(line[strlen("include")])
         ) {
+            log_parser("found include");
+
             /* "include GLOB_PATTERN" */
-            p = skip_whitespace(p + strlen("include"));
+            const char *p = libreport_skip_whitespace(line + strlen("include"));
 
             const char *last_slash;
-            char *name_to_glob;
+            g_autofree char *name_to_glob = NULL;
             if (*p != '/'
              && (last_slash = strrchr(conf_file_name, '/')) != NULL
             )
                 /* GLOB_PATTERN is relative, and this include is in path/to/file.conf
                  * Construct path/to/GLOB_PATTERN:
                  */
-                name_to_glob = xasprintf("%.*s%s", (int)(last_slash - conf_file_name + 1), conf_file_name, p);
+                name_to_glob = g_strdup_printf("%.*s%s", (int)(last_slash - conf_file_name + 1), conf_file_name, p);
             else
                 /* Either GLOB_PATTERN is absolute, or this include is in file.conf
                  * (no slashes in its name). Use unchanged GLOB_PATTERN:
                  */
-                name_to_glob = xstrdup(p);
+                name_to_glob = g_strdup(p);
 
             glob_t globbuf;
             memset(&globbuf, 0, sizeof(globbuf));
-            //log_debug("%s: globbing '%s'", __func__, name_to_glob);
+            log_parser("globbing '%s'", name_to_glob);
             glob(name_to_glob, 0, NULL, &globbuf);
-            free(name_to_glob);
             char **name = globbuf.gl_pathv;
             if (name) while (*name)
             {
-                //log_debug("%s: recursing into '%s'", __func__, *name);
+                log_parser("recursing into '%s'", *name);
                 rule_list = load_rule_list(rule_list, *name, recursion_depth + 1);
-                //log_debug("%s: returned from '%s'", __func__, *name);
+                log_parser("returned from '%s'", *name);
                 name++;
             }
             globfree(&globbuf);
-            goto next_line;
         }
+        else
+            log_parser("Unknown line found, ignoring: '%s'", line);
 
-        /* Rule has form: [VAR=VAL]... PROG [ARGS] */
-        struct rule *cur_rule = xzalloc(sizeof(*cur_rule));
-
-        while (1) /* word loop */
-        {
-            char *end_word = skip_non_whitespace(p);
-
-            /* If there is no '=' in this word... */
-            char *line_val = strchr(p, '=');
-            if (!line_val || line_val >= end_word)
-                break; /* ...we found the start of a command */
-
-            cur_rule->conditions = g_list_append(cur_rule->conditions, xstrndup(p, end_word - p));
-
-            /* Go to next word */
-            p = skip_whitespace(end_word);
-        } /* end of word loop */
-
-        if (cur_rule->conditions == NULL)
-            log_warning("%s:%d: warning: command without conditions, this command will be executed for all events", conf_file_name, line_counter);
-
-        log_notice("Adding '%s'", p);
-        cur_rule->command = xstrdup(p);
-
-        rule_list = g_list_append(rule_list, cur_rule);
-
- next_line:
-        free(line);
+        g_free(line);
     } /* end of line loop */
 
     fclose(conffile);
@@ -256,7 +315,7 @@ static int regcmp_lines(char *val, const char *regex)
 {
     regex_t rx;
     int r = regcomp(&rx, regex, REG_NOSUB); //TODO: and REG_EXTENDED?
-    //log("REGEX:'%s':%d", regex, r);
+    //log_warning("REGEX:'%s':%d", regex, r);
     if (r)
     {
         //char errbuf[256];
@@ -272,7 +331,7 @@ static int regcmp_lines(char *val, const char *regex)
         if (eol)
             *eol = '\0';
         r = regexec(&rx, val, 0, NULL, /*eflags:*/ 0);
-        //log("REGCMP:'%s':%d", val, r);
+        //log_warning("REGCMP:'%s':%d", val, r);
         if (eol)
             *eol = '\n';
         if (r == 0 || !eol)
@@ -298,11 +357,17 @@ static int regcmp_lines(char *val, const char *regex)
 static char* pop_next_command(GList **pp_rule_list,
         char **pp_event_name,    /* reports EVENT value thru this, if not NULL on entry */
         struct dump_dir **pp_dd, /* use *pp_dd for access to dump dir, if non-NULL */
+        problem_data_t *pd,      /* use *pd for access to problem data, if non-NULL */
         const char *dump_dir_name,
         const char *pfx,
         unsigned pfx_len
 )
 {
+    /* It is an error to pass both, but we can recover from it and use only
+     * problem_data_t in that case */
+    if (pp_dd != NULL && pd != NULL)
+        error_msg("BUG: both dump dir and problem data passed to %s()", __func__);
+
     char *command = NULL;
     struct dump_dir *dd = pp_dd ? *pp_dd : NULL;
 
@@ -324,14 +389,14 @@ static char* pop_next_command(GList **pp_rule_list,
                     goto next_rule; /* prefix doesn't match */
                 if (pp_event_name)
                 {
-                    free(*pp_event_name);
-                    *pp_event_name = xstrdup(eq_sign + 1);
+                    g_free(*pp_event_name);
+                    *pp_event_name = g_strdup(eq_sign + 1);
                 }
             }
             else
             {
                 /* Read from dump dir and compare */
-                if (!dd)
+                if (!dd && pd == NULL)
                 {
                     /* Without dir to match, we assume match for all conditions */
                     if (!dump_dir_name)
@@ -348,21 +413,24 @@ static char* pop_next_command(GList **pp_rule_list,
                 int regex = (eq_sign > cond_str && eq_sign[-1] == '~');
                 /* Is it "VAR!=VAL"? */
                 int inverted = (eq_sign > cond_str && eq_sign[-1] == '!');
-                char *var_name = xstrndup(cond_str, eq_sign - cond_str - (regex|inverted));
-                char *real_val = dd_load_text_ext(dd, var_name, DD_FAIL_QUIETLY_ENOENT);
-                free(var_name);
+                g_autofree char *var_name = g_strndup(cond_str, eq_sign - cond_str - (regex|inverted));
+                char *real_val = NULL;
+                g_autofree char *free_me = NULL;
+                if (pd == NULL)
+                    free_me = real_val = dd_load_text_ext(dd, var_name, DD_FAIL_QUIETLY_ENOENT);
+                else
+                {
+                    real_val = problem_data_get_content_or_NULL(pd, var_name);
+                    if (real_val == NULL)
+                        free_me = real_val = g_strdup("");
+                }
                 int vals_differ = regex ? regcmp_lines(real_val, eq_sign + 1) : strcmp(real_val, eq_sign + 1);
-                free(real_val);
                 if (inverted)
                     vals_differ = !vals_differ;
 
                 /* Do values match? */
                 if (vals_differ) /* no */
                 {
-                    //log_debug("var '%s': '%.*s'!='%s', skipping line",
-                    //        p,
-                    //        (int)(strchrnul(real_val, '\n') - real_val), real_val,
-                    //        eq_sign);
                     goto next_rule;
                 }
             }
@@ -374,10 +442,9 @@ static char* pop_next_command(GList **pp_rule_list,
         /* We are here if all conditions are satisfied */
         /* IOW, we found rule to run, delete it and return its command */
         *pp_rule_list = g_list_remove(*pp_rule_list, cur_rule);
-        list_free_with_free(cur_rule->conditions);
+        g_list_free_full(cur_rule->conditions, free);
         command = cur_rule->command;
-        /*free(cur_rule->command); - WRONG! we are returning it! */
-        free(cur_rule);
+        g_free(cur_rule);
         break;
 
  next_rule:
@@ -400,14 +467,12 @@ void free_commands(struct run_event_state *state)
     state->command_pid = 0;
 }
 
-int prepare_commands(struct run_event_state *state,
-                const char *dump_dir_name,
-                const char *event
-) {
+int prepare_commands(struct run_event_state *state)
+{
     free_commands(state);
 
     state->children_count = 0;
-    strbuf_clear(state->command_output);
+    g_string_erase(state->command_output, 0, -1);
 
     GList *rule_list = load_rule_list(NULL, CONF_DIR"/report_event.conf", /*recursion_depth:*/ 0);
     state->rule_list = rule_list;
@@ -419,9 +484,10 @@ int spawn_next_command(struct run_event_state *state,
                 const char *event,
                 unsigned execflags
 ) {
-    char *cmd = pop_next_command(&state->rule_list,
+    g_autofree char *cmd = pop_next_command(&state->rule_list,
                 NULL,          /* don't return event_name */
                 NULL,          /* NULL dd: we match by... */
+                NULL,          /* no problem data */
                 dump_dir_name, /* ...dirname */
                 event, strlen(event)+1 /* for this event name exactly (not prefix) */
     );
@@ -434,20 +500,30 @@ int spawn_next_command(struct run_event_state *state,
      */
     state->children_count++;
 
-    log_notice("Executing '%s'", cmd);
+    log_info("Next command: '%s'", cmd);
 
-    /* Export some useful environment variables for children */
-    char *env_vec[4];
     /* Just exporting dump_dir_name isn't always ok: it can be "."
      * and some children want to cd to other directory but still
      * be able to find problem directory by using $DUMP_DIR...
      */
-    char *full_name = realpath(dump_dir_name, NULL);
-    env_vec[0] = xasprintf("DUMP_DIR=%s", (full_name ? full_name : dump_dir_name));
-    free(full_name);
-    env_vec[1] = xasprintf("EVENT=%s", event);
-    env_vec[2] = xasprintf("REPORT_CLIENT_SLAVE=1");
-    env_vec[3] = NULL;
+    g_autofree char *full_name = realpath(dump_dir_name, NULL);
+    /* Export some useful environment variables for children */
+    GPtrArray *env_array;
+
+    env_array = g_ptr_array_new();
+
+    g_ptr_array_add(env_array, g_strdup_printf("DUMP_DIR=%s", (full_name ? full_name : dump_dir_name)));
+    g_ptr_array_add(env_array, g_strdup_printf("EVENT=%s", event));
+    g_ptr_array_add(env_array, g_strdup_printf("REPORT_CLIENT_SLAVE=1"));
+    for (unsigned int i = 0; i < state->extra_environment->len; i++)
+    {
+        char *variable;
+
+        variable = g_ptr_array_index(state->extra_environment, i);
+
+        g_ptr_array_add(env_array, g_strdup(variable));
+    }
+    g_ptr_array_add(env_array, NULL);
 
     char *argv[4];
     argv[0] = (char*)"/bin/sh"; // TODO: honor $SHELL?
@@ -456,21 +532,18 @@ int spawn_next_command(struct run_event_state *state,
     argv[3] = NULL;
 
     int pipefds[2];
-    state->command_pid = fork_execv_on_steroids(
+    state->command_pid = libreport_fork_execv_on_steroids(
                 EXECFLG_INPUT | EXECFLG_OUTPUT | EXECFLG_ERR2OUT | execflags,
                 argv,
                 pipefds,
-                /* env_vec: */ env_vec,
+                /* env_vec: */ (char **)env_array->pdata,
                 /* dir: */ dump_dir_name,
                 /* uid(unused): */ 0
     );
     state->command_out_fd = pipefds[0];
     state->command_in_fd = pipefds[1];
 
-    free(env_vec[0]);
-    free(env_vec[1]);
-    free(env_vec[2]);
-    free(cmd);
+    g_ptr_array_free(env_array, TRUE);
 
     return 0;
 }
@@ -480,8 +553,8 @@ int consume_event_command_output(struct run_event_state *state, const char *dump
     int r = 0;
     char buf[256];
     errno = 0;
-    struct strbuf *cmd_output = state->command_output;
-    while ((r = safe_read(state->command_out_fd, buf, sizeof(buf) - 1)) > 0)
+    GString *cmd_output = state->command_output;
+    while ((r = libreport_safe_read(state->command_out_fd, buf, sizeof(buf) - 1)) > 0)
     {
         char *newline;
         char *raw;
@@ -491,18 +564,18 @@ int consume_event_command_output(struct run_event_state *state, const char *dump
         while ((newline = strchr(raw, '\n')) != NULL)
         {
             *newline = '\0';
-            strbuf_append_str(cmd_output, raw);
-            char *msg = cmd_output->buf;
+            g_string_append(cmd_output, raw);
+            char *msg = cmd_output->str;
 
-            char *response = NULL;
+            g_autofree char *response = NULL;
 
             /* just cut off prefix, no waiting */
-            if (prefixcmp(msg, REPORT_PREFIX_ALERT) == 0)
+            if (g_str_has_prefix(msg, REPORT_PREFIX_ALERT))
             {
                 state->alert_callback(msg + sizeof(REPORT_PREFIX_ALERT) - 1 , state->interaction_param);
             }
             /* wait for y/N/f response on the same line */
-            else if (prefixcmp(msg, REPORT_PREFIX_ASK_YES_NO_YESFOREVER) == 0)
+            else if (g_str_has_prefix(msg, REPORT_PREFIX_ASK_YES_NO_YESFOREVER))
             {
                 /* example:
                  *   ASK_YES_NO_YESFOREVER ask_before_delete Do you want to delete selected files?
@@ -530,21 +603,52 @@ int consume_event_command_output(struct run_event_state *state, const char *dump
                     key_end[0] = ' '; /* restore original message, not sure if it is necessary */
                 }
 
-                response = xstrdup(ans ? "y" : "N");
+                response = g_strdup(ans ? "y" : "N");
+            }
+            /* wait for y/N/f/e response on the same line */
+            else if (g_str_has_prefix(msg, REPORT_PREFIX_ASK_YES_NO_SAVE_RESULT))
+            {
+                /* example:
+                 *   ASK_YES_NO_SAVE_RESULT ask_before_delete Do you want to delete selected files?
+                 */
+                char *key = msg + sizeof(REPORT_PREFIX_ASK_YES_NO_SAVE_RESULT) - 1;
+                char *key_end = strchr(key, ' ');
+
+                bool ans = false;
+
+                if (!key_end)
+                {   /* example:
+                     *  ASK_YES_NO_YESFOREVER Continue?
+                     *
+                     * Print a wraning only and do not scary users with error messages.
+                     */
+                    log_warning("invalid input format (missing option name), using simple ask yes/no");
+
+                    /* can't simply use 'goto ask_yes_no' because of different lenght of prefixes */
+                    ans = state->ask_yes_no_callback(key, state->interaction_param);
+                }
+                else
+                {
+                    key_end[0] = '\0'; /* split 'key msg' to 'key' and 'msg' */
+                    ans = state->ask_yes_no_save_result_callback(key, key + strlen(key) + 1, state->interaction_param);
+                    key_end[0] = ' '; /* restore original message, not sure if it is necessary */
+                }
+
+                response = g_strdup(ans ? "y" : "N");
             }
             /* wait for y/N response on the same line */
-            else if (prefixcmp(msg, REPORT_PREFIX_ASK_YES_NO) == 0)
+            else if (g_str_has_prefix(msg, REPORT_PREFIX_ASK_YES_NO))
             {
                 const bool ans = state->ask_yes_no_callback(msg + sizeof(REPORT_PREFIX_ASK_YES_NO) - 1, state->interaction_param);
-                response = xstrdup(ans ? "y" : "N");
+                response = g_strdup(ans ? "y" : "N");
             }
             /* wait for the string on the same line */
-            else if (prefixcmp(msg, REPORT_PREFIX_ASK) == 0)
+            else if (g_str_has_prefix(msg, REPORT_PREFIX_ASK))
             {
                 response = state->ask_callback(msg + sizeof(REPORT_PREFIX_ASK) - 1, state->interaction_param);
             }
             /* set echo off and wait for password on the same line */
-            else if (prefixcmp(msg, REPORT_PREFIX_ASK_PASSWORD) == 0)
+            else if (g_str_has_prefix(msg, REPORT_PREFIX_ASK_PASSWORD))
             {
                 response = state->ask_password_callback(msg + sizeof(REPORT_PREFIX_ASK_PASSWORD) - 1, state->interaction_param);
             }
@@ -552,8 +656,7 @@ int consume_event_command_output(struct run_event_state *state, const char *dump
              * note that callback may take ownership of buf by returning NULL */
             else if (state->logging_callback)
             {
-                char *logged = state->logging_callback(xstrdup(msg), state->logging_param);
-                free(logged);
+                g_autofree char *logged = state->logging_callback(g_strdup(msg), state->logging_param);
             }
 
             if (response)
@@ -561,35 +664,33 @@ int consume_event_command_output(struct run_event_state *state, const char *dump
                 size_t len = strlen(response);
                 response[len++] = '\n';
 
-                if (full_write(state->command_in_fd, response, len) != len)
+                if (libreport_full_write(state->command_in_fd, response, len) != len)
                 {
                     if (state->error_callback)
                         state->error_callback("<WRITE ERROR>", state->error_param);
                     else
                         perror_msg_and_die("Can't write %zu bytes to child's stdin", len);
                 }
-
-                free(response);
             }
 
-            strbuf_clear(cmd_output);
+            g_string_erase(cmd_output, 0, -1);
 
             /* jump to next line */
             raw = newline + 1;
         }
 
         /* beginning of next line. the line continues by next read() */
-        strbuf_append_str(cmd_output, raw);
+        g_string_append(cmd_output, raw);
     }
 
     /* Hope that child's stdout fd was set to O_NONBLOCK */
     if (r == -1 && errno == EAGAIN)
         return -1;
 
-    strbuf_clear(cmd_output);
+    g_string_erase(cmd_output, 0, -1);
 
     /* Wait for child to actually exit, collect status */
-    safe_waitpid(state->command_pid, &(state->process_status), 0);
+    libreport_safe_waitpid(state->command_pid, &(state->process_status), 0);
 
     int retval = WEXITSTATUS(state->process_status);
     if (WIFSIGNALED(state->process_status))
@@ -607,7 +708,7 @@ int run_event_on_dir_name(struct run_event_state *state,
                 const char *dump_dir_name,
                 const char *event
 ) {
-    prepare_commands(state, dump_dir_name, event);
+    prepare_commands(state);
 
     /* Execute every command in shell */
 
@@ -631,14 +732,13 @@ int run_event_on_problem_data(struct run_event_state *state, problem_data_t *dat
     struct dump_dir *dd = create_dump_dir_from_problem_data(data, NULL);
     if (!dd)
         return -1;
-    char *dir_name = xstrdup(dd->dd_dirname);
+    g_autofree char *dir_name = g_strdup(dd->dd_dirname);
     dd_close(dd);
 
     int r = run_event_on_dir_name(state, dir_name, event);
 
     g_hash_table_remove_all(data);
     dd = dd_opendir(dir_name, /*flags:*/ 0);
-    free(dir_name);
     if (dd)
     {
         problem_data_load_from_dump_dir(data, dd, NULL);
@@ -648,9 +748,10 @@ int run_event_on_problem_data(struct run_event_state *state, problem_data_t *dat
     return r;
 }
 
-char *list_possible_events(struct dump_dir *dd, const char *dump_dir_name, const char *pfx)
+
+static char *_list_possible_events(struct dump_dir **dd, problem_data_t *pd, const char *dump_dir_name, const char *pfx)
 {
-    struct strbuf *result = strbuf_new();
+    GString *result = g_string_new(NULL);
 
     GList *rule_list = load_rule_list(NULL, CONF_DIR"/report_event.conf", /*recursion_depth:*/ 0);
 
@@ -658,26 +759,25 @@ char *list_possible_events(struct dump_dir *dd, const char *dump_dir_name, const
     for (;;)
     {
         /* Retrieve each cmd, and fetch its EVENT=foo value */
-        char *event_name = NULL;
-        char *cmd = pop_next_command(&rule_list,
+        g_autofree char *event_name = NULL;
+        g_autofree char *cmd = pop_next_command(&rule_list,
                 &event_name,       /* return event_name */
-                (dd ? &dd : NULL), /* match this dd... */
+                dd,                /* match this dd... */
+                pd,                /* no problem data */
                 dump_dir_name,     /* ...or if NULL, this dirname */
                 pfx, pfx_len       /* for events with this prefix */
         );
         if (!cmd)
         {
             free_rule_list(rule_list);
-            free(event_name);
             break;
         }
-        free(cmd);
 
         if (event_name)
         {
             /* Append "EVENT\n" - only if it is not there yet */
             unsigned e_len = strlen(event_name);
-            char *p = result->buf;
+            char *p = result->str;
             while (p && *p)
             {
                 if (strncmp(p, event_name, e_len) == 0 && p[e_len] == '\n')
@@ -686,66 +786,79 @@ char *list_possible_events(struct dump_dir *dd, const char *dump_dir_name, const
                 if (p)
                     p++;
             }
-            strbuf_append_strf(result, "%s\n", event_name);
+            g_string_append_printf(result, "%s\n", event_name);
  skip:
-            free(event_name);
+            continue;
         }
     }
 
-    return strbuf_free_nobuf(result);
+    return g_string_free(result, FALSE);
+}
+
+char *list_possible_events(struct dump_dir *dd, const char *dump_dir_name, const char *pfx)
+{
+    return _list_possible_events((dd ? &dd : NULL), NULL, dump_dir_name, pfx);
+}
+
+char *list_possible_events_problem_data(problem_data_t *pd, const char *dump_dir_name, const char *pfx)
+{
+    return _list_possible_events(NULL, pd, dump_dir_name, pfx);
 }
 
 GList *list_possible_events_glist(const char *problem_dir_name,
                                   const char *pfx)
 {
     struct dump_dir *dd = dd_opendir(problem_dir_name, DD_OPEN_READONLY);
-    GList *l = NULL;
-    char *events = list_possible_events(dd, problem_dir_name, pfx);
-    char *start = events;
-    char *end = strchr(events, '\n');
-
-    while(end)
-    {
-        *end = '\0';
-        l = g_list_append(l, xstrdup(start));
-        start = end + 1;
-        end = strchr(start, '\n');
-    }
-
+    g_autofree char *events = list_possible_events(dd, problem_dir_name, pfx);
+    GList *l = libreport_parse_delimited_list(events, "\n");
     dd_close(dd);
-    free(events);
+
+    return l;
+}
+
+GList *list_possible_events_problem_data_glist(problem_data_t *pd,
+                                  const char *problem_dir_name,
+                                  const char *pfx)
+{
+    g_autofree char *events = list_possible_events_problem_data(pd, problem_dir_name, pfx);
+    GList *l = libreport_parse_delimited_list(events, "\n");
 
     return l;
 }
 
 void run_event_stdio_alert(const char *msg, void *param)
 {
-    alert(msg);
+    libreport_alert(msg);
 }
 
 char *run_event_stdio_ask(const char *msg, void *param)
 {
-    return ask(msg);
+    return libreport_ask(msg);
 }
 
 int run_event_stdio_ask_yes_no(const char *msg, void *param)
 {
-    return ask_yes_no(msg);
+    return libreport_ask_yes_no(msg);
 }
 
 int run_event_stdio_ask_yes_no_yesforever(const char *key, const char *msg, void *param)
 {
-    return ask_yes_no_yesforever(key, msg);
+    return libreport_ask_yes_no_yesforever(key, msg);
+}
+
+int run_event_stdio_ask_yes_no_save_result(const char *key, const char *msg, void *param)
+{
+    return libreport_ask_yes_no_save_result(key, msg);
 }
 
 char *run_event_stdio_ask_password(const char *msg, void *param)
 {
-    return ask_password(msg);
+    return libreport_ask_password(msg);
 }
 
 static char *run_event_stdio_log(char *log_line, void *param)
 {
-    client_log(log_line);
+    libreport_client_log(log_line);
     return log_line;
 }
 
@@ -754,16 +867,16 @@ static void run_event_stdio_error_and_die(const char *error_line, void *param)
     error_msg_and_die("Can't write bytes to child's stdin");
 }
 
-char *exit_status_as_string(const char *progname, int status)
+char *libreport_exit_status_as_string(const char *progname, int status)
 {
     INITIALIZE_LIBREPORT();
 
     char *msg;
     if (WIFSIGNALED(status))
-        msg = xasprintf(_("('%s' was killed by signal %u)\n"), progname, WTERMSIG(status));
+        msg = g_strdup_printf(_("('%s' was killed by signal %u)\n"), progname, WTERMSIG(status));
     else if (status == 0)
-        msg = xasprintf(_("('%s' completed successfully)\n"), progname);
+        msg = g_strdup_printf(_("('%s' completed successfully)\n"), progname);
     else
-        msg = xasprintf(_("('%s' exited with %u)\n"), progname, WEXITSTATUS(status));
+        msg = g_strdup_printf(_("('%s' exited with %u)\n"), progname, WEXITSTATUS(status));
     return msg;
 }
