@@ -30,23 +30,13 @@
 
 #define BTHASH_URL_SFX "reports/bthash/"
 
-#define RHSM_WEB_SERVICE_URL "https://api.access.redhat.com/rs/telemetry/abrt"
-
-#define RHSMENT_PEM_DIR_PATH "/etc/pki/entitlement"
-#define RHSMENT_ENT_DATA_BEGIN_TAG "-----BEGIN ENTITLEMENT DATA-----"
-#define RHSMENT_ENT_DATA_END_TAG "-----END ENTITLEMENT DATA-----"
-#define RHSMENT_SIG_DATA_BEGIN_TAG "-----BEGIN RSA SIGNATURE-----"
-#define RHSMENT_SIG_DATA_END_TAG "-----END RSA SIGNATURE-----"
-
-
 static char *
 puppet_config_print(const char *key)
 {
-    char *command = xasprintf("puppet config print %s", key);
-    char *result = run_in_shell_and_save_output(0, command, NULL, NULL);
-    free(command);
+    g_autofree char *command = g_strdup_printf("puppet config print %s", key);
+    char *result = libreport_run_in_shell_and_save_output(0, command, NULL, NULL);
 
-    /* run_in_shell_and_save_output always returns non-NULL */
+    /* libreport_run_in_shell_and_save_output always returns non-NULL */
     if (result[0] != '/')
         goto error;
 
@@ -57,47 +47,20 @@ puppet_config_print(const char *key)
     *newline = '\0';
     return result;
 error:
-    free(result);
+    g_free(result);
     error_msg_and_die("Unable to determine puppet %s path (puppet not installed?)", key);
 }
 
 void
-ureport_server_config_set_url(struct ureport_server_config *config,
+libreport_ureport_server_config_set_url(struct ureport_server_config *config,
                               char *server_url)
 {
-    free(config->ur_url);
+    g_free(config->ur_url);
     config->ur_url = server_url;
 }
 
-static char *
-rhsm_config_get_entitlement_cert_dir(void)
-{
-    char *result = getenv("LIBREPORT_DEBUG_RHSMENT_PEM_DIR_PATH");
-    if (result != NULL)
-        return xstrdup(result);
-
-    result = run_in_shell_and_save_output(0,
-            "python -c \"from rhsm.config import initConfig; print(initConfig().get('rhsm', 'entitlementCertDir'))\"",
-            NULL, NULL);
-
-    /* run_in_shell_and_save_output always returns non-NULL */
-    if (result[0] != '/')
-        goto error;
-
-    char *newline = strchrnul(result, '\n');
-    if (!newline)
-        goto error;
-
-    *newline = '\0';
-    return result;
-error:
-    free(result);
-    error_msg("Failed to get 'rhsm':'entitlementCertDir' from rhsm.config python module. Using "RHSMENT_PEM_DIR_PATH);
-    return xstrdup(RHSMENT_PEM_DIR_PATH);
-}
-
 void
-ureport_server_config_set_client_auth(struct ureport_server_config *config,
+libreport_ureport_server_config_set_client_auth(struct ureport_server_config *config,
                                       const char *client_auth)
 {
     if (client_auth == NULL)
@@ -105,107 +68,14 @@ ureport_server_config_set_client_auth(struct ureport_server_config *config,
 
     if (strcmp(client_auth, "") == 0)
     {
-        free(config->ur_client_cert);
+        g_free(config->ur_client_cert);
         config->ur_client_cert = NULL;
 
-        free(config->ur_client_key);
+        g_free(config->ur_client_key);
         config->ur_client_key = NULL;
 
         log_notice("Not using client authentication");
     }
-    else if (strcmp(client_auth, "rhsm") == 0)
-    {
-        if (config->ur_url == NULL)
-            ureport_server_config_set_url(config, xstrdup(RHSM_WEB_SERVICE_URL));
-
-        char *rhsm_dir = rhsm_config_get_entitlement_cert_dir();
-        if (rhsm_dir == NULL)
-        {
-            log_notice("Not using client authentication");
-            return;
-        }
-
-        GList *certs = get_file_list(rhsm_dir, "pem");
-        if (g_list_length(certs) < 2)
-        {
-            g_list_free_full(certs, (GDestroyNotify)free_file_obj);
-
-            log_notice("'%s' does not contain a cert-key files pair", rhsm_dir);
-            log_notice("Not using client authentication");
-            free(rhsm_dir);
-            return;
-        }
-
-        /* Use the last non-key file found. */
-        file_obj_t *cert = NULL;
-        for (GList *iter = certs; iter != NULL; iter = g_list_next(iter))
-        {
-            file_obj_t *tmp = (file_obj_t *)iter->data;
-            const char *file_name = fo_get_filename(tmp);
-
-            if (suffixcmp(file_name, "-key"))
-                cert = tmp;
-        }
-
-        if (cert == NULL)
-        {
-            g_list_free_full(certs, (GDestroyNotify)free_file_obj);
-
-            log_notice("'%s' does not contain a cert file (only keys)", rhsm_dir);
-            log_notice("Not using client authentication");
-            free(rhsm_dir);
-            return;
-        }
-
-        config->ur_client_cert = xstrdup(fo_get_fullpath(cert));
-        /* Yes, the key file may not exists. I over took this code from
-         * sos-uploader and they are pretty happy with this approach, so why
-         * shouldn't we?. */
-        config->ur_client_key = xasprintf("%s/%s-key.pem", rhsm_dir, fo_get_filename(cert));
-        free(rhsm_dir);
-
-        log_debug("Using cert files: '%s' : '%s'", config->ur_client_cert, config->ur_client_key);
-
-        g_list_free_full(certs, (GDestroyNotify)free_file_obj);
-
-        char *certdata = xmalloc_open_read_close(config->ur_client_cert, /*no size limit*/NULL);
-        if (certdata != NULL)
-        {
-            char *ent_data = xstrdup_between(certdata,
-                    RHSMENT_ENT_DATA_BEGIN_TAG, RHSMENT_ENT_DATA_END_TAG);
-
-            char *sig_data = xstrdup_between(certdata,
-                    RHSMENT_SIG_DATA_BEGIN_TAG, RHSMENT_SIG_DATA_END_TAG);
-
-            if (ent_data != NULL && sig_data != NULL)
-            {
-                ent_data = strremovech(ent_data, '\n');
-                insert_map_string(config->ur_http_headers,
-                        xstrdup("X-RH-Entitlement-Data"),
-                        xasprintf(RHSMENT_ENT_DATA_BEGIN_TAG"%s"RHSMENT_ENT_DATA_END_TAG, ent_data));
-
-                sig_data = strremovech(sig_data, '\n');
-                insert_map_string(config->ur_http_headers,
-                        xstrdup("X-RH-Entitlement-Sig"),
-                        xasprintf(RHSMENT_SIG_DATA_BEGIN_TAG"%s"RHSMENT_SIG_DATA_END_TAG, sig_data));
-            }
-            else
-            {
-                log_notice("Cert file '%s' doesn't contain Entitlement and RSA Signature sections", config->ur_client_cert);
-                log_notice("Not using client authentication");
-
-                free(config->ur_client_cert);
-                config->ur_client_cert = NULL;
-
-                free(config->ur_client_key);
-                config->ur_client_key = NULL;
-            }
-
-            free(sig_data);
-            free(ent_data);
-            free(certdata);
-        }
-    }
     else if (strcmp(client_auth, "puppet") == 0)
     {
         config->ur_client_cert = puppet_config_print("hostcert");
@@ -213,10 +83,9 @@ ureport_server_config_set_client_auth(struct ureport_server_config *config,
     }
     else
     {
-        char *scratch = xstrdup(client_auth);
-        config->ur_client_cert = xstrdup(strtok(scratch, ":"));
-        config->ur_client_key = xstrdup(strtok(NULL, ":"));
-        free(scratch);
+        g_autofree char *scratch = g_strdup(client_auth);
+        config->ur_client_cert = g_strdup(strtok(scratch, ":"));
+        config->ur_client_key = g_strdup(strtok(NULL, ":"));
 
         if (config->ur_client_cert == NULL || config->ur_client_key == NULL)
             error_msg_and_die("Invalid client authentication specification");
@@ -227,25 +96,25 @@ ureport_server_config_set_client_auth(struct ureport_server_config *config,
         log_notice("Using client certificate: %s", config->ur_client_cert);
         log_notice("Using client private key: %s", config->ur_client_key);
 
-        free(config->ur_username);
+        g_free(config->ur_username);
         config->ur_username = NULL;
 
-        free(config->ur_password);
+        g_free(config->ur_password);
         config->ur_password = NULL;
     }
 }
 
 void
-ureport_server_config_set_basic_auth(struct ureport_server_config *config,
+libreport_ureport_server_config_set_basic_auth(struct ureport_server_config *config,
                                      const char *login, const char *password)
 {
-    ureport_server_config_set_client_auth(config, "");
+    libreport_ureport_server_config_set_client_auth(config, "");
 
-    free(config->ur_username);
-    config->ur_username = xstrdup(login);
+    g_free(config->ur_username);
+    config->ur_username = g_strdup(login);
 
-    free(config->ur_password);
-    config->ur_password = xstrdup(password);
+    g_free(config->ur_password);
+    config->ur_password = g_strdup(password);
 }
 
 void
@@ -255,59 +124,40 @@ ureport_server_config_load_basic_auth(struct ureport_server_config *config,
     if (http_auth_pref == NULL)
         return;
 
-    map_string_t *settings = NULL;
+    g_autoptr(GHashTable) settings = NULL;
 
-    char *tmp_password = NULL;
-    char *tmp_username = NULL;
-    const char *username = NULL;
-    const char *password = NULL;
+    g_autofree const char *username = NULL;
+    const char *password_from_http_auth_pref = NULL;
 
-    if (strcmp(http_auth_pref, "rhts-credentials") == 0)
-    {
-        settings = new_map_string();
+    username = g_strdup(http_auth_pref);
+    password_from_http_auth_pref = strchr(username, ':');
 
-        if (!load_plugin_conf_file("rhtsupport.conf", settings, /*skip key w/o values:*/ false))
-            error_msg_and_die("Could not get RHTSupport credentials");
-
-        username = get_map_string_item_or_NULL(settings, "Login");
-        password = get_map_string_item_or_NULL(settings, "Password");
-
-        if (config->ur_url == NULL)
-            ureport_server_config_set_url(config, xstrdup(RHSM_WEB_SERVICE_URL));
-    }
-    else
-    {
-        username = tmp_username = xstrdup(http_auth_pref);
-        password = strchr(tmp_username, ':');
-
-        if (password != NULL)
-            /* It is "char *", see strchr() few lines above. */
-            *((char *)(password++)) = '\0';
+    if (password_from_http_auth_pref != NULL) {
+        /* It is "char *", see strchr() few lines above. */
+        *((char *)(password_from_http_auth_pref++)) = '\0';
     }
 
-    if (password == NULL)
-    {
-        char *message = xasprintf("Please provide uReport server password for user '%s':", username);
-        password = tmp_password = ask_password(message);
-        free(message);
+    if (password_from_http_auth_pref == NULL) {
+        g_autofree char *message = g_strdup_printf("Please provide uReport server password for user '%s':", username);
+        g_autofree char *password_from_input = libreport_ask_password(message);
 
-        if (strcmp(password, "") == 0)
+        if (strcmp(password_from_input, "") == 0) {
             error_msg_and_die("Cannot continue without uReport server password!");
+        }
+
+        libreport_ureport_server_config_set_basic_auth(config, username, password_from_input);
+    } else {
+        libreport_ureport_server_config_set_basic_auth(config, username, password_from_http_auth_pref);
     }
 
-    ureport_server_config_set_basic_auth(config, username, password);
-
-    free(tmp_password);
-    free(tmp_username);
-    free_map_string(settings);
 }
 
 void
-ureport_server_config_load(struct ureport_server_config *config,
-                           map_string_t *settings)
+libreport_ureport_server_config_load(struct ureport_server_config *config,
+                           GHashTable *settings)
 {
-    UREPORT_OPTION_VALUE_FROM_CONF(settings, "URL", config->ur_url, xstrdup);
-    UREPORT_OPTION_VALUE_FROM_CONF(settings, "SSLVerify", config->ur_ssl_verify, string_to_bool);
+    UREPORT_OPTION_VALUE_FROM_CONF(settings, "URL", config->ur_url, g_strdup);
+    UREPORT_OPTION_VALUE_FROM_CONF(settings, "SSLVerify", config->ur_ssl_verify, libreport_string_to_bool);
 
     const char *http_auth_pref = NULL;
     UREPORT_OPTION_VALUE_FROM_CONF(settings, "HTTPAuth", http_auth_pref, (const char *));
@@ -317,18 +167,18 @@ ureport_server_config_load(struct ureport_server_config *config,
     if (http_auth_pref == NULL)
     {
         UREPORT_OPTION_VALUE_FROM_CONF(settings, "SSLClientAuth", client_auth, (const char *));
-        ureport_server_config_set_client_auth(config, client_auth);
+        libreport_ureport_server_config_set_client_auth(config, client_auth);
     }
 
     /* If SSLClientAuth is configured, include the auth items by default. */
     bool include_auth = config->ur_client_cert != NULL || config->ur_username != NULL;
-    UREPORT_OPTION_VALUE_FROM_CONF(settings, "IncludeAuthData", include_auth, string_to_bool);
+    UREPORT_OPTION_VALUE_FROM_CONF(settings, "IncludeAuthData", include_auth, libreport_string_to_bool);
 
     if (include_auth)
     {
         const char *auth_items = NULL;
         UREPORT_OPTION_VALUE_FROM_CONF(settings, "AuthDataItems", auth_items, (const char *));
-        config->ur_prefs.urp_auth_items = parse_list(auth_items);
+        config->ur_prefs.urp_auth_items = libreport_parse_delimited_list(auth_items, ",");
 
         if (config->ur_prefs.urp_auth_items == NULL)
             log_warning("IncludeAuthData set to 'yes' but AuthDataItems is empty.");
@@ -336,65 +186,67 @@ ureport_server_config_load(struct ureport_server_config *config,
 }
 
 void
-ureport_server_config_init(struct ureport_server_config *config)
+libreport_ureport_server_config_init(struct ureport_server_config *config)
 {
     config->ur_url = NULL;
     config->ur_ssl_verify = true;
     config->ur_client_cert = NULL;
     config->ur_client_key = NULL;
+    config->ur_cert_authority_cert = NULL;
     config->ur_username = NULL;
     config->ur_password = NULL;
-    config->ur_http_headers = new_map_string();
+
     config->ur_prefs.urp_auth_items = NULL;
+    config->ur_prefs.urp_flags = 0;
 }
 
 void
-ureport_server_config_destroy(struct ureport_server_config *config)
+libreport_ureport_server_config_destroy(struct ureport_server_config *config)
 {
-    free(config->ur_url);
+    g_free(config->ur_url);
     config->ur_url = DESTROYED_POINTER;
 
-    free(config->ur_client_cert);
+    g_free(config->ur_client_cert);
     config->ur_client_cert = DESTROYED_POINTER;
 
-    free(config->ur_client_key);
+    g_free(config->ur_client_key);
     config->ur_client_key = DESTROYED_POINTER;
 
-    free(config->ur_username);
+    g_free(config->ur_cert_authority_cert);
+    config->ur_cert_authority_cert = DESTROYED_POINTER;
+
+    g_free(config->ur_username);
     config->ur_username = DESTROYED_POINTER;
 
-    free(config->ur_password);
+    g_free(config->ur_password);
     config->ur_password = DESTROYED_POINTER;
 
-    g_list_free_full(config->ur_prefs.urp_auth_items, free);
+    g_list_free_full(config->ur_prefs.urp_auth_items, g_free);
     config->ur_prefs.urp_auth_items = DESTROYED_POINTER;
-
-    free_map_string(config->ur_http_headers);
-    config->ur_http_headers = DESTROYED_POINTER;
 }
 
 void
-ureport_server_response_free(struct ureport_server_response *resp)
+libreport_ureport_server_response_free(struct ureport_server_response *resp)
 {
     if (!resp)
         return;
 
-    free(resp->urr_solution);
+    g_free(resp->urr_solution);
     resp->urr_solution = DESTROYED_POINTER;
 
     g_list_free_full(resp->urr_reported_to_list, g_free);
     resp->urr_reported_to_list = DESTROYED_POINTER;
 
-    free(resp->urr_bthash);
+    g_free(resp->urr_bthash);
     resp->urr_bthash = DESTROYED_POINTER;
 
-    free(resp->urr_message);
+    g_free(resp->urr_message);
     resp->urr_message = DESTROYED_POINTER;
 
-    free(resp->urr_value);
+    g_free(resp->urr_value);
     resp->urr_value = DESTROYED_POINTER;
 
-    free(resp);
+    g_free(resp);
 }
 
 static char *
@@ -403,14 +255,14 @@ parse_solution_from_json_list(struct json_object *list,
 {
     json_object *list_elem, *struct_elem;
     const char *cause, *note, *url;
-    struct strbuf *solution_buf = strbuf_new();
+    GString *solution_buf = g_string_new(NULL);
 
     const unsigned length = json_object_array_length(list);
 
     const char *one_format = _("Your problem seems to be caused by %s\n\n%s\n");
     if (length > 1)
     {
-        strbuf_append_str(solution_buf, _("Your problem seems to be caused by one of the following:\n"));
+        g_string_append(solution_buf, _("Your problem seems to be caused by one of the following:\n"));
         one_format = "\n* %s\n\n%s\n";
     }
 
@@ -436,7 +288,7 @@ parse_solution_from_json_list(struct json_object *list,
             continue;
 
         empty = false;
-        strbuf_append_strf(solution_buf, one_format, cause, note);
+        g_string_append_printf(solution_buf, one_format, cause, note);
 
         if (!json_object_object_get_ex(list_elem, "url", &struct_elem))
             continue;
@@ -444,18 +296,18 @@ parse_solution_from_json_list(struct json_object *list,
         url = json_object_get_string(struct_elem);
         if (url)
         {
-            char *reported_to_line = xasprintf("%s: URL=%s", cause, url);
+            char *reported_to_line = g_strdup_printf("%s: URL=%s", cause, url);
             *reported_to = g_list_append(*reported_to, reported_to_line);
         }
     }
 
     if (empty)
     {
-        strbuf_free(solution_buf);
+        g_string_free(solution_buf, TRUE);
         return NULL;
     }
 
-    return strbuf_free_nobuf(solution_buf);
+    return g_string_free(solution_buf, FALSE);
 }
 
 /* reported_to json element should be a list of structures
@@ -471,7 +323,8 @@ parse_reported_to_from_json_list(struct json_object *list)
     int i;
     json_object *list_elem, *struct_elem;
     const char *reporter, *value, *type;
-    char *reported_to_line, *prefix;
+    char *reported_to_line = NULL;
+    g_autofree char *prefix = NULL;
     GList *result = NULL;
 
     for (i = 0; i < json_object_array_length(list); ++i)
@@ -502,16 +355,15 @@ parse_reported_to_from_json_list(struct json_object *list)
         if (type)
         {
             if (strcasecmp("url", type) == 0)
-                prefix = xstrdup("URL=");
+                prefix = g_strdup("URL=");
             else if (strcasecmp("bthash", type) == 0)
-                prefix = xstrdup("BTHASH=");
+                prefix = g_strdup("BTHASH=");
         }
 
         if (!prefix)
-            prefix = xstrdup("");
+            prefix = g_strdup("");
 
-        reported_to_line = xasprintf("%s: %s%s", reporter, prefix, value);
-        free(prefix);
+        reported_to_line = g_strdup_printf("%s: %s%s", reporter, prefix, value);
 
         result = g_list_append(result, reported_to_line);
     }
@@ -531,28 +383,28 @@ ureport_server_parse_json(json_object *json)
     json_object *obj = NULL;
     if (json_object_object_get_ex(json, "error", &obj))
     {
-        struct ureport_server_response *out_response = xzalloc(sizeof(*out_response));
+        struct ureport_server_response *out_response = g_new0(struct ureport_server_response, 1);
         out_response->urr_is_error = true;
         /*
          * Used to use json_object_to_json_string(obj), but it returns
          * the string in quote marks (") - IOW, json-formatted string.
          */
-        out_response->urr_value = xstrdup(json_object_get_string(obj));
+        out_response->urr_value = g_strdup(json_object_get_string(obj));
         return out_response;
     }
 
     if (json_object_object_get_ex(json, "result", &obj))
     {
-        struct ureport_server_response *out_response = xzalloc(sizeof(*out_response));
-        out_response->urr_value = xstrdup(json_object_get_string(obj));
+        struct ureport_server_response *out_response = g_new0(struct ureport_server_response, 1);
+        out_response->urr_value = g_strdup(json_object_get_string(obj));
 
         json_object *message = NULL;
         if (json_object_object_get_ex(json, "message", &message))
-            out_response->urr_message = xstrdup(json_object_get_string(message));
+            out_response->urr_message = g_strdup(json_object_get_string(message));
 
         json_object *bthash = NULL;
         if (json_object_object_get_ex(json, "bthash", &bthash))
-            out_response->urr_bthash = xstrdup(json_object_get_string(bthash));
+            out_response->urr_bthash = g_strdup(json_object_get_string(bthash));
 
         json_object *reported_to_list = NULL;
         if (json_object_object_get_ex(json, "reported_to", &reported_to_list))
@@ -571,7 +423,7 @@ ureport_server_parse_json(json_object *json)
 }
 
 struct ureport_server_response *
-ureport_server_response_from_reply(post_state_t *post_state,
+libreport_ureport_server_response_from_reply(post_state_t *post_state,
                                    struct ureport_server_config *config)
 {
     /* Previously, the condition here was (post_state->errmsg[0] != '\0')
@@ -581,7 +433,16 @@ ureport_server_response_from_reply(post_state_t *post_state,
      */
     if (post_state->curl_result != CURLE_OK)
     {
-        error_msg(_("Failed to upload uReport to the server '%s' with curl: %s"), config->ur_url, post_state->errmsg);
+        if (strcmp(post_state->errmsg, "") != 0)
+            error_msg(_("Failed to upload uReport to the server '%s' with curl: %s"),
+                                                                    config->ur_url,
+                                                                    post_state->errmsg);
+        else
+            error_msg(_("Failed to upload uReport to the server '%s'"), config->ur_url);
+
+        if (post_state->curl_error_msg != NULL && strcmp(post_state->curl_error_msg, "") != 0)
+            error_msg(_("Error: %s"), post_state->curl_error_msg);
+
         return NULL;
     }
 
@@ -615,7 +476,7 @@ ureport_server_response_from_reply(post_state_t *post_state,
 
     json_object *const json = json_tokener_parse(post_state->body);
 
-    if (is_error(json))
+    if (json == NULL)
     {
         error_msg(_("Unable to parse response from ureport server at '%s'"), config->ur_url);
         log_notice("%s", post_state->body);
@@ -640,7 +501,7 @@ ureport_server_response_from_reply(post_state_t *post_state,
 }
 
 bool
-ureport_server_response_save_in_dump_dir(struct ureport_server_response *resp,
+libreport_ureport_server_response_save_in_dump_dir(struct ureport_server_response *resp,
                                          const char *dump_dir_path,
                                          struct ureport_server_config *config)
 {
@@ -650,21 +511,53 @@ ureport_server_response_save_in_dump_dir(struct ureport_server_response *resp,
 
     if (resp->urr_bthash)
     {
-        char *msg = xasprintf("uReport: BTHASH=%s", resp->urr_bthash);
-        add_reported_to(dd, msg);
-        free(msg);
+        {
+            report_result_t *result;
 
-        char *report_url = ureport_server_response_get_report_url(resp, config);
-        msg = xasprintf("ABRT Server: URL=%s", report_url);
-        add_reported_to(dd, msg);
-        free(msg);
-        free(report_url);
+            result = report_result_new_with_label_from_env("uReport");
+
+            report_result_set_bthash(result, resp->urr_bthash);
+
+            libreport_add_reported_to_entry(dd, result);
+
+            report_result_free(result);
+        }
+
+        {
+            report_result_t *result;
+            g_autofree char *url = NULL;
+
+            result = report_result_new_with_label_from_env("ABRT Server");
+            url = libreport_ureport_server_response_get_report_url(resp, config);
+
+            report_result_set_url(result, url);
+
+            libreport_add_reported_to_entry(dd, result);
+
+            report_result_free(result);
+        }
     }
 
     if (resp->urr_reported_to_list)
     {
         for (GList *e = resp->urr_reported_to_list; e; e = g_list_next(e))
-            add_reported_to(dd, e->data);
+        {
+            char *workflow;
+
+            workflow = getenv("LIBREPORT_WORKFLOW");
+            if (NULL == workflow)
+            {
+                libreport_add_reported_to(dd, e->data);
+            }
+            else
+            {
+                g_autofree char *line = NULL;
+
+                line = g_strdup_printf("%s WORKFLOW=%s", (const char *)e->data, workflow);
+
+                libreport_add_reported_to(dd, line);
+            }
+        }
     }
 
     if (resp->urr_solution)
@@ -675,12 +568,11 @@ ureport_server_response_save_in_dump_dir(struct ureport_server_response *resp,
 }
 
 char *
-ureport_server_response_get_report_url(struct ureport_server_response *resp,
+libreport_ureport_server_response_get_report_url(struct ureport_server_response *resp,
                                        struct ureport_server_config *config)
 {
-    char *bthash_url = concat_path_file(config->ur_url, BTHASH_URL_SFX);
-    char *report_url = concat_path_file(bthash_url, resp->urr_bthash);
-    free(bthash_url);
+    g_autofree char *bthash_url = g_build_filename(config->ur_url ? config->ur_url : "", BTHASH_URL_SFX, NULL);
+    char *report_url = g_build_filename(bthash_url, resp->urr_bthash, NULL);
     return report_url;
 }
 
@@ -689,32 +581,38 @@ ureport_add_str(struct json_object *ur, const char *key, const char *s)
 {
     struct json_object *jstring = json_object_new_string(s);
     if (!jstring)
-        die_out_of_memory();
+        libreport_die_out_of_memory();
 
     json_object_object_add(ur, key, jstring);
 }
 
 char *
-ureport_from_dump_dir_ext(const char *dump_dir_path, const struct ureport_preferences *preferences)
+libreport_ureport_from_dump_dir_ext(const char *dump_dir_path, const struct ureport_preferences *preferences)
 {
-    char *error_message;
+    g_autofree char *error_message = NULL;
     struct sr_report *report = sr_abrt_report_from_dir(dump_dir_path,
                                                        &error_message);
 
     if (!report)
-        error_msg_and_die("%s", error_message);
+    {
+        if (NULL == preferences || !(preferences->urp_flags & UREPORT_PREF_FLAG_RETURN_ON_FAILURE))
+            error_msg_and_die("%s", error_message);
+
+        log_notice("%s", error_message);
+        return NULL;
+    }
 
     if (preferences != NULL && preferences->urp_auth_items != NULL)
     {
         struct dump_dir *dd = dd_opendir(dump_dir_path, DD_OPEN_READONLY);
         if (!dd)
-            xfunc_die(); /* dd_opendir() already printed an error message */
+            libreport_xfunc_die(); /* dd_opendir() already printed an error message */
 
         GList *iter = preferences->urp_auth_items;
         for ( ; iter != NULL; iter = g_list_next(iter))
         {
             const char *key = (const char *)iter->data;
-            char *value = dd_load_text_ext(dd, key,
+            g_autofree char *value = dd_load_text_ext(dd, key,
                     DD_LOAD_TEXT_RETURN_NULL_ON_FAILURE | DD_FAIL_QUIETLY_ENOENT);
 
             if (value == NULL)
@@ -724,7 +622,6 @@ ureport_from_dump_dir_ext(const char *dump_dir_path, const struct ureport_prefer
             }
 
             sr_report_add_auth(report, key, value);
-            free(value);
         }
 
         dd_close(dd);
@@ -736,14 +633,8 @@ ureport_from_dump_dir_ext(const char *dump_dir_path, const struct ureport_prefer
     return json_ureport;
 }
 
-char *
-ureport_from_dump_dir(const char *dump_dir_path)
-{
-    return ureport_from_dump_dir_ext(dump_dir_path, /*no preferences*/NULL);
-}
-
 struct post_state *
-ureport_do_post(const char *json, struct ureport_server_config *config,
+libreport_ureport_do_post(const char *json, struct ureport_server_config *config,
                 const char *url_sfx)
 {
     int flags = POST_WANT_BODY | POST_WANT_ERROR_MSG;
@@ -757,6 +648,7 @@ ureport_do_post(const char *json, struct ureport_server_config *config,
     {
         post_state->client_cert_path = config->ur_client_cert;
         post_state->client_key_path = config->ur_client_key;
+        post_state->cert_authority_cert_path = config->ur_cert_authority_cert;
     }
     else if (config->ur_username && config->ur_password)
     {
@@ -764,27 +656,16 @@ ureport_do_post(const char *json, struct ureport_server_config *config,
         post_state->password = config->ur_password;
     }
 
-    char **headers = xmalloc(sizeof(char *) * (3 + size_map_string(config->ur_http_headers)));
-    headers[0] = (char *)"Accept: application/json";
-    headers[1] = (char *)"Connection: close";
-    headers[2] = NULL;
-
-    if (config->ur_http_headers != NULL)
+    const char *headers[] =
     {
-        unsigned i = 2;
-        const char *header;
-        const char *value;
-        map_string_iter_t iter;
-        init_map_string_iter(&iter, config->ur_http_headers);
-        while (next_map_string_iter(&iter, &header, &value))
-            headers[i++] = xasprintf("%s: %s", header, value);
-        headers[i] = NULL;
-    }
-
-    char *dest_url = concat_path_file(config->ur_url, url_sfx);
+        "Accept: application/json",
+        "Connection: close",
+        NULL,
+    };
+    g_autofree char *dest_url = g_build_filename(config->ur_url ? config->ur_url : "", url_sfx, NULL);
 
     post_string_as_form_data(post_state, dest_url, "application/json",
-                     (const char **)headers, json);
+                             headers, json);
 
     /* Client authentication failed. Try again without client auth.
      * CURLE_SSL_CONNECT_ERROR - cert not found/server doesnt trust the CA
@@ -799,23 +680,17 @@ ureport_do_post(const char *json, struct ureport_server_config *config,
         post_state = new_post_state(flags);
 
         post_string_as_form_data(post_state, dest_url, "application/json",
-                         (const char **)headers, json);
+                                 headers, json);
 
     }
 
-    free(dest_url);
-
-    for (unsigned i = size_map_string(config->ur_http_headers); i != 0; --i)
-        free(headers[i + 1]);
-    free(headers);
-
     return post_state;
 }
 
 struct ureport_server_response *
-ureport_submit(const char *json, struct ureport_server_config *config)
+libreport_ureport_submit(const char *json, struct ureport_server_config *config)
 {
-    struct post_state *post_state = ureport_do_post(json, config, UREPORT_SUBMIT_ACTION);
+    struct post_state *post_state = libreport_ureport_do_post(json, config, UREPORT_SUBMIT_ACTION);
 
     if (post_state == NULL)
     {
@@ -823,8 +698,8 @@ ureport_submit(const char *json, struct ureport_server_config *config)
         return NULL;
     }
 
-    struct ureport_server_response *resp = ureport_server_response_from_reply(post_state, config);
-    free(post_state);
+    struct ureport_server_response *resp = libreport_ureport_server_response_from_reply(post_state, config);
+    free_post_state(post_state);
 
     return resp;
 }
@@ -834,28 +709,32 @@ ureport_json_attachment_new(const char *bthash, const char *type, const char *da
 {
     struct json_object *attachment = json_object_new_object();
     if (!attachment)
-        die_out_of_memory();
+        libreport_die_out_of_memory();
 
     ureport_add_str(attachment, "bthash", bthash);
     ureport_add_str(attachment, "type", type);
     ureport_add_str(attachment, "data", data);
 
-    char *result = xstrdup(json_object_to_json_string(attachment));
+    char *result = g_strdup(json_object_to_json_string(attachment));
     json_object_put(attachment);
 
     return result;
 }
 
 bool
-ureport_attach_string(const char *bthash, const char *type, const char *data,
-               struct ureport_server_config *config)
+ureport_attach_string(struct ureport_server_config *config,
+                      const char *bthash,
+                      const char *type,
+                      const char *data)
 {
-    char *json = ureport_json_attachment_new(bthash, type, data);
-    post_state_t *post_state = ureport_do_post(json, config, UREPORT_ATTACH_ACTION);
-    free(json);
+    g_autofree char *json = NULL;
+
+    json = ureport_json_attachment_new(bthash, type, data);
+
+    post_state_t *post_state = libreport_ureport_do_post(json, config, UREPORT_ATTACH_ACTION);
 
     struct ureport_server_response *resp =
-        ureport_server_response_from_reply(post_state, config);
+        libreport_ureport_server_response_from_reply(post_state, config);
     free_post_state(post_state);
     /* don't use str_bo_bool() because we require "true" string */
     const int result = !resp || resp->urr_is_error || strcmp(resp->urr_value, "true") != 0;
@@ -864,18 +743,25 @@ ureport_attach_string(const char *bthash, const char *type, const char *data,
         error_msg(_("The server at '%s' responded with an error: '%s'"),
                 config->ur_url, resp->urr_value);
 
-    ureport_server_response_free(resp);
+    libreport_ureport_server_response_free(resp);
 
     return result;
 }
 
 bool
-ureport_attach_int(const char *bthash, const char *type, int data,
-                    struct ureport_server_config *config)
+ureport_attach(struct ureport_server_config *config,
+               const char *bthash,
+               const char *type,
+               const char *format,
+               ...)
 {
-    char *data_str = xasprintf("%d", data);
-    const bool result = ureport_attach_string(bthash, type, data_str, config);
-    free(data_str);
+    va_list args;
+    g_autofree char *string = NULL;
 
-    return result;
+    va_start(args, format);
+
+    string = g_strdup_vprintf(format, args);
+    va_end(args);
+
+    return ureport_attach_string(config, bthash, type, string);
 }
