@@ -73,11 +73,11 @@ mantisbt_settings_free(mantisbt_settings_t *s)
     if (s == NULL)
         return;
 
-    free(s->m_login);
-    free(s->m_password);
-    free(s->m_project);
-    free(s->m_project_id);
-    free(s->m_project_version);
+    g_free(s->m_login);
+    g_free(s->m_password);
+    g_free(s->m_project);
+    g_free(s->m_project_id);
+    g_free(s->m_project_version);
 }
 
 /*
@@ -86,7 +86,7 @@ mantisbt_settings_free(mantisbt_settings_t *s)
 mantisbt_issue_info_t *
 mantisbt_issue_info_new()
 {
-    mantisbt_issue_info_t *info = xzalloc(sizeof(mantisbt_issue_info_t));
+    mantisbt_issue_info_t *info = g_malloc0(sizeof(*info));
     info->mii_id = -1;
     info->mii_dup_id = -1;
 
@@ -104,8 +104,8 @@ mantisbt_issue_info_free(mantisbt_issue_info_t *info)
     free(info->mii_reporter);
     free(info->mii_project);
 
-    list_free_with_free(info->mii_notes);
-    list_free_with_free(info->mii_attachments);
+    g_list_free_full(info->mii_notes, free);
+    g_list_free_full(info->mii_attachments, free);
 
     free(info);
 }
@@ -122,11 +122,13 @@ mantisbt_find_origin_bug_closed_duplicate(mantisbt_settings_t *settings, mantisb
         if (ii == MAX_HOPS)
             error_msg_and_die(_("MantisBT couldn't find parent of issue %d"), info->mii_id);
 
-        log("Issue %d is a duplicate, using parent issue %d", info_tmp->mii_id, info_tmp->mii_dup_id);
+        log_warning("Issue %d is a duplicate, using parent issue %d", info_tmp->mii_id, info_tmp->mii_dup_id);
         int issue_id = info_tmp->mii_dup_id;
 
         mantisbt_issue_info_free(info_tmp);
         info_tmp = mantisbt_get_issue_info(settings, issue_id);
+        if (info_tmp == NULL)
+            return NULL;
 
         // found a issue which is not CLOSED as DUPLICATE
         if (info_tmp->mii_dup_id == -1)
@@ -142,7 +144,7 @@ mantisbt_find_origin_bug_closed_duplicate(mantisbt_settings_t *settings, mantisb
 static soap_request_t *
 soap_request_new()
 {
-    soap_request_t *req = xzalloc(sizeof(*req));
+    soap_request_t *req = g_malloc0(sizeof(*req));
 
     return req;
 }
@@ -208,10 +210,9 @@ soap_node_get_child_node(xmlNodePtr parent, const char *name)
 soap_request_t *
 soap_request_new_for_method(const char *method)
 {
-    char *xml_str = xasprintf(SOAP_TEMPLATE, method, method);
+    g_autofree char *xml_str = g_strdup_printf(SOAP_TEMPLATE, method, method);
 
     xmlDocPtr doc = xmlParseDoc(BAD_CAST xml_str);
-    free(xml_str);
 
     if (doc == NULL)
         error_msg_and_die(_("SOAP: Failed to parse xml during creating request."));
@@ -328,12 +329,10 @@ soap_add_new_issue_parameters(soap_request_t *req,
 
     if (custom_fields_count > 0)
     {
-        char *type = xasprintf("%s[%i]", SOAP_CUSTOMFIELD, custom_fields_count);
+        g_autofree char *type = g_strdup_printf("%s[%i]", SOAP_CUSTOMFIELD, custom_fields_count);
 
         if (xmlNewProp(duphash_node, BAD_CAST "ns3:arrayType", BAD_CAST type) == NULL)
             error_msg_and_die(_("SOAP: Failed to create a new property in custom fields."));
-
-        free(type);
     }
 
     soap_node_add_child_node(issue_node, "os_build", SOAP_STRING, version);
@@ -359,34 +358,12 @@ soap_request_to_str(const soap_request_t *req)
         error_msg_and_die(_("SOAP: Failed to dump xml node."));
     }
 
-    char *ret = xasprintf("%s%s", XML_VERSION, (const char *) xmlBufferContent(buffer));
+    char *ret = g_strdup_printf("%s%s", XML_VERSION, (const char *) xmlBufferContent(buffer));
     xmlBufferFree(buffer);
 
     return ret;
 }
 
-#if 0
-void
-soap_request_print(soap_request_t *req)
-{
-    if (req == NULL || req->sr_root == NULL || req->sr_root->doc == NULL)
-        error_msg_and_die(_("SOAP: Failed to print SOAP string."));
-
-    xmlBufferPtr buffer = xmlBufferCreate();
-    int err = xmlNodeDump(buffer, req->sr_root->doc, req->sr_root, 1, /* formatting */ 0);
-    if (err == -1)
-    {
-        xmlBufferFree(buffer);
-        error_msg_and_die(_("Failed to dump xml node."));
-    }
-
-    puts((const char *) xmlBufferContent(buffer));
-
-    xmlBufferFree(buffer);
-    return;
-}
-#endif
-
 static bool
 reader_move_reader_if_node_type_is_element_with_name_and_verify_its_value(xmlTextReaderPtr reader, const char *name)
 {
@@ -472,7 +449,7 @@ response_values_at_depth_by_name(const char *xml, const char *name, int depth)
             continue;
 
         if ((value = xmlTextReaderConstValue(reader)) != NULL)
-            result = g_list_append(result, xstrdup((const char *) value));
+            result = g_list_append(result, g_strdup((const char *) value));
     }
     xmlFreeTextReader(reader);
 
@@ -598,14 +575,14 @@ static char *
 response_get_error_msg(const char *xml)
 {
     GList *l = response_values_at_depth_by_name(xml, "faultstring", 3);
-    return (l != NULL) ? xstrdup(l->data) : NULL;
+    return (l != NULL) ? g_strdup(l->data) : NULL;
 }
 
 static char *
 response_get_additioanl_information(const char *xml)
 {
     GList *l = response_values_at_depth_by_name(xml, "additional_information", -1);
-    return (l != NULL) ? xstrdup(l->data) : NULL;
+    return (l != NULL) ? g_strdup(l->data) : NULL;
 }
 
 void
@@ -633,22 +610,21 @@ mantisbt_result_free(mantisbt_result_t *result)
 mantisbt_result_t *
 mantisbt_soap_call(const mantisbt_settings_t *settings, const soap_request_t *req)
 {
-    char *request = soap_request_to_str(req);
+    g_autofree char *request = soap_request_to_str(req);
 
     const char *url = settings->m_mantisbt_soap_url;
 
-    mantisbt_result_t *result = xzalloc(sizeof(*result));
+    mantisbt_result_t *result = g_malloc0(sizeof(*result));
 
     if (url == NULL || request == NULL)
     {
         result->mr_error = -2;
-        result->mr_msg = xasprintf(_("Url or request isn't specified."));
-        free(request);
+        result->mr_msg = g_strdup_printf(_("Url or request isn't specified."));
 
         return result;
     }
 
-    char *url_copy = NULL;
+    g_autofree char *url_copy = NULL;
 
     int redirect_count = 0;
     char *errmsg;
@@ -670,7 +646,7 @@ redirect:
     {
     case 404:
         result->mr_error = -1;
-        result->mr_msg = xasprintf(_("Error in HTTP POST, "
+        result->mr_msg = g_strdup_printf(_("Error in HTTP POST, "
                         "HTTP code: 404 (Not found), URL:'%s'"), url);
         break;
     case 500:
@@ -683,8 +659,7 @@ redirect:
     case 305: /* "305 Use Proxy" */
         if (++redirect_count < 10 && location)
         {
-            free(url_copy);
-            url = url_copy = xstrdup(location);
+            url = url_copy = g_strdup(location);
             free_post_state(post_state);
             goto redirect;
         }
@@ -694,15 +669,15 @@ redirect:
         result->mr_error = -1;
         errmsg = post_state->curl_error_msg;
         if (errmsg && errmsg[0])
-            result->mr_msg = xasprintf(_("Error in MantisBT request at '%s': %s"), url, errmsg);
+            result->mr_msg = g_strdup_printf(_("Error in MantisBT request at '%s': %s"), url, errmsg);
         else
-            result->mr_msg = xasprintf(_("Error in MantisBT request at '%s'"), url);
+            result->mr_msg = g_strdup_printf(_("Error in MantisBT request at '%s'"), url);
         break;
 
     case 200:
     case 201:
         /* sent successfully */
-        result->mr_url = xstrdup(location); /* note: xstrdup(NULL) returns NULL */
+        result->mr_url = g_strdup(location); /* note: g_strdup(NULL) returns NULL */
     } /* switch (HTTP code) */
 
     result->mr_http_resp_code = post_state->http_resp_code;
@@ -710,8 +685,6 @@ redirect:
     post_state->body = NULL;
 
     free_post_state(post_state);
-    free(url_copy);
-    free(request);
 
     return result;
 }
@@ -727,7 +700,7 @@ mantisbt_attach_data(const mantisbt_settings_t *settings, const char *bug_id,
     soap_request_add_method_parameter(req, "name", SOAP_STRING, att_name);
 
     soap_request_add_method_parameter(req, "file_type", SOAP_STRING, "text");
-    soap_request_add_method_parameter(req, "content", SOAP_BASE64, encode_base64(data, size));
+    soap_request_add_method_parameter(req, "content", SOAP_BASE64, libreport_encode_base64(data, size));
 
     mantisbt_result_t *result = mantisbt_soap_call(settings, req);
     soap_request_free(req);
@@ -768,17 +741,15 @@ mantisbt_attach_fd(const mantisbt_settings_t *settings, const char *bug_id,
     }
     lseek(fd, 0, SEEK_SET);
 
-    char *data = xmalloc(size + 1);
-    ssize_t r = full_read(fd, data, size);
+    g_autofree char *data = g_malloc(size + 1);
+    ssize_t r = libreport_full_read(fd, data, size);
     if (r < 0)
     {
-        free(data);
         perror_msg(_("Can't read '%s'"), att_name);
         return -1;
     }
 
     int res = mantisbt_attach_data(settings, bug_id, att_name, data, size);
-    free(data);
     return res;
 }
 
@@ -862,6 +833,7 @@ mantisbt_search_by_abrt_hash(mantisbt_settings_t *settings, const char *abrt_has
     }
 
     GList *ids = response_get_main_ids_list(result->mr_body);
+    mantisbt_result_free(result);
 
     return ids;
 }
@@ -906,6 +878,7 @@ mantisbt_search_duplicate_issues(mantisbt_settings_t *settings, const char *cate
     }
 
     GList *ids = response_get_main_ids_list(result->mr_body);
+    mantisbt_result_free(result);
 
     return ids;
 }
@@ -927,14 +900,13 @@ custom_field_get_id_from_name(GList *ids, GList *names, const char *name)
 static void
 custom_field_ask(const char *name)
 {
-    char *msg = xasprintf(_("MantisBT doesn't contain custom field '%s', which is required for full functionality of the reporter. Do you still want to create a new issue?"), name);
-    int yes = ask_yes_no(msg);
-    free(msg);
+    g_autofree char *msg = g_strdup_printf(_("MantisBT doesn't contain custom field '%s', which is required for full functionality of the reporter. Do you still want to create a new issue?"), name);
+    int yes = libreport_ask_yes_no(msg);
 
     if (!yes)
     {
-        set_xfunc_error_retval(EXIT_CANCEL_BY_USER);
-        xfunc_die();
+        libreport_set_xfunc_error_retval(EXIT_CANCEL_BY_USER);
+        libreport_xfunc_die();
     }
 
     return;
@@ -977,7 +949,7 @@ mantisbt_create_new_issue(const mantisbt_settings_t *settings,
     const char *category = problem_data_get_content_or_NULL(problem_data, FILENAME_COMPONENT);
     const char *duphash = problem_data_get_content_or_NULL(problem_data, FILENAME_DUPHASH);
 
-    char *summary = shorten_string_to_length(problem_report_get_summary(pr), MAX_SUMMARY_LENGTH);
+    g_autofree char *summary = libreport_shorten_string_to_length(problem_report_get_summary(pr), MAX_SUMMARY_LENGTH);
 
     const char *description = problem_report_get_description(pr);
     const char *additional_information = problem_report_get_section(pr, PR_SEC_ADDITIONAL_INFO);
@@ -991,7 +963,6 @@ mantisbt_create_new_issue(const mantisbt_settings_t *settings,
 
     mantisbt_result_t *result = mantisbt_soap_call(settings, req);
     soap_request_free(req);
-    free(summary);
 
     if (result->mr_error == -1)
     {
@@ -1012,9 +983,8 @@ mantisbt_get_issue_info(const mantisbt_settings_t *settings, int issue_id)
     soap_request_t *req = soap_request_new_for_method("mc_issue_get");
     soap_request_add_credentials_parameter(req, settings);
 
-    char *issue_id_str = xasprintf("%d", issue_id);
+    g_autofree char *issue_id_str = g_strdup_printf("%d", issue_id);
     soap_request_add_method_parameter(req, "issue_id", SOAP_INTEGER, issue_id_str);
-    free(issue_id_str);
 
     mantisbt_result_t *result = mantisbt_soap_call(settings, req);
     soap_request_free(req);
@@ -1040,7 +1010,7 @@ mantisbt_get_issue_info(const mantisbt_settings_t *settings, int issue_id)
     issue_info->mii_dup_id = response_get_id_of_relatedto_issue(result->mr_body);
 
     if (strcmp(issue_info->mii_status, "closed") == 0
-        && strcmp(issue_info->mii_resolution, "duplicate") == 0
+        && (issue_info->mii_resolution != NULL && strcmp(issue_info->mii_resolution, "duplicate") == 0)
         && issue_info->mii_dup_id == -1 )
     {
         error_msg(_("Issue %i is CLOSED as DUPLICATE, but it has no DUPLICATE_ID"),
@@ -1055,7 +1025,7 @@ mantisbt_get_issue_info(const mantisbt_settings_t *settings, int issue_id)
     if (add_info != NULL)
         issue_info->mii_notes = g_list_append (issue_info->mii_notes, add_info);
     issue_info->mii_attachments = response_values_at_depth_by_name(result->mr_body, "filename", -1);
-    issue_info->mii_best_bt_rating = comments_find_best_bt_rating(issue_info->mii_notes);
+    issue_info->mii_best_bt_rating = libreport_comments_find_best_bt_rating(issue_info->mii_notes);
 
     mantisbt_result_free(result);
     return issue_info;
@@ -1067,7 +1037,7 @@ mantisbt_add_issue_note(const mantisbt_settings_t *settings, int issue_id, const
     soap_request_t *req = soap_request_new_for_method("mc_issue_note_add");
     soap_request_add_credentials_parameter(req, settings);
 
-    char *issue_id_str = xasprintf("%i", issue_id);
+    g_autofree char *issue_id_str = g_strdup_printf("%i", issue_id);
     soap_node_add_child_node(req->sr_method, "issue_id", SOAP_INTEGER, issue_id_str);
 
     xmlNodePtr note_node = soap_node_add_child_node(req->sr_method, "note", SOAP_ISSUENOTE, /* content */ NULL);
@@ -1075,7 +1045,6 @@ mantisbt_add_issue_note(const mantisbt_settings_t *settings, int issue_id, const
 
     mantisbt_result_t *result = mantisbt_soap_call(settings, req);
 
-    free(issue_id_str);
     soap_request_free(req);
 
     if (result->mr_error == -1)
@@ -1101,11 +1070,16 @@ mantisbt_get_project_id_from_name(mantisbt_settings_t *settings)
     soap_node_add_child_node(req->sr_method, "project_name", SOAP_STRING, settings->m_project);
 
     mantisbt_result_t *result = mantisbt_soap_call(settings, req);
+    soap_request_free(req);
 
     if (result->mr_http_resp_code != 200)
+    {
+        mantisbt_result_free(result);
         error_msg_and_die(_("Failed to get project id from name"));
+    }
 
     settings->m_project_id = response_get_return_value_as_string(result->mr_body);
+    mantisbt_result_free(result);
 
     return;
 }
