@@ -23,20 +23,20 @@
 #include "problem_report.h"
 
 static void
-parse_osinfo_for_mantisbt(map_string_t *osinfo, char** project, char** version)
+parse_osinfo_for_mantisbt(GHashTable *osinfo, char** project, char** version)
 {
-    const char *name = get_map_string_item_or_NULL(osinfo, "CENTOS_MANTISBT_PROJECT");
+    const char *name = g_hash_table_lookup(osinfo, "CENTOS_MANTISBT_PROJECT");
     if (!name)
-        name = get_map_string_item_or_NULL(osinfo, OSINFO_NAME);
+        name = g_hash_table_lookup(osinfo, OSINFO_NAME);
 
-    const char *version_id = get_map_string_item_or_NULL(osinfo, "CENTOS_MANTISBT_PROJECT_VERSION");
+    const char *version_id = g_hash_table_lookup(osinfo, "CENTOS_MANTISBT_PROJECT_VERSION");
     if (!version_id)
-        version_id = get_map_string_item_or_NULL(osinfo, OSINFO_VERSION_ID);
+        version_id = g_hash_table_lookup(osinfo, OSINFO_VERSION_ID);
 
     if (name && version_id)
     {
-        *project = xstrdup(name);
-        *version = xstrdup(version_id);
+        *project = g_strdup(name);
+        *version = g_strdup(version_id);
         return;
     }
 
@@ -48,10 +48,10 @@ parse_osinfo_for_mantisbt(map_string_t *osinfo, char** project, char** version)
 static char *
 ask_mantisbt_login(const char *message)
 {
-    char *login = ask(message);
+    char *login = libreport_ask(message);
     if (login == NULL || login[0] == '\0')
     {
-        set_xfunc_error_retval(EXIT_CANCEL_BY_USER);
+        libreport_set_xfunc_error_retval(EXIT_CANCEL_BY_USER);
         error_msg_and_die(_("Can't continue without login"));
     }
 
@@ -61,12 +61,10 @@ ask_mantisbt_login(const char *message)
 static char *
 ask_mantisbt_password(const char *message)
 {
-    char *password = ask_password(message);
-    /* TODO: this should be fixed in ask_password() as other tools have the same problem */
-    putchar('\n');
+    char *password = libreport_ask_password(message);
     if (password == NULL || password[0] == '\0')
     {
-        set_xfunc_error_retval(EXIT_CANCEL_BY_USER);
+        libreport_set_xfunc_error_retval(EXIT_CANCEL_BY_USER);
         error_msg_and_die(_("Can't continue without password"));
     }
 
@@ -76,16 +74,14 @@ ask_mantisbt_password(const char *message)
 static void
 ask_mantisbt_credentials(mantisbt_settings_t *settings, const char *pre_message)
 {
-    free(settings->m_login);
-    free(settings->m_password);
+    g_free(settings->m_login);
+    g_free(settings->m_password);
 
-    char *question = xasprintf("%s %s", pre_message, _("Please enter your MantisBT login:"));
+    g_autofree char *question = g_strdup_printf("%s %s", pre_message, _("Please enter your MantisBT login:"));
     settings->m_login = ask_mantisbt_login(question);
-    free(question);
 
-    question = xasprintf("%s %s '%s':", pre_message, _("Please enter the password for"), settings->m_login);
+    question = g_strdup_printf("%s %s '%s':", pre_message, _("Please enter the password for"), settings->m_login);
     settings->m_password = ask_mantisbt_password(question);
-    free(question);
 
     return;
 }
@@ -104,11 +100,11 @@ verify_credentials(mantisbt_settings_t *settings)
         mantisbt_result_t *result = mantisbt_soap_call(settings, req);
         soap_request_free(req);
 
-        if (g_verbose > 2)
+        if (libreport_g_verbose > 2)
         {
             GList *ids = response_get_main_ids_list(result->mr_body);
             if (ids != NULL)
-                log("%s", (char *)ids->data);
+                log_warning("%s", (char *)ids->data);
             response_values_free(ids);
         }
 
@@ -123,18 +119,24 @@ verify_credentials(mantisbt_settings_t *settings)
 }
 
 static void
-set_settings(mantisbt_settings_t *m, map_string_t *settings, struct dump_dir *dd)
+set_settings(mantisbt_settings_t *m, GHashTable *settings, struct dump_dir *dd)
 {
     const char *environ;
 
     environ = getenv("Mantisbt_Login");
-    m->m_login = xstrdup(environ ? environ : get_map_string_item_or_empty(settings, "Login"));
+    if (!environ)
+        environ = g_hash_table_lookup(settings, "Login");
+    m->m_login = g_strdup(environ ? environ : "");
 
     environ = getenv("Mantisbt_Password");
-    m->m_password = xstrdup(environ ? environ : get_map_string_item_or_empty(settings, "Password"));
+    if (!environ)
+        environ = g_hash_table_lookup(settings, "Password");
+    m->m_password = g_strdup(environ ? environ : "");
 
     environ = getenv("Mantisbt_MantisbtURL");
-    m->m_mantisbt_url = environ ? environ : get_map_string_item_or_empty(settings, "MantisbtURL");
+    if (!environ)
+        environ = g_hash_table_lookup(settings, "MantisbtURL");
+    m->m_mantisbt_url = environ ? environ : "";
     if (!m->m_mantisbt_url[0])
         m->m_mantisbt_url = "http://localhost/mantisbt";
     else
@@ -144,58 +146,60 @@ set_settings(mantisbt_settings_t *m, map_string_t *settings, struct dump_dir *dd
         if (last_slash && last_slash[1] == '\0')
             *last_slash = '\0';
     }
-    m->m_mantisbt_soap_url = concat_path_file(m->m_mantisbt_url, "api/soap/mantisconnect.php");
+    m->m_mantisbt_soap_url = g_build_filename(m->m_mantisbt_url, "api/soap/mantisconnect.php", NULL);
 
     environ = getenv("Mantisbt_Project");
     if (environ)
     {
-        m->m_project = xstrdup(environ);
+        m->m_project = g_strdup(environ);
         environ = getenv("Mantisbt_ProjectVersion");
         if (environ)
-            m->m_project_version = xstrdup(environ);
+            m->m_project_version = g_strdup(environ);
     }
     else
     {
-        const char *option = get_map_string_item_or_NULL(settings, "Project");
+        const char *option = g_hash_table_lookup(settings, "Project");
         if (option)
-            m->m_project = xstrdup(option);
-        option = get_map_string_item_or_NULL(settings, "ProjectVersion");
+            m->m_project = g_strdup(option);
+        option = g_hash_table_lookup(settings, "ProjectVersion");
         if (option)
-            m->m_project_version = xstrdup(option);
+            m->m_project_version = g_strdup(option);
     }
 
     if (!m->m_project || !*m->m_project) /* if not overridden or empty... */
     {
-        free(m->m_project);
-        free(m->m_project_version);
+        g_free(m->m_project);
+        g_free(m->m_project_version);
 
         if (dd != NULL)
         {
-            map_string_t *osinfo = new_map_string();
+            g_autoptr(GHashTable) osinfo = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
 
-            char *os_info_data = dd_load_text(dd, FILENAME_OS_INFO);
-            parse_osinfo(os_info_data, osinfo);
-            free(os_info_data);
+            g_autofree char *os_info_data = dd_load_text(dd, FILENAME_OS_INFO);
+            libreport_parse_osinfo(os_info_data, osinfo);
 
             parse_osinfo_for_mantisbt(osinfo, &m->m_project, &m->m_project_version);
-            free_map_string(osinfo);
         }
     }
 
     environ = getenv("Mantisbt_SSLVerify");
-    m->m_ssl_verify = string_to_bool(environ ? environ : get_map_string_item_or_empty(settings, "SSLVerify"));
+    if (!environ)
+        environ = g_hash_table_lookup(settings, "SSLVerify");
+    m->m_ssl_verify = libreport_string_to_bool(environ ? environ : "");
 
     environ = getenv("Mantisbt_DontMatchComponents");
-    m->m_DontMatchComponents = environ ? environ : get_map_string_item_or_empty(settings, "DontMatchComponents");
+    if (!environ)
+        environ = g_hash_table_lookup(settings, "DontMatchComponents");
+    m->m_DontMatchComponents = environ ? environ : "";
 
-    environ = getenv(CREATE_PRIVATE_TICKET);
-    if (environ)
-        m->m_create_private = string_to_bool(environ);
+    m->m_create_private = libreport_get_global_create_private_ticket();
 
     if (!m->m_create_private)
     {
         environ = getenv("Mantisbt_CreatePrivate");
-        m->m_create_private = string_to_bool(environ ? environ : get_map_string_item_or_empty(settings, "CreatePrivate"));
+        if (!environ)
+            environ = g_hash_table_lookup(settings, "CreatePrivate");
+        m->m_create_private = libreport_string_to_bool(environ ? environ : "");
     }
     log_notice("create private MantisBT ticket: '%s'", m->m_create_private ? "YES": "NO");
 }
@@ -211,7 +215,12 @@ int main(int argc, char **argv)
     textdomain(PACKAGE);
 #endif
 
-    const char *program_usage_string = _(
+    /* user's config, ~/.config/libreport/mantisbt.conf */
+    g_autofree char *user_conf = g_build_filename(g_get_user_config_dir(),
+                                                  "libreport", "mantisbt.conf",
+                                                  NULL);
+
+    g_autofree char *program_usage_string = g_strdup_printf(_(
         "\n& [-vf] [-c CONFFILE]... [-F FMTFILE] [-A FMTFILE2] -d DIR"
         "\nor:"
         "\n& [-v] [-c CONFFILE]... [-d DIR] -t[ID] FILE..."
@@ -247,15 +256,18 @@ int main(int argc, char **argv)
         "\nTRACKER_NAME to URL field. This option is applied only when a new issue is to be"
         "\nfiled. The default value is 'ABRT Server'"
         "\n"
-        "\nIf not specified, CONFFILE defaults to "CONF_DIR"/plugins/mantisb.conf"
+        "\nIf not specified, CONFFILE defaults to %1$s/plugins/mantisbt.conf"
+        "\nand %2$s."
         "\nIts lines should have 'PARAM = VALUE' format."
         "\nRecognized string parameters: MantisbtURL, Login, Password, Project, ProjectVersion."
         "\nRecognized boolean parameter (VALUE should be 1/0, yes/no): SSLVerify, CreatePrivate."
+        "\nUser's local configuration overrides the system wide configuration."
         "\nParameters can be overridden via $Mantisbt_PARAM environment variables."
         "\n"
-        "\nFMTFILE default to "CONF_DIR"/plugins/mantisbt_format.conf."
-        "\nFMTFILE2 default to "CONF_DIR"/plugins/mantisbt_formatdup.conf."
-    );
+        "\nFMTFILE default to %1$s/plugins/mantisbt_format.conf."
+        "\nFMTFILE2 default to %1$s/plugins/mantisbt_formatdup.conf."),
+        CONF_DIR,
+        user_conf);
 
     enum {
         OPT_v = 1 << 0,
@@ -281,7 +293,7 @@ int main(int argc, char **argv)
     mantisbt_settings_t mbt_settings = { 0 };
     /* Keep enum above and order of options below in sync! */
     struct options program_options[] = {
-        OPT__VERBOSE(&g_verbose),
+        OPT__VERBOSE(&libreport_g_verbose),
         OPT_STRING(   'd', NULL, &dump_dir_name , "DIR"    , _("Problem directory")),
         OPT_LIST(     'c', NULL, &conf_file     , "FILE"   , _("Configuration file (may be given many times)")),
         OPT_STRING(   'F', NULL, &fmt_file      , "FILE"   , _("Formatting file for initial comment")),
@@ -295,25 +307,28 @@ int main(int argc, char **argv)
         OPT_END()
     };
 
-    unsigned opts = parse_opts(argc, argv, program_options, program_usage_string);
+    unsigned opts = libreport_parse_opts(argc, argv, program_options, program_usage_string);
     argv += optind;
 
-    export_abrt_envvars(0);
+    libreport_export_abrt_envvars(0);
 
-    map_string_t *settings = new_map_string();
+    GHashTable *settings = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
+
+    if (!conf_file)
+    {
+        conf_file = g_list_append(conf_file, (char*) CONF_DIR"/plugins/mantisbt.conf");
+        conf_file = g_list_append(conf_file, user_conf);
+    }
+    while (conf_file)
+    {
+        char *fn = (char *)conf_file->data;
+        log_notice("Loading settings from '%s'", fn);
+        libreport_load_conf_file(fn, settings, /*skip key w/o values:*/ false);
+        log_debug("Loaded '%s'", fn);
+        conf_file = g_list_delete_link(conf_file, conf_file);
+    }
 
     {
-        if (!conf_file)
-            conf_file = g_list_append(conf_file, (char*) CONF_DIR"/plugins/mantisbt.conf");
-        while (conf_file)
-        {
-            char *fn = (char *)conf_file->data;
-            log_notice("Loading settings from '%s'", fn);
-            load_conf_file(fn, settings, /*skip key w/o values:*/ false);
-            log_debug("Loaded '%s'", fn);
-            conf_file = g_list_delete_link(conf_file, conf_file);
-        }
-
         struct dump_dir *dd = NULL;
         if (abrt_hash == NULL)
         {
@@ -327,7 +342,8 @@ int main(int argc, char **argv)
         /* WRONG! set_settings() does not copy the strings, it merely sets up pointers
          * to settings[] dictionary:
          */
-        /*free_map_string(settings);*/
+        /*if (settings)
+         * g_hash_table_destroy(settings);*/
     }
 
     /* No connection is opened between client and server. Users authentication
@@ -341,7 +357,7 @@ int main(int argc, char **argv)
 
     if (abrt_hash)
     {
-        log(_("Looking for similar problems in MantisBT"));
+        log_warning(_("Looking for similar problems in MantisBT"));
         GList *ids = mantisbt_search_by_abrt_hash(&mbt_settings, abrt_hash);
         mantisbt_settings_free(&mbt_settings);
 
@@ -358,24 +374,27 @@ int main(int argc, char **argv)
     if (opts & OPT_t)
     {
         if (!argv[0])
-            show_usage_and_die(program_usage_string, program_options);
+            libreport_show_usage_and_die(program_usage_string, program_options);
 
         if (!ticket_no)
         {
             struct dump_dir *dd = dd_opendir(dump_dir_name, /*flags:*/ 0);
+            g_autoptr(report_result_t) reported_to = NULL;
+            char *url;
+
             if (!dd)
-                xfunc_die();
-            report_result_t *reported_to = find_in_reported_to(dd, "MantisBT");
+                libreport_xfunc_die();
+
+            reported_to = libreport_find_in_reported_to(dd, "MantisBT");
+
             dd_close(dd);
 
-            if (!reported_to || !reported_to->url)
+            if (NULL == reported_to)
                 error_msg_and_die(_("Can't get MantisBT ID because this problem has not yet been reported to MantisBT."));
 
-            char *url = reported_to->url;
-            reported_to->url = NULL;
-            free_report_result(reported_to);
+            url = report_result_get_url(reported_to);
 
-            if (prefixcmp(url, mbt_settings.m_mantisbt_url) != 0)
+            if (!g_str_has_prefix(url, mbt_settings.m_mantisbt_url))
                 error_msg_and_die(_("This problem has been reported to MantisBT '%s' which differs from the configured MantisBT '%s'."), url, mbt_settings.m_mantisbt_url);
 
             ticket_no = strrchr(url, '=');
@@ -383,8 +402,8 @@ int main(int argc, char **argv)
                 error_msg_and_die(_("Malformed url to MantisBT '%s'."), url);
 
             /* won't ever call free on it - it simplifies the code a lot */
-            ticket_no = xstrdup(ticket_no + 1);
-            log(_("Using MantisBT ID '%s'"), ticket_no);
+            ticket_no = g_strdup(ticket_no + 1);
+            log_warning(_("Using MantisBT ID '%s'"), ticket_no);
         }
 
         /* Attach files to existing MantisBT issues */
@@ -392,7 +411,7 @@ int main(int argc, char **argv)
         {
             const char *path = *argv++;
             char *filename = basename(path);
-            log(_("Attaching file '%s' to issue %s"), filename, ticket_no);
+            log_warning(_("Attaching file '%s' to issue %s"), filename, ticket_no);
             mantisbt_attach_file(&mbt_settings, ticket_no, filename, path);
         }
 
@@ -404,40 +423,49 @@ int main(int argc, char **argv)
     if (!(opts & OPT_f))
     {
         struct dump_dir *dd = dd_opendir(dump_dir_name, /*flags:*/ 0);
+        g_autoptr(report_result_t) reported_to = NULL;
+        g_autofree char *url = NULL;
+
         if (!dd)
-            xfunc_die();
-        report_result_t *reported_to = find_in_reported_to(dd, "MantisBT");
+            libreport_xfunc_die();
+
+        reported_to = libreport_find_in_reported_to(dd, "MantisBT");
+
         dd_close(dd);
 
-        if (reported_to && reported_to->url)
+        if (NULL != reported_to)
         {
-            char *msg = xasprintf(_("This problem was already reported to MantisBT (see '%s')."
+            url = report_result_get_url(reported_to);
+        }
+        if (NULL != url)
+        {
+            g_autofree char *msg = NULL;
+
+            msg = g_strdup_printf(_("This problem was already reported to MantisBT (see '%s')."
                             " Do you still want to create a new issue?"),
-                            reported_to->url);
-            int yes = ask_yes_no(msg);
-            free(msg);
-            if (!yes)
+                            url);
+
+            if (!libreport_ask_yes_no(msg))
                 return 0;
         }
-        free_report_result(reported_to);
     }
 
-    problem_data_t *problem_data = create_problem_data_for_reporting(dump_dir_name);
+    g_autoptr(problem_data_t) problem_data = create_problem_data_for_reporting(dump_dir_name);
     if (!problem_data)
-        xfunc_die(); /* create_problem_data_for_reporting already emitted error msg */
+        libreport_xfunc_die(); /* create_problem_data_for_reporting already emitted error msg */
 
     const char *category = problem_data_get_content_or_die(problem_data, FILENAME_COMPONENT);
     const char *duphash   = problem_data_get_content_or_die(problem_data, FILENAME_DUPHASH);
 
     if (opts & OPT_D)
     {
-        problem_formatter_t *pf = problem_formatter_new();
+        g_autoptr(problem_formatter_t) pf = problem_formatter_new();
         problem_formatter_add_section(pf, PR_SEC_ADDITIONAL_INFO, /* optional section */ 0);
 
         if (problem_formatter_load_file(pf, fmt_file))
             error_msg_and_die("Invalid format file: %s", fmt_file);
 
-        problem_report_t *pr = NULL;
+        g_autoptr(problem_report_t) pr = NULL;
         if (problem_formatter_generate_report(pf, problem_data, &pr))
             error_msg_and_die("Failed to format issue report from problem data");
 
@@ -454,12 +482,10 @@ int main(int argc, char **argv)
         for (GList *a = problem_report_get_attachments(pr); a != NULL; a = g_list_next(a))
             printf(" %s\n", (const char *)a->data);
 
-        problem_report_free(pr);
-        problem_formatter_free(pf);
         exit(0);
     }
 
-     int bug_id = 0;
+     unsigned long bug_id = 0;
 
     /* If REMOTE_RESULT contains "DUPLICATE 12345", we consider it a dup of 12345
      * and won't search on MantisBT server.
@@ -471,7 +497,7 @@ int main(int argc, char **argv)
         char *cmd = strtok(remote_result, " \n");
         char *id = strtok(NULL, " \n");
 
-        if (!prefixcmp(cmd, "DUPLICATE"))
+        if (g_str_has_prefix(cmd, "DUPLICATE"))
         {
             errno = 0;
             char *e;
@@ -487,7 +513,7 @@ int main(int argc, char **argv)
     mantisbt_issue_info_t *ii;
     if (!bug_id)
     {
-        log(_("Checking for duplicates"));
+        log_warning(_("Checking for duplicates"));
 
         int existing_id = -1;
         int crossver_id = -1;
@@ -495,7 +521,7 @@ int main(int argc, char **argv)
             /* Figure out whether we want to match category
              * when doing dup search.
              */
-            const char *category_substitute = is_in_comma_separated_list(category, mbt_settings.m_DontMatchComponents) ? NULL : category;
+            const char *category_substitute = libreport_is_in_comma_separated_list(category, mbt_settings.m_DontMatchComponents) ? NULL : category;
 
             /* We don't do dup detection across versions (see below why),
              * but we do add a note if cross-version potential dup exists.
@@ -526,14 +552,14 @@ int main(int argc, char **argv)
         if (existing_id < 0)
         {
             /* Create new issue */
-            log(_("Creating a new issue"));
-            problem_formatter_t *pf = problem_formatter_new();
+            log_warning(_("Creating a new issue"));
+            g_autoptr(problem_formatter_t) pf = problem_formatter_new();
             problem_formatter_add_section(pf, PR_SEC_ADDITIONAL_INFO, 0);
 
             if (problem_formatter_load_file(pf, fmt_file))
                 error_msg_and_die(_("Invalid format file: %s"), fmt_file);
 
-            problem_report_t *pr = NULL;
+            g_autoptr(problem_report_t) pr = NULL;
             if (problem_formatter_generate_report(pf, problem_data, &pr))
                 error_msg_and_die(_("Failed to format problem data"));
 
@@ -542,33 +568,36 @@ int main(int argc, char **argv)
                         problem_report_get_buffer(pr, PR_SEC_DESCRIPTION),
                         "\nPotential duplicate: issue %u\n", crossver_id);
 
-            problem_formatter_free(pf);
-
             /* get tracker URL if exists */
             struct dump_dir *dd = dd_opendir(dump_dir_name, 0);
-            char *tracker_url = NULL;
+            g_autofree char *tracker_url = NULL;
             if (dd)
             {
-                report_result_t *reported_to = find_in_reported_to(dd, tracker_str);
+                g_autoptr(report_result_t) reported_to = NULL;
+                g_autofree char *url = NULL;
+
+                reported_to = libreport_find_in_reported_to(dd, tracker_str);
+
                 dd_close(dd);
 
-                if (reported_to && reported_to->url)
+                if (NULL != reported_to)
                 {
-                    log(_("Adding External URL to issue"));
-                    tracker_url = xstrdup(reported_to->url);
-                    free_report_result(reported_to);
+                    url = report_result_get_url(reported_to);
+                }
+                if (NULL != url)
+                {
+                    log_warning(_("Adding External URL to issue"));
+                    tracker_url = g_steal_pointer(&url);
                 }
             }
 
             int new_id = mantisbt_create_new_issue(&mbt_settings, problem_data, pr, tracker_url);
 
-            free(tracker_url);
-
             if (new_id == -1)
                 return EXIT_FAILURE;
 
-            log(_("Adding attachments to issue %i"), new_id);
-            char *new_id_str = xasprintf("%u", new_id);
+            log_warning(_("Adding attachments to issue %i"), new_id);
+            g_autofree char *new_id_str = g_strdup_printf("%u", new_id);
 
             for (GList *a = problem_report_get_attachments(pr); a != NULL; a = g_list_next(a))
             {
@@ -582,11 +611,9 @@ int main(int argc, char **argv)
                     mantisbt_attach_file(&mbt_settings, new_id_str, item_name, item->content);
             }
 
-            free(new_id_str);
-            problem_report_free(pr);
             ii = mantisbt_issue_info_new();
             ii->mii_id = new_id;
-            ii->mii_status = xstrdup("new");
+            ii->mii_status = g_strdup("new");
 
             goto finish;
         }
@@ -596,7 +623,7 @@ int main(int argc, char **argv)
 
     ii = mantisbt_get_issue_info(&mbt_settings, bug_id);
 
-    log(_("Bug is already reported: %i"), ii->mii_id);
+    log_warning(_("Bug is already reported: %i"), ii->mii_id);
 
     /* Follow duplicates */
     if ((strcmp(ii->mii_status, "closed") == 0)
@@ -619,21 +646,21 @@ int main(int argc, char **argv)
     const char *comment = problem_data_get_content_or_NULL(problem_data, FILENAME_COMMENT);
     if (comment && comment[0])
     {
-        problem_formatter_t *pf = problem_formatter_new();
+        g_autoptr(problem_formatter_t) pf = problem_formatter_new();
 
         if (problem_formatter_load_file(pf, fmt_file2))
             error_msg_and_die(_("Invalid duplicate format file: '%s"), fmt_file2);
 
-        problem_report_t *pr;
+        g_autoptr(problem_report_t) pr = NULL;
         if (problem_formatter_generate_report(pf, problem_data, &pr))
             error_msg_and_die(_("Failed to format duplicate comment from problem data"));
 
         const char *mbtcomment = problem_report_get_description(pr);
 
-        int dup_comment = is_comment_dup(ii->mii_notes, mbtcomment);
+        int dup_comment = libreport_is_comment_dup(ii->mii_notes, mbtcomment);
         if (!dup_comment)
         {
-            log(_("Adding new comment to issue %d"), ii->mii_id);
+            log_warning(_("Adding new comment to issue %d"), ii->mii_id);
             mantisbt_add_issue_note(&mbt_settings, ii->mii_id, mbtcomment);
 
             const char *bt = problem_data_get_content_or_NULL(problem_data, FILENAME_BACKTRACE);
@@ -641,42 +668,43 @@ int main(int argc, char **argv)
             const char *rating_str = problem_data_get_content_or_NULL(problem_data, FILENAME_RATING);
             /* python doesn't have rating file */
             if (rating_str)
-                rating = xatou(rating_str);
+            {
+                char *endptr;
+                long rtg = g_ascii_strtoull(rating_str, &endptr, 10);
+                if (rtg >= 0 && rtg <= UINT_MAX && rating_str != endptr)
+                    rating = (unsigned)rtg;
+                else
+                    error_msg_and_die("expected number in range <0, %d>: '%s'", UINT_MAX, rating_str);
+            }
             if (bt && rating > ii->mii_best_bt_rating)
             {
-                char *bug_id_str = xasprintf("%i", ii->mii_id);
+                g_autofree char *bug_id_str = g_strdup_printf("%i", ii->mii_id);
 
-                log(_("Attaching better backtrace"));
+                log_warning(_("Attaching better backtrace"));
 
                 // find unique filename of attachment
-                char *name = NULL;
+                g_autofree char *name = NULL;
                 for (int i = 0;; ++i)
                 {
                     if (i == 0)
-                        name = xasprintf("%s", FILENAME_BACKTRACE);
+                        name = g_strdup_printf("%s", FILENAME_BACKTRACE);
                     else
-                        name = xasprintf("%s%d", FILENAME_BACKTRACE, i);
+                        name = g_strdup_printf("%s%d", FILENAME_BACKTRACE, i);
 
                     if (g_list_find_custom(ii->mii_attachments, name, (GCompareFunc) strcmp) == NULL)
                         break;
 
-                    free(name);
+                    g_free(name);
                 }
                 mantisbt_attach_data(&mbt_settings, bug_id_str, name, bt, strlen(bt));
-
-                free(name);
-                free(bug_id_str);
             }
         }
         else
-            log(_("Found the same comment in the issue history, not adding a new one"));
-
-        problem_report_free(pr);
-        problem_formatter_free(pf);
+            log_warning(_("Found the same comment in the issue history, not adding a new one"));
     }
 
 finish:
-    log(_("Status: %s%s%s %s/view.php?id=%u"),
+    log_warning(_("Status: %s%s%s %s/view.php?id=%u"),
                 ii->mii_status,
                 ii->mii_resolution ? " " : "",
                 ii->mii_resolution ? ii->mii_resolution : "",
@@ -686,9 +714,16 @@ finish:
     struct dump_dir *dd = dd_opendir(dump_dir_name, /*flags:*/ 0);
     if (dd)
     {
-        char *msg = xasprintf("MantisBT: URL=%s/view.php?id=%u", mbt_settings.m_mantisbt_url, ii->mii_id);
-        add_reported_to(dd, msg);
-        free(msg);
+        report_result_t *result;
+
+        result = report_result_new_with_label_from_env("MantisBT");
+        g_autofree char *url = g_strdup_printf("%s/view.php?id=%u", mbt_settings.m_mantisbt_url, ii->mii_id);
+
+        report_result_set_url(result, url);
+
+        libreport_add_reported_to_entry(dd, result);
+
+        report_result_free(result);
         dd_close(dd);
     }
 
