@@ -19,7 +19,6 @@
 
 /* Bugzilla API doc:
  * http://www.bugzilla.org/docs/4.2/en/html/api/Bugzilla/WebService.html
- * http://www.bugzilla.org/docs/tip/en/html/api/Bugzilla/WebService.html
  *
  * To make libxmlrpc print debug info to stdout (stderr?),
  * export XMLRPC_TRACE_XML=1
@@ -34,8 +33,8 @@
 
 //#define DEBUG
 #ifdef DEBUG
-#define func_entry() log("-- %s", __func__)
-#define func_entry_str(x) log("-- %s\t%s", __func__, (x))
+#define func_entry() log_warning("-- %s", __func__)
+#define func_entry_str(x) log_warning("-- %s\t%s", __func__, (x))
 #else
 #define func_entry()
 #define func_entry_str(x)
@@ -45,7 +44,7 @@ struct bug_info *new_bug_info()
 {
     func_entry();
 
-    struct bug_info *bi = xzalloc(sizeof(struct bug_info));
+    struct bug_info *bi = g_new0(struct bug_info, 1);
     bi->bi_dup_id = -1;
 
     return bi;
@@ -58,14 +57,15 @@ void free_bug_info(struct bug_info *bi)
     if (!bi)
         return;
 
-    free(bi->bi_status);
-    free(bi->bi_resolution);
-    free(bi->bi_reporter);
-    free(bi->bi_product);
+    g_free(bi->bi_status);
+    g_free(bi->bi_resolution);
+    g_free(bi->bi_reporter);
+    g_free(bi->bi_product);
+    g_free(bi->bi_platform);
 
-    list_free_with_free(bi->bi_cc_list);
+    g_list_free_full(bi->bi_cc_list, g_free);
 
-    free(bi);
+    g_free(bi);
 }
 
 static GList *rhbz_comments(struct abrt_xmlrpc *ax, int bug_id)
@@ -95,9 +95,8 @@ static GList *rhbz_comments(struct abrt_xmlrpc *ax, int bug_id)
     xmlrpc_value *bugs_memb = rhbz_get_member("bugs", xml_response);
 
     /* Get hash value assigned to bug_id key */
-    char *item = xasprintf("%d", bug_id);
+    g_autofree char *item = g_strdup_printf("%d", bug_id);
     xmlrpc_value *item_memb = rhbz_get_member(item, bugs_memb);
-    free(item);
 
     /* Get array of comments */
     xmlrpc_value *comments_memb = rhbz_get_member("comments", item_memb);
@@ -193,12 +192,11 @@ bool rhbz_login(struct abrt_xmlrpc *ax, const char *login, const char *password)
         return false;
     }
 
-    char *token = rhbz_bug_read_item("token", result, RHBZ_READ_STR);
+    g_autofree char *token = rhbz_bug_read_item("token", result, RHBZ_READ_STR);
     if (token != NULL)
     {
         log_debug("Adding session param Bugzilla_token");
         abrt_xmlrpc_client_add_session_param_string(&env, ax, "Bugzilla_token", token);
-        free(token);
     }
 
 //TODO: with URL like http://bugzilla.redhat.com (that is, with http: instead of https:)
@@ -210,6 +208,21 @@ bool rhbz_login(struct abrt_xmlrpc *ax, const char *login, const char *password)
     return true;
 }
 
+
+bool rhbz_add_session_api_key(struct abrt_xmlrpc *ax, const char *api_key)
+{
+    func_entry();
+
+    xmlrpc_env env;
+    xmlrpc_env_init(&env);
+
+    log_debug("Adding session param Bugzilla_api_key");
+    abrt_xmlrpc_client_add_session_param_string(&env, ax, "Bugzilla_api_key", api_key);
+
+    return true;
+}
+
+
 xmlrpc_value *rhbz_get_member(const char *member, xmlrpc_value *xml)
 {
     func_entry_str(member);
@@ -271,7 +284,7 @@ unsigned rhbz_version(struct abrt_xmlrpc *ax)
 
     xmlrpc_value *result;
     result = abrt_xmlrpc_call(ax, "Bugzilla.version", "{}");
-    char *version = NULL;
+    g_autofree char *version = NULL;
     if (result)
         version = rhbz_bug_read_item("version", result, RHBZ_READ_STR);
     if (!result || !version)
@@ -293,8 +306,6 @@ unsigned rhbz_version(struct abrt_xmlrpc *ax)
         v[i] = strtoul(tok, NULL, 10);
     }
 
-    free(version);
-
     return BUGZILLA_VERSION(v[0], v[1], v[2]);
 }
 
@@ -311,17 +322,15 @@ int rhbz_get_bug_id_from_array0(xmlrpc_value* xml, unsigned ver)
     if (env.fault_occurred)
         abrt_xmlrpc_die(&env);
 
-    const char *id;
-    if (ver >= BUGZILLA_VERSION(4,2,0))
-        id = "id";
-    else
-        id = "bug_id";
+    const char *item_id = "id";
+    if (ver < BUGZILLA_VERSION(4,2,0))
+        item_id = "bug_id";
 
     xmlrpc_value *bug;
-    bug = rhbz_get_member(id, item);
+    bug = rhbz_get_member(item_id, item);
     xmlrpc_DECREF(item);
     if (!bug)
-        error_msg_and_die("Can't get member '%s' from bug data", id);
+        error_msg_and_die("Can't get member '%s' from bug data", item_id);
 
     int bug_id = -1;
     xmlrpc_read_int(&env, bug, &bug_id);
@@ -367,7 +376,7 @@ void *rhbz_bug_read_item(const char *memb, xmlrpc_value *xml, int flags)
 
     if (IS_READ_INT(flags))
     {
-        int *integer = xmalloc(sizeof(int));
+        int *integer = g_malloc(sizeof(*integer));
         xmlrpc_read_int(&env, member, integer);
         xmlrpc_DECREF(member);
         if (env.fault_occurred)
@@ -411,19 +420,20 @@ GList *rhbz_bug_cc(xmlrpc_value* result_xml)
         if (!item)
             continue;
 
-        const char* cc = NULL;
-        xmlrpc_read_string(&env, item, &cc);
+        char *cc = NULL;
+        xmlrpc_read_string(&env, item, (const char **)&cc);
         xmlrpc_DECREF(item);
         if (env.fault_occurred)
             abrt_xmlrpc_die(&env);
 
         if (*cc != '\0')
         {
-            cc_list = g_list_append(cc_list, (char*)cc);
+            cc_list = g_list_append(cc_list, cc);
             log_debug("member on cc is %s", cc);
             continue;
         }
-        free((char*)cc);
+
+        free(cc);
     }
     xmlrpc_DECREF(cc_member);
     return cc_list;
@@ -454,7 +464,7 @@ struct bug_info *rhbz_bug_info(struct abrt_xmlrpc *ax, int bug_id)
     int *ret = (int*)rhbz_bug_read_item("id", bug_item,
                                         RHBZ_MANDATORY_MEMB | RHBZ_READ_INT);
     bz->bi_id = *ret;
-    free(ret);
+    g_free(ret);
     bz->bi_product = rhbz_bug_read_item("product", bug_item,
                                         RHBZ_MANDATORY_MEMB | RHBZ_READ_STR);
     bz->bi_reporter = rhbz_bug_read_item("creator", bug_item,
@@ -480,7 +490,7 @@ struct bug_info *rhbz_bug_info(struct abrt_xmlrpc *ax, int bug_id)
     }
 
     bz->bi_dup_id = (ret) ? *ret: -1;
-    free(ret);
+    g_free(ret);
 
     bz->bi_cc_list = rhbz_bug_cc(bug_item);
 
@@ -514,28 +524,71 @@ int rhbz_new_bug(struct abrt_xmlrpc *ax,
                                                                 FILENAME_ARCHITECTURE);
     const char *duphash      = problem_data_get_content_or_NULL(problem_data,
                                                                 FILENAME_DUPHASH);
-//COMPAT, remove after 2.1 release
-    if (!duphash) duphash    = problem_data_get_content_or_NULL(problem_data,
-                                                                "global_uuid");
 
-    char *summary = shorten_string_to_length(bzsummary, MAX_SUMMARY_LENGTH);
+    g_autofree char *summary = libreport_shorten_string_to_length(bzsummary, MAX_SUMMARY_LENGTH);
 
-    char *status_whiteboard = xasprintf("abrt_hash:%s", duphash);
+    g_autoptr(GString) status_whiteboard = g_string_new(NULL);
+    g_string_append_printf(status_whiteboard, "abrt_hash:%s;", duphash);
+
+    {   /* Add fields from /etc/os-release to Whiteboard for simple metrics. */
+        g_autoptr(GHashTable) osinfo = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
+        problem_data_get_osinfo(problem_data, osinfo);
+
+        /* This is the highest abstraction level I am willing to introduce now.
+         *
+         * The lines below can be either reduced to the body of the for loop
+         * or the opts variable can be dynamically initialized
+         * or you can simply add an another /etc/os-release option name
+         * (e.g. BUILD_ID).
+         */
+        const char *const opts[] = { "VARIANT_ID", NULL };
+        for (const char *const *iter = opts; *iter != NULL; ++iter)
+        {
+            const char *v = g_hash_table_lookup(osinfo, *iter);
+            if (v != NULL)
+            {
+                /* semi-colon (;) is the delimiter because /etc/os-release *_ID
+                 * options does not permit the ';' character in values
+                 */
+                g_string_append_printf(status_whiteboard, "%s=%s;", *iter, v);
+            }
+        }
+    }
 
     xmlrpc_env env;
     xmlrpc_env_init(&env);
 
-    xmlrpc_value *params = abrt_xmlrpc_params_new(&env);
+    xmlrpc_value *params = abrt_xmlrpc_struct_new(&env);
 
-    abrt_xmlrpc_params_add_string(&env, params, "product", product);
-    abrt_xmlrpc_params_add_string(&env, params, "component", component);
-    abrt_xmlrpc_params_add_string(&env, params, "version", version);
-    abrt_xmlrpc_params_add_string(&env, params, "summary", summary);
-    abrt_xmlrpc_params_add_string(&env, params, "description", bzcomment);
-    abrt_xmlrpc_params_add_string(&env, params, "status_whiteboard", status_whiteboard);
+    abrt_xmlrpc_params_set_value_str(&env, params, "product", product);
+    abrt_xmlrpc_params_set_value_str(&env, params, "component", component);
+
+    if (strncmp(product, "Red Hat Enterprise Linux", strlen("Red Hat Enterprise Linux")) == 0)
+    {
+        xmlrpc_value *sc_array = rhbz_get_sub_components(ax, product, component);
+        const char *sub_component = NULL;
+        if (sc_array)
+            sub_component = rhbz_get_default_sub_component(component, sc_array);
+        if (sub_component)
+        {
+            xmlrpc_value *sub_components_struct = abrt_xmlrpc_struct_new(&env);
+            xmlrpc_value *sub_components_array = abrt_xmlrpc_array_new(&env);
+            abrt_xmlrpc_array_append_string(&env, sub_components_array, sub_component);
+            abrt_xmlrpc_params_set_value(&env, sub_components_struct, component, sub_components_array);
+            abrt_xmlrpc_params_set_value(&env, params, "sub_components", sub_components_struct);
+            xmlrpc_DECREF(sub_components_struct);
+            xmlrpc_DECREF(sub_components_array);
+            free((void *)sub_component);
+        }
+    }
+
+    abrt_xmlrpc_params_set_value_str(&env, params, "version", version);
+    abrt_xmlrpc_params_set_value_str(&env, params, "summary", summary);
+    abrt_xmlrpc_params_set_value_str(&env, params, "description", bzcomment);
+    abrt_xmlrpc_params_set_value_str(&env, params, "status_whiteboard", status_whiteboard->str);
 
     if(arch)
-        abrt_xmlrpc_params_add_string(&env, params, "platform", arch);
+        abrt_xmlrpc_params_set_value_str(&env, params, "platform", arch);
 
     if (private)
     {
@@ -546,13 +599,16 @@ int rhbz_new_bug(struct abrt_xmlrpc *ax,
             for (GList *l = group; l; l = l->next)
                 abrt_xmlrpc_array_append_string(&env, xmlrpc_groups, l->data);
 
-            abrt_xmlrpc_params_add_array(&env, params, "groups", xmlrpc_groups);
+            abrt_xmlrpc_params_set_value(&env, params, "groups", xmlrpc_groups);
 
             xmlrpc_DECREF(xmlrpc_groups);
         }
         else
         {
-            error_msg(_("A private ticket creation has been requested, but no groups were specified, please see https://github.com/abrt/abrt/wiki/FAQ#creating-private-bugzilla-tickets for more info"));
+            error_msg(_("A private ticket creation has been requested, but no groups were"
+                        " specified, please see"
+                        " https://github.com/abrt/abrt/wiki/FAQ#creating-private-bugzilla-tickets"
+                        " for more info"));
             return -1;
         }
     }
@@ -562,37 +618,33 @@ int rhbz_new_bug(struct abrt_xmlrpc *ax,
     xmlrpc_DECREF(params);
     xmlrpc_env_clean(&env);
 
-    free(status_whiteboard);
-    free(summary);
-
     if (!result)
         return -1;
 
-    int *r = rhbz_bug_read_item("id", result, RHBZ_MANDATORY_MEMB | RHBZ_READ_INT);
+    g_autofree int *r = rhbz_bug_read_item("id", result, RHBZ_MANDATORY_MEMB | RHBZ_READ_INT);
     xmlrpc_DECREF(result);
     int new_bug_id = *r;
-    free(r);
 
-    log(_("New bug id: %i"), new_bug_id);
+    log_warning(_("New bug id: %i"), new_bug_id);
     return new_bug_id;
 }
 
-/* suppress mail notify by {s:i} (nomail:1) (driven by flag) */
+/* suppress mail notify by {s:i} (minor_update:1) (driven by flag) */
 int rhbz_attach_blob(struct abrt_xmlrpc *ax, const char *bug_id,
                 const char *filename, const char *data, int data_len, int flags)
 {
     func_entry();
 
-    if (strlen(data) == 0)
+    if (0 == data_len)
     {
         log_notice("not attaching an empty file: '%s'", filename);
         /* Return SUCCESS */
         return 0;
     }
 
-    char *fn = xasprintf("File: %s", filename);
+    g_autofree char *fn = g_strdup_printf("File: %s", filename);
     xmlrpc_value* result;
-    int nomail_notify = !!IS_NOMAIL_NOTIFY(flags);
+    int minor_update = !!IS_MINOR_UPDATE(flags);
 
     /* http://www.bugzilla.org/docs/4.2/en/html/api/Bugzilla/WebService/Bug.html#add_attachment
      *
@@ -601,8 +653,10 @@ int rhbz_attach_blob(struct abrt_xmlrpc *ax, const char *bug_id,
      *   i -> integer, single argument (int value)
      *   6 -> base64,  two arguments (char* plain data which will be encoded by xmlrpc-c to base64,
      *                                size_t number of bytes to encode)
+     *
+     * Retry if another user/bot attempted to change the same data.
      */
-    result = abrt_xmlrpc_call(ax, "Bug.add_attachment", "{s:(s),s:s,s:s,s:s,s:6,s:i}",
+    result = abrt_xmlrpc_call_with_retry("query serialization error", ax, "Bug.add_attachment", "{s:(s),s:s,s:s,s:s,s:6,s:i}",
                 "ids", bug_id,
                 "summary", fn,
                 "file_name", filename,
@@ -612,13 +666,13 @@ int rhbz_attach_blob(struct abrt_xmlrpc *ax, const char *bug_id,
                  */
                 "data", data, (size_t)data_len,
 
-                /* Undocumented argument but it works with Red Hat Bugzilla version 4.2.4-7
-                 * and version 4.4.rc1.b02
+                /* If set to true, this is considered a minor update and no mail is sent to users who do not want
+                 * minor update emails. If current user is not in the minor_update_group, this parameter is simply
+                 * ignored.
                  */
-                "nomail", nomail_notify
+                "minor_update", minor_update
     );
 
-    free(fn);
     if (!result)
         return -1;
 
@@ -653,17 +707,15 @@ int rhbz_attach_fd(struct abrt_xmlrpc *ax, const char *bug_id,
 
 //TODO: need to have a method of attaching huge files (IOW: 1Gb read isn't good).
 
-    char *data = xmalloc(size + 1);
-    ssize_t r = full_read(fd, data, size);
+    g_autofree char *data = g_malloc(size);
+    ssize_t r = libreport_full_read(fd, data, size);
     if (r < 0)
     {
-        free(data);
         perror_msg("Can't read '%s'", att_name);
         return -1;
     }
 
     int res = rhbz_attach_blob(ax, bug_id, att_name, data, size, flags);
-    free(data);
     return res;
 }
 
@@ -695,7 +747,7 @@ struct bug_info *rhbz_find_origin_bug_closed_duplicate(struct abrt_xmlrpc *ax,
         if (ii == MAX_HOPS)
             error_msg_and_die(_("Bugzilla couldn't find parent of bug %d"), bi->bi_id);
 
-        log("Bug %d is a duplicate, using parent bug %d", bi_tmp->bi_id, bi_tmp->bi_dup_id);
+        log_warning("Bug %d is a duplicate, using parent bug %d", bi_tmp->bi_id, bi_tmp->bi_dup_id);
         int bug_id = bi_tmp->bi_dup_id;
 
         free_bug_info(bi_tmp);
@@ -709,25 +761,18 @@ struct bug_info *rhbz_find_origin_bug_closed_duplicate(struct abrt_xmlrpc *ax,
     return bi_tmp;
 }
 
-/* suppress mail notify by {s:i} (nomail:1) */
+/* suppress mail notify by {s:i} (minor_update:1) */
 void rhbz_mail_to_cc(struct abrt_xmlrpc *ax, int bug_id, const char *mail, int flags)
 {
     func_entry();
 
     xmlrpc_value *result;
-    int nomail_notify = !!IS_NOMAIL_NOTIFY(flags);
-#if 0 /* Obsolete API */
-    result = abrt_xmlrpc_call(ax, "Bug.update", "({s:i,s:{s:(s),s:i}})",
-                              "ids", bug_id,
-                              "updates", "add_cc", mail,
-                                         "nomail", nomail_notify
-    );
-#endif
+    int minor_update = !!IS_MINOR_UPDATE(flags);
     /* Bugzilla 4.0+ uses this API: */
-    result = abrt_xmlrpc_call(ax, "Bug.update", "{s:i,s:{s:(s),s:i}}",
+    result = abrt_xmlrpc_call(ax, "Bug.update", "{s:(i),s:{s:(s)},s:i}",
                               "ids", bug_id,
                               "cc", "add", mail,
-                                    "nomail", nomail_notify
+                                    "minor_update", minor_update
     );
     if (result)
         xmlrpc_DECREF(result);
@@ -758,12 +803,12 @@ void rhbz_add_comment(struct abrt_xmlrpc *ax, int bug_id, const char *comment,
     func_entry();
 
     int private = !!IS_PRIVATE(flags);
-    int nomail_notify = !!IS_NOMAIL_NOTIFY(flags);
+    int minor_update = !!IS_MINOR_UPDATE(flags);
 
     xmlrpc_value *result;
     result = abrt_xmlrpc_call(ax, "Bug.add_comment", "{s:i,s:s,s:b,s:i}",
                               "id", bug_id, "comment", comment,
-                              "private", private, "nomail", nomail_notify);
+                              "private", private, "minor_update", minor_update);
 
     if (result)
         xmlrpc_DECREF(result);
@@ -773,15 +818,40 @@ void rhbz_set_url(struct abrt_xmlrpc *ax, int bug_id, const char *url, int flags
 {
     func_entry();
 
-    const int nomail_notify = !!IS_NOMAIL_NOTIFY(flags);
-    xmlrpc_value *result = abrt_xmlrpc_call(ax, "Bug.update", "{s:i,s:s,s:i}",
+    const int minor_update = !!IS_MINOR_UPDATE(flags);
+    xmlrpc_value *result = abrt_xmlrpc_call(ax, "Bug.update", "{s:(i),s:s,s:i}",
                               "ids", bug_id,
                               "url", url,
 
-                /* Undocumented argument but it works with Red Hat Bugzilla version 4.2.4-7
-                 * and version 4.4.rc1.b02
+                /* If set to true, this is considered a minor update and no mail is sent to users who do not want
+                 * minor update emails. If current user is not in the minor_update_group, this parameter is simply
+                 * ignored.
                  */
-                              "nomail", nomail_notify
+                              "minor_update", minor_update
+    );
+
+    if (result)
+        xmlrpc_DECREF(result);
+}
+
+void rhbz_close_as_duplicate(struct abrt_xmlrpc *ax, int bug_id,
+                        int duplicate_bug,
+                        int flags)
+{
+    func_entry();
+
+    const int minor_update = !!IS_MINOR_UPDATE(flags);
+    xmlrpc_value *result = abrt_xmlrpc_call(ax, "Bug.update", "{s:(i),s:s,s:s,s:i,s:i}",
+                              "ids", bug_id,
+                              "status", "CLOSED",
+                              "resolution", "DUPLICATE",
+                              "dupe_of", duplicate_bug,
+
+                /* If set to true, this is considered a minor update and no mail is sent to users who do not want
+                 * minor update emails. If current user is not in the minor_update_group, this parameter is simply
+                 * ignored.
+                 */
+                              "minor_update", minor_update
     );
 
     if (result)
@@ -794,24 +864,23 @@ xmlrpc_value *rhbz_search_duphash(struct abrt_xmlrpc *ax,
                         const char *component,
                         const char *duphash)
 {
-    struct strbuf *query = strbuf_new();
+    GString *query = g_string_new(NULL);
 
-    strbuf_append_strf(query, "ALL whiteboard:\"%s\"", duphash);
+    g_string_append_printf(query, "ALL whiteboard:\"%s\"", duphash);
 
     if (product)
-        strbuf_append_strf(query, " product:\"%s\"", product);
+        g_string_append_printf(query, " product:\"%s\"", product);
 
     if (version)
-        strbuf_append_strf(query, " version:\"%s\"", version);
+        g_string_append_printf(query, " version:\"%s\"", version);
 
     if (component)
-        strbuf_append_strf(query, " component:\"%s\"", component);
+        g_string_append_printf(query, " component:\"%s\"", component);
 
-    char *s = strbuf_free_nobuf(query);
+    g_autofree char *s = g_string_free(query, FALSE);
     log_debug("search for '%s'", s);
-    xmlrpc_value *search = abrt_xmlrpc_call(ax, "Bug.search", "{s:s}", "quicksearch", s);
+    xmlrpc_value *search = abrt_xmlrpc_call(ax, "Bug.search", "{s:s,s:(s)}", "quicksearch", s, "include_fields", "id");
 
-    free(s);
     xmlrpc_value *bugs = rhbz_get_member("bugs", search);
     xmlrpc_DECREF(search);
 
@@ -820,3 +889,177 @@ xmlrpc_value *rhbz_search_duphash(struct abrt_xmlrpc *ax,
 
     return bugs;
 }
+
+xmlrpc_value *rhbz_get_sub_components(struct abrt_xmlrpc *ax, const char *product, const char *component)
+{
+    xmlrpc_env env;
+    xmlrpc_env_init(&env);
+
+    /* The "components" struct itself doesn't have to be specified
+     * in the include_fields array because we specify the struct's members.
+     * By specifying "components.name", we get the "components" struct ok because
+     * it is implied.
+     * For some reason, the same doesn't work for "components.sub_components".
+     * If we only specify "components.sub_components.name" in include_fields,
+     * we don't get the "sub_components" struct at all. That's why include_fields
+     * has to contain both "components.sub_components" _and_
+     * "components.sub_components.name" even though the latter is a member of the
+     * former.
+     */
+    xmlrpc_value *products_array = abrt_xmlrpc_call(ax, "Product.get", "{s:(s),s:(sss)}",
+                                                    "names", product,
+                                                    "include_fields",
+                                                    "components.name",
+                                                    "components.sub_components",
+                                                    "components.sub_components.name");
+    if (env.fault_occurred)
+        abrt_xmlrpc_die(&env);
+
+    /* Extract the one product struct from the array.
+     * 'Actually, in the present implementation, you must specify this asterisk.
+     * That's because xmlrpc_decompose_value() is not capable of determining
+     * whether your format string extracts all the members of a structure,
+     * which is what no asterisk would imply. Some day, this will get fixed.'
+     *
+     * http://xmlrpc-c.sourceforge.net/doc/libxmlrpc.html#decomposevalue
+     */
+    xmlrpc_value *product_struct = NULL;
+    xmlrpc_decompose_value(&env, products_array, "{s:(S),*}", "products", &product_struct);
+    if (env.fault_occurred)
+        abrt_xmlrpc_die(&env);
+    xmlrpc_DECREF(products_array);
+
+    xmlrpc_value *components = rhbz_get_member("components", product_struct);
+    if (env.fault_occurred)
+        abrt_xmlrpc_die(&env);
+
+    unsigned array_size = rhbz_array_size(components);
+
+    const char *current_component = NULL;
+    xmlrpc_value *component_info = NULL;
+
+    xmlrpc_value *component_nameV = NULL;
+    for (unsigned i = 0; i < array_size; i++)
+    {
+        xmlrpc_array_read_item(&env, components, i, &component_info);
+        if (env.fault_occurred)
+            abrt_xmlrpc_die(&env);
+
+        if (!component_info)
+            continue;
+
+        if (component_nameV)
+            xmlrpc_DECREF(component_nameV);
+        if (current_component)
+            free((void *)current_component);
+
+        xmlrpc_struct_find_value(&env, component_info, "name", &component_nameV);
+        if (env.fault_occurred)
+            abrt_xmlrpc_die(&env);
+        xmlrpc_read_string(&env, component_nameV, &current_component);
+
+        if (strcmp(current_component, component) == 0)
+            break;
+    }
+    if (component_nameV)
+        xmlrpc_DECREF(component_nameV);
+    if (current_component)
+        free((void *)current_component);
+    xmlrpc_DECREF(components);
+
+    xmlrpc_value *sub_components = NULL;
+    xmlrpc_struct_find_value(&env, component_info, "sub_components", &sub_components);
+    if (env.fault_occurred)
+        abrt_xmlrpc_die(&env);
+
+    xmlrpc_DECREF(component_info);
+
+    return sub_components;
+}
+
+char *rhbz_get_default_sub_component(const char *component, xmlrpc_value *sub_components)
+{
+    /* Here are some reasonably-looking default subcomponents for components
+     * that have those as of this writing. See if we have a hardcoded deafult,
+     * and check if it actually exists in the sub_components array. Otherwise
+     * just pick whatever comes last.
+     */
+
+    const char *defaults[][2] =
+    {
+            {"binutils",      "system-version"},
+            {"Documentation", "default"},
+            {"dwz",           "system-version"},
+            {"dynist",        "system-version"},
+            {"elfutils",      "system-version"},
+            {"gcc",           "system-version"},
+            {"gdb",           "system-version"},
+            {"kernel",        "Other"},
+            {"kernel-rt",     "Other"},
+            {"kpatch",        "kpatch-utility"},
+            {"ltrace",        "system-version"},
+            {"lvm2",          "Default / Unclassified"},
+            {"make",          "system-version"},
+            {"systemtap",     "system-version"},
+            {"test",          "sub1"},
+            {"valgrind",      "system-version"},
+            {"virtio-win",    "distribution"}
+    };
+    char *candidate = NULL;
+
+    for (int i = 0; i < G_N_ELEMENTS(defaults); i++)
+        if (strcmp(component, defaults[i][0]) == 0)
+        {
+            candidate = (char*)defaults[i][1];
+            break;
+        }
+
+    const char *sc_str_name = NULL;
+    xmlrpc_value *sc_struct = NULL;
+    xmlrpc_value *sc_name = NULL;
+
+    xmlrpc_env env;
+    xmlrpc_env_init(&env);
+
+    unsigned sc_array_size = rhbz_array_size(sub_components);
+    if (candidate)
+    {
+        for (unsigned i = 0; i < sc_array_size; i++)
+        {
+            if (sc_struct)
+                xmlrpc_DECREF(sc_struct);
+            if (sc_name)
+                xmlrpc_DECREF(sc_name);
+            if (sc_str_name)
+                free((void *)sc_str_name);
+
+            xmlrpc_array_read_item(&env, sub_components, i, &sc_struct);
+            xmlrpc_struct_find_value(&env, sc_struct, "name", &sc_name);
+            xmlrpc_read_string(&env, sc_name, &sc_str_name);
+            if (strcmp(candidate, sc_str_name) == 0)
+            {
+                if (sc_struct)
+                    xmlrpc_DECREF(sc_struct);
+                if (sc_name)
+                    xmlrpc_DECREF(sc_name);
+                break;
+            }
+        }
+    }
+    else
+    {
+        if (sc_array_size) {
+            xmlrpc_array_read_item(&env, sub_components, sc_array_size - 1, &sc_struct);
+            if (xmlrpc_struct_has_key(&env, sc_struct, "name")) {
+                xmlrpc_struct_find_value(&env, sc_struct, "name", &sc_name);
+                xmlrpc_read_string(&env, sc_name, &sc_str_name);
+            }
+            if (sc_struct)
+                xmlrpc_DECREF(sc_struct);
+            if (sc_name)
+                xmlrpc_DECREF(sc_name);
+        }
+    }
+
+    return (char *)sc_str_name;
+}
